@@ -2,7 +2,8 @@ package gatekeeper
 
 import (
 	"context"
-	"fmt"
+	"reflect"
+	"strings"
 	"time"
 
 	. "github.com/onsi/ginkgo"
@@ -11,29 +12,43 @@ import (
 	testClient "github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/client"
 	"github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/namespaces"
 	"github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/pods"
+	"github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/utils"
 	corev1 "k8s.io/api/core/v1"
-	apiextensionsv1beta1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
+	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
-	"k8s.io/apimachinery/pkg/runtime"
+	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
+	"k8s.io/apimachinery/pkg/types"
 	k8sClient "sigs.k8s.io/controller-runtime/pkg/client"
 
 	gkopv1alpha "github.com/gatekeeper/gatekeeper-operator/api/v1alpha1"
+
+	constraints "github.com/open-policy-agent/frameworks/constraint/pkg/apis/templates/v1beta1"
 	gkv1alpha "github.com/open-policy-agent/gatekeeper/apis/mutations/v1alpha1"
+	gkmatch "github.com/open-policy-agent/gatekeeper/pkg/mutation/match"
+	gktypes "github.com/open-policy-agent/gatekeeper/pkg/mutation/types"
+	gkutil "github.com/open-policy-agent/gatekeeper/pkg/util"
+
+	admission "k8s.io/api/admissionregistration/v1"
 )
 
 const (
-	// TestingNamespace is the namespace for resources in this test
-	TestingNamespace = "gatekeeper-testing"
+	testingNamespace              = utils.GatekeeperTestingNamespace
+	mutationIncludedNamespace     = utils.GatekeeperMutationIncludedNamespace
+	mutationExcludedNamespace     = utils.GatekeeperMutationExcludedNamespace
+	mutationEnabledNamespace      = utils.GatekeeperMutationEnabledNamespace
+	mutationDisabledNamespace     = utils.GatekeeperMutationDisabledNamespace
+	testObjectNamespace           = utils.GatekeeperTestObjectNamespace
+	constraintValidationNamespace = utils.GatekeeperConstraintValidationNamespace
 )
 
 var _ = Describe("gatekeeper", func() {
 	client := testClient.Client
 
 	AfterEach(func() {
-		err := deletePods(TestingNamespace, client)
+		err := deletePods(testingNamespace, client)
 		Expect(err).NotTo(HaveOccurred())
 
-		namespacesUsed := []string{"mutation-included", "mutation-excluded", "gk-test-object", "test-mutate", "test-no-mutate"}
+		namespacesUsed := []string{mutationIncludedNamespace, mutationExcludedNamespace, mutationEnabledNamespace, mutationDisabledNamespace, testObjectNamespace, constraintValidationNamespace}
 
 		for _, namespace := range namespacesUsed {
 			if namespaces.Exists(namespace, client) {
@@ -56,6 +71,24 @@ var _ = Describe("gatekeeper", func() {
 			err := namespaces.WaitForDeletion(client, namespace, time.Minute)
 			Expect(err).ToNot(HaveOccurred())
 		}
+
+		gkConfig := &gkopv1alpha.Gatekeeper{
+			ObjectMeta: metav1.ObjectMeta{
+				Name: "gatekeeper",
+			},
+		}
+
+		gkConfigKey := k8sClient.ObjectKeyFromObject(gkConfig)
+		Expect(err).ToNot(HaveOccurred())
+
+		err = client.Get(context.Background(), gkConfigKey, gkConfig)
+		Expect(err).ToNot(HaveOccurred())
+
+		if gkConfig.Spec.Webhook != nil && gkConfig.Spec.Webhook.NamespaceSelector != nil {
+			gkConfig.Spec.Webhook.NamespaceSelector = nil
+			err = client.Update(context.Background(), gkConfig, &k8sClient.UpdateOptions{})
+			Expect(err).ToNot(HaveOccurred())
+		}
 	})
 
 	Context("mutation", func() {
@@ -67,10 +100,10 @@ var _ = Describe("gatekeeper", func() {
 							Name: "add-label",
 						},
 						Spec: gkv1alpha.AssignMetadataSpec{
-							Match: gkv1alpha.Match{
-								Scope:      apiextensionsv1beta1.NamespaceScoped,
-								Namespaces: []string{TestingNamespace},
-								Kinds: []gkv1alpha.Kinds{
+							Match: gkmatch.Match{
+								Scope:      apiextensionsv1.NamespaceScoped,
+								Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+								Kinds: []gkmatch.Kinds{
 									{
 										APIGroups: []string{""},
 										Kinds:     []string{"Pod"},
@@ -79,8 +112,10 @@ var _ = Describe("gatekeeper", func() {
 							},
 							Location: "metadata.labels.mutated",
 							Parameters: gkv1alpha.MetadataParameters{
-								Assign: runtime.RawExtension{
-									Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+								Assign: gkv1alpha.AssignField{
+									Value: &gktypes.Anything{
+										Value: "true",
+									},
 								},
 							},
 						},
@@ -90,10 +125,10 @@ var _ = Describe("gatekeeper", func() {
 							Name: "add-annotation",
 						},
 						Spec: gkv1alpha.AssignMetadataSpec{
-							Match: gkv1alpha.Match{
-								Scope:      apiextensionsv1beta1.NamespaceScoped,
-								Namespaces: []string{TestingNamespace},
-								Kinds: []gkv1alpha.Kinds{
+							Match: gkmatch.Match{
+								Scope:      apiextensionsv1.NamespaceScoped,
+								Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+								Kinds: []gkmatch.Kinds{
 									{
 										APIGroups: []string{""},
 										Kinds:     []string{"Pod"},
@@ -102,8 +137,10 @@ var _ = Describe("gatekeeper", func() {
 							},
 							Location: "metadata.annotations.mutated",
 							Parameters: gkv1alpha.MetadataParameters{
-								Assign: runtime.RawExtension{
-									Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+								Assign: gkv1alpha.AssignField{
+									Value: &gktypes.Anything{
+										Value: "true",
+									},
 								},
 							},
 						},
@@ -115,16 +152,18 @@ var _ = Describe("gatekeeper", func() {
 					Expect(err).NotTo(HaveOccurred())
 				}
 
-				pod := pods.DefinePod(TestingNamespace)
+				pod := pods.DefinePod(testingNamespace)
 				err := client.Create(context.Background(), pod)
 				Expect(err).NotTo(HaveOccurred())
 
-				podKey, err := k8sClient.ObjectKeyFromObject(pod)
+				podKey := k8sClient.ObjectKeyFromObject(pod)
 				Expect(err).ToNot(HaveOccurred())
-				err = client.Get(context.Background(), podKey, pod)
-				Expect(err).ToNot(HaveOccurred())
-				Expect(pod.GetLabels()["mutated"]).To(Equal("true"))
-				Expect(pod.GetAnnotations()["mutated"]).To(Equal("true"))
+
+				Eventually(func() bool {
+					err := client.Get(context.Background(), podKey, pod)
+					Expect(err).ToNot(HaveOccurred())
+					return pod.GetLabels()["mutated"] == "true" && pod.GetAnnotations()["mutated"] == "true"
+				}, 10*time.Second, 2*time.Second).Should(Equal(true), "Mutations are not applied")
 			},
 		)
 
@@ -136,10 +175,10 @@ var _ = Describe("gatekeeper", func() {
 							Name: "mutate-label",
 						},
 						Spec: gkv1alpha.AssignMetadataSpec{
-							Match: gkv1alpha.Match{
-								Scope:      apiextensionsv1beta1.NamespaceScoped,
-								Namespaces: []string{TestingNamespace},
-								Kinds: []gkv1alpha.Kinds{
+							Match: gkmatch.Match{
+								Scope:      apiextensionsv1.NamespaceScoped,
+								Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+								Kinds: []gkmatch.Kinds{
 									{
 										APIGroups: []string{""},
 										Kinds:     []string{"Pod"},
@@ -148,8 +187,10 @@ var _ = Describe("gatekeeper", func() {
 							},
 							Location: "metadata.labels.mutated",
 							Parameters: gkv1alpha.MetadataParameters{
-								Assign: runtime.RawExtension{
-									Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+								Assign: gkv1alpha.AssignField{
+									Value: &gktypes.Anything{
+										Value: "true",
+									},
 								},
 							},
 						},
@@ -159,10 +200,10 @@ var _ = Describe("gatekeeper", func() {
 							Name: "mutate-annotation",
 						},
 						Spec: gkv1alpha.AssignMetadataSpec{
-							Match: gkv1alpha.Match{
-								Scope:      apiextensionsv1beta1.NamespaceScoped,
-								Namespaces: []string{TestingNamespace},
-								Kinds: []gkv1alpha.Kinds{
+							Match: gkmatch.Match{
+								Scope:      apiextensionsv1.NamespaceScoped,
+								Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+								Kinds: []gkmatch.Kinds{
 									{
 										APIGroups: []string{""},
 										Kinds:     []string{"Pod"},
@@ -171,8 +212,10 @@ var _ = Describe("gatekeeper", func() {
 							},
 							Location: "metadata.annotations.mutated",
 							Parameters: gkv1alpha.MetadataParameters{
-								Assign: runtime.RawExtension{
-									Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+								Assign: gkv1alpha.AssignField{
+									Value: &gktypes.Anything{
+										Value: "true",
+									},
 								},
 							},
 						},
@@ -184,7 +227,7 @@ var _ = Describe("gatekeeper", func() {
 					Expect(err).NotTo(HaveOccurred())
 				}
 
-				pod := pods.DefinePod(TestingNamespace)
+				pod := pods.DefinePod(testingNamespace)
 				labels := map[string]string{
 					"mutated": "false",
 				}
@@ -196,7 +239,7 @@ var _ = Describe("gatekeeper", func() {
 				err := client.Create(context.Background(), pod)
 				Expect(err).NotTo(HaveOccurred())
 
-				podKey, err := k8sClient.ObjectKeyFromObject(pod)
+				podKey := k8sClient.ObjectKeyFromObject(pod)
 				Expect(err).ToNot(HaveOccurred())
 				err = client.Get(context.Background(), podKey, pod)
 				Expect(err).ToNot(HaveOccurred())
@@ -212,10 +255,10 @@ var _ = Describe("gatekeeper", func() {
 					Name: "mutation-b",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{TestingNamespace},
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -224,8 +267,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutated-by",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"b\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "b",
+							},
 						},
 					},
 				},
@@ -234,13 +279,13 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Creating test-pod-b")
-			testPodB := pods.DefinePod(TestingNamespace)
+			testPodB := pods.DefinePod(testingNamespace)
 			testPodB.SetName("test-pod-b")
 			err = client.Create(context.Background(), testPodB)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that test-pod-b was labeled")
-			testPodBKey, err := k8sClient.ObjectKeyFromObject(testPodB)
+			testPodBKey := k8sClient.ObjectKeyFromObject(testPodB)
 			Expect(err).NotTo(HaveOccurred())
 			err = client.Get(context.Background(), testPodBKey, testPodB)
 			Expect(err).NotTo(HaveOccurred())
@@ -252,10 +297,10 @@ var _ = Describe("gatekeeper", func() {
 					Name: "mutation-a",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{TestingNamespace},
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -264,8 +309,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutated-by",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"a\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "a",
+							},
 						},
 					},
 				},
@@ -274,13 +321,13 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Creating test-pod-a")
-			testPodA := pods.DefinePod(TestingNamespace)
+			testPodA := pods.DefinePod(testingNamespace)
 			testPodA.SetName("test-pod-a")
 			err = client.Create(context.Background(), testPodA)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that test-pod-a was labeled")
-			testPodAKey, err := k8sClient.ObjectKeyFromObject(testPodA)
+			testPodAKey := k8sClient.ObjectKeyFromObject(testPodA)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), testPodAKey, testPodA)
 			Expect(err).ToNot(HaveOccurred())
@@ -292,10 +339,10 @@ var _ = Describe("gatekeeper", func() {
 					Name: "mutation-c",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{TestingNamespace},
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -304,8 +351,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutated-by",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"c\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "c",
+							},
 						},
 					},
 				},
@@ -314,13 +363,13 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Creating test-pod-c")
-			testPodC := pods.DefinePod(TestingNamespace)
+			testPodC := pods.DefinePod(testingNamespace)
 			testPodC.SetName("test-pod-c")
 			err = client.Create(context.Background(), testPodC)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that test-pod-c was labeled")
-			testPodCKey, err := k8sClient.ObjectKeyFromObject(testPodC)
+			testPodCKey := k8sClient.ObjectKeyFromObject(testPodC)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), testPodCKey, testPodC)
 			Expect(err).ToNot(HaveOccurred())
@@ -334,10 +383,10 @@ var _ = Describe("gatekeeper", func() {
 					Name: "mutation-version",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{TestingNamespace},
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -346,8 +395,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutation-version",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"0\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "0",
+							},
 						},
 					},
 				},
@@ -356,41 +407,48 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Creating test-pod-version-0")
-			testPodA := pods.DefinePod(TestingNamespace)
+			testPodA := pods.DefinePod(testingNamespace)
 			testPodA.SetName("test-pod-version-0")
 			err = client.Create(context.Background(), testPodA)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that test-pod-version-0 was labeled with mutation-version: 0")
-			testPodAKey, err := k8sClient.ObjectKeyFromObject(testPodA)
+			testPodAKey := k8sClient.ObjectKeyFromObject(testPodA)
 			Expect(err).NotTo(HaveOccurred())
 			err = client.Get(context.Background(), testPodAKey, testPodA)
 			Expect(err).NotTo(HaveOccurred())
 			Expect(testPodA.GetLabels()["mutation-version"]).To(Equal("0"))
 
 			By("Updating assignMetadata to mutation-version: 1")
-			newValue := runtime.RawExtension{
-				Raw: []byte(fmt.Sprintf("{\"value\":\"1\"}")),
+			newValue := gkv1alpha.AssignField{
+				Value: &gktypes.Anything{
+					Value: "1",
+				},
 			}
+
+			amKey := k8sClient.ObjectKeyFromObject(am)
+			err = client.Get(context.Background(), amKey, am)
+			Expect(err).ToNot(HaveOccurred())
+
 			am.Spec.Parameters.Assign = newValue
 			err = client.Update(context.Background(), am)
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Asserting that mutation-vesrion was updated to: 1")
-			amKey, err := k8sClient.ObjectKeyFromObject(am)
+			amKey = k8sClient.ObjectKeyFromObject(am)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), amKey, am)
 			Expect(err).NotTo(HaveOccurred())
 			Expect(am.Spec.Parameters.Assign).To(Equal(newValue))
 
 			By("Creating test-pod-version-1")
-			testPodB := pods.DefinePod(TestingNamespace)
+			testPodB := pods.DefinePod(testingNamespace)
 			testPodB.SetName("test-pod-version-1")
 			err = client.Create(context.Background(), testPodB)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that test-pod-version-1 was labeled with mutation-version: 1")
-			testPodBKey, err := k8sClient.ObjectKeyFromObject(testPodB)
+			testPodBKey := k8sClient.ObjectKeyFromObject(testPodB)
 			Expect(err).NotTo(HaveOccurred())
 			err = client.Get(context.Background(), testPodBKey, testPodB)
 			Expect(err).NotTo(HaveOccurred())
@@ -404,10 +462,10 @@ var _ = Describe("gatekeeper", func() {
 					Name: "mutation-deleted",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{TestingNamespace},
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{testingNamespace},
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -416,8 +474,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutated",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
@@ -426,13 +486,13 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Creating pod before-delete")
-			podBeforeDelete := pods.DefinePod(TestingNamespace)
+			podBeforeDelete := pods.DefinePod(testingNamespace)
 			podBeforeDelete.SetName("before-delete")
 			err = client.Create(context.Background(), podBeforeDelete)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that pod before-delete was labeled")
-			podBeforeDeleteKey, err := k8sClient.ObjectKeyFromObject(podBeforeDelete)
+			podBeforeDeleteKey := k8sClient.ObjectKeyFromObject(podBeforeDelete)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), podBeforeDeleteKey, podBeforeDelete)
 			Expect(err).ToNot(HaveOccurred())
@@ -443,13 +503,13 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Creating pod after-delete")
-			podAfterDelete := pods.DefinePod(TestingNamespace)
+			podAfterDelete := pods.DefinePod(testingNamespace)
 			podAfterDelete.SetName("after-delete")
 			err = client.Create(context.Background(), podAfterDelete)
 			Expect(err).NotTo(HaveOccurred())
 
 			By("Asserting that pod after-delete was not labeld by the deleted policy")
-			podAfterDeleteKey, err := k8sClient.ObjectKeyFromObject(podAfterDelete)
+			podAfterDeleteKey := k8sClient.ObjectKeyFromObject(podAfterDelete)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), podAfterDeleteKey, podAfterDelete)
 			Expect(err).ToNot(HaveOccurred())
@@ -487,8 +547,8 @@ var _ = Describe("gatekeeper", func() {
 					Name: "all-selected",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope: apiextensionsv1beta1.ResourceScope("*"),
+					Match: gkmatch.Match{
+						Scope: apiextensionsv1.ResourceScope("*"),
 						LabelSelector: &metav1.LabelSelector{
 							MatchLabels: map[string]string{
 								"selected": "true",
@@ -497,8 +557,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.all-selected",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
@@ -511,18 +573,21 @@ var _ = Describe("gatekeeper", func() {
 					Name: "namespace-included",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope:      apiextensionsv1beta1.NamespaceScoped,
-						Namespaces: []string{includedNamepsace.GetName()},
+					Match: gkmatch.Match{
+						Scope:      apiextensionsv1.NamespaceScoped,
+						Namespaces: []gkutil.PrefixWildcard{gkutil.PrefixWildcard(includedNamepsace.GetName())},
 					},
 					Location: "metadata.labels.namespace-included",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
 			}
+
 			err = client.Create(context.Background(), namespaceIncluded)
 			Expect(err).NotTo(HaveOccurred())
 
@@ -531,8 +596,8 @@ var _ = Describe("gatekeeper", func() {
 					Name: "cluster-selected",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope: apiextensionsv1beta1.ClusterScoped,
+					Match: gkmatch.Match{
+						Scope: apiextensionsv1.ClusterScoped,
 						LabelSelector: &metav1.LabelSelector{
 							MatchLabels: map[string]string{
 								"selected": "true",
@@ -541,8 +606,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.cluster-selected",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
@@ -555,19 +622,22 @@ var _ = Describe("gatekeeper", func() {
 					Name: "namespace-selected",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope: apiextensionsv1beta1.NamespaceScoped,
+					Match: gkmatch.Match{
+						Scope: apiextensionsv1.NamespaceScoped,
 						NamespaceSelector: &metav1.LabelSelector{
 							MatchLabels: map[string]string{
 								"ns-selected": "true",
 							},
 						},
-						ExcludedNamespaces: []string{excludedNamespace.Name},
+
+						ExcludedNamespaces: []gkutil.PrefixWildcard{gkutil.PrefixWildcard(excludedNamespace.Name)},
 					},
 					Location: "metadata.labels.namespace-selected",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
@@ -575,6 +645,22 @@ var _ = Describe("gatekeeper", func() {
 			err = client.Create(context.Background(), namespaceSelected)
 			Expect(err).NotTo(HaveOccurred())
 
+			for _, am := range []*gkv1alpha.AssignMetadata{allSelected, namespaceIncluded, clusterSelected, namespaceSelected} {
+				Eventually(func() bool {
+
+					getAm := &gkv1alpha.AssignMetadata{}
+					err := client.Get(context.Background(), k8sClient.ObjectKeyFromObject(am), getAm)
+					Expect(err).ToNot(HaveOccurred())
+					podStatuses := getAm.Status.ByPod
+					for _, podStatus := range podStatuses {
+						if !podStatus.Enforced {
+							return false
+						}
+					}
+					return true
+				}, 10*time.Second, 2*time.Second).Should(Equal(true), "Mutations are not applied")
+			}
+
 			By("Creating all test objects")
 			clusterObject := &corev1.Namespace{
 				ObjectMeta: metav1.ObjectMeta{
@@ -604,7 +690,7 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).ToNot(HaveOccurred())
 
 			By("Asserting that cluster object mutations were applied")
-			clusterObjectKey, err := k8sClient.ObjectKeyFromObject(clusterObject)
+			clusterObjectKey := k8sClient.ObjectKeyFromObject(clusterObject)
 			Expect(err).ToNot(HaveOccurred())
 			err = client.Get(context.Background(), clusterObjectKey, clusterObject)
 			Expect(err).ToNot(HaveOccurred())
@@ -626,7 +712,7 @@ var _ = Describe("gatekeeper", func() {
 			Expect(ok).To(Equal(false))
 
 			By("Asserting that excludedNamespaced object mutations were applied")
-			excludedNamespacedObjectKey, err := k8sClient.ObjectKeyFromObject(excludedNamespacedObject)
+			excludedNamespacedObjectKey := k8sClient.ObjectKeyFromObject(excludedNamespacedObject)
 			Expect(err).NotTo(HaveOccurred())
 			err = client.Get(context.Background(), excludedNamespacedObjectKey, excludedNamespacedObject)
 			Expect(err).NotTo(HaveOccurred())
@@ -647,7 +733,7 @@ var _ = Describe("gatekeeper", func() {
 			Expect(ok).To(Equal(false))
 
 			By("Asserting that includedNamespaced object mutations were applied")
-			includedNamespacedObjectKey, err := k8sClient.ObjectKeyFromObject(includedNamespacedObject)
+			includedNamespacedObjectKey := k8sClient.ObjectKeyFromObject(includedNamespacedObject)
 			Expect(err).NotTo(HaveOccurred())
 			err = client.Get(context.Background(), includedNamespacedObjectKey, includedNamespacedObject)
 			Expect(err).NotTo(HaveOccurred())
@@ -670,6 +756,57 @@ var _ = Describe("gatekeeper", func() {
 		})
 	})
 
+	Context("constraints", func() {
+		AfterEach(func() {
+			err := deletePods(constraintValidationNamespace, client)
+			Expect(err).NotTo(HaveOccurred())
+			err = deleteConstraintTemplates(client)
+			Expect(err).NotTo(HaveOccurred())
+		})
+
+		It("should apply constraints", func() {
+			By("adding a constraint template")
+			constraintTemplate := constraintTemplateDef()
+			err := client.Create(context.Background(), constraintTemplate)
+			Expect(err).NotTo(HaveOccurred())
+
+			Eventually(func() bool {
+				crdKey := types.NamespacedName{Name: "denyall.constraints.gatekeeper.sh"}
+				crd := &apiextensionsv1.CustomResourceDefinition{}
+				err := client.Get(context.Background(), crdKey, crd)
+				if err != nil {
+					return false
+				}
+				return crd.Name == "denyall.constraints.gatekeeper.sh"
+			}, 1*time.Minute, 2*time.Second).Should(Equal(true))
+
+			By("adding a constraint based on the template")
+			constraint := constraintDefinition("denyall")
+			err = client.Create(context.Background(), constraint)
+			Expect(err).NotTo(HaveOccurred())
+
+			By("creating a test namespace labeled for constraint validation")
+			ns := &corev1.Namespace{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: constraintValidationNamespace,
+					Labels: map[string]string{
+						"admission.gatekeeper.sh/denyall": "true",
+					},
+				},
+			}
+			err = client.Create(context.Background(), ns)
+			Expect(err).NotTo(HaveOccurred())
+
+			By("failing to add a pod due to validation error")
+			Eventually(func() bool { // Need to wait until constrain is ready in Gatekeeper
+				pod := pods.DefinePod(constraintValidationNamespace)
+				err = client.Create(context.Background(), pod)
+				return err != nil && strings.Contains(err.Error(), "denied!")
+			}, 20*time.Second, 2*time.Second).Should(Equal(true),
+				"Pod creation should be denied with a predefined error message")
+		})
+	})
+
 	Context("operator", func() {
 		It("should be able to select mutation namespaces", func() {
 			var err error
@@ -679,14 +816,21 @@ var _ = Describe("gatekeeper", func() {
 				ObjectMeta: metav1.ObjectMeta{
 					Name: "gatekeeper",
 				},
+				Spec: gkopv1alpha.GatekeeperSpec{
+					Webhook: &gkopv1alpha.WebhookConfig{},
+				},
 			}
 
-			gkConfigKey, err := k8sClient.ObjectKeyFromObject(gkConfig)
+			gkConfigKey := k8sClient.ObjectKeyFromObject(gkConfig)
 			Expect(err).ToNot(HaveOccurred())
 
 			err = client.Get(context.Background(), gkConfigKey, gkConfig)
 			Expect(err).ToNot(HaveOccurred())
 
+			if gkConfig.Spec.Webhook == nil {
+				gkConfig.Spec.Webhook = &gkopv1alpha.WebhookConfig{}
+			}
+
 			gkConfig.Spec.Webhook.NamespaceSelector = &metav1.LabelSelector{
 				MatchLabels: map[string]string{"mutate": "enabled"},
 			}
@@ -694,15 +838,32 @@ var _ = Describe("gatekeeper", func() {
 			err = client.Update(context.Background(), gkConfig, &k8sClient.UpdateOptions{})
 			Expect(err).ToNot(HaveOccurred())
 
+			mutatinWebhookConfiguration := &admission.MutatingWebhookConfiguration{
+				ObjectMeta: metav1.ObjectMeta{
+					Name: "gatekeeper-mutating-webhook-configuration",
+				},
+			}
+			mwConfigKey := k8sClient.ObjectKeyFromObject(mutatinWebhookConfiguration)
+			Expect(err).ToNot(HaveOccurred())
+
+			Eventually(func() bool {
+				err := client.Get(context.Background(), mwConfigKey, mutatinWebhookConfiguration)
+				Expect(err).ToNot(HaveOccurred())
+				// Webhook must update its namespaceSelector to match gatekeeper namespaceSelector
+				return len(mutatinWebhookConfiguration.Webhooks) == 1 &&
+					mutatinWebhookConfiguration.Webhooks[0].NamespaceSelector != nil &&
+					reflect.DeepEqual(mutatinWebhookConfiguration.Webhooks[0].NamespaceSelector.MatchLabels, gkConfig.Spec.Webhook.NamespaceSelector.MatchLabels)
+			}, 1*time.Minute, 1*time.Second).Should(BeTrue())
+
 			By("Creating an all pod mutation")
 			podMutation := &gkv1alpha.AssignMetadata{
 				ObjectMeta: metav1.ObjectMeta{
 					Name: "all-pod-mutation",
 				},
 				Spec: gkv1alpha.AssignMetadataSpec{
-					Match: gkv1alpha.Match{
-						Scope: apiextensionsv1beta1.NamespaceScoped,
-						Kinds: []gkv1alpha.Kinds{
+					Match: gkmatch.Match{
+						Scope: apiextensionsv1.NamespaceScoped,
+						Kinds: []gkmatch.Kinds{
 							{
 								APIGroups: []string{""},
 								Kinds:     []string{"Pod"},
@@ -711,8 +872,10 @@ var _ = Describe("gatekeeper", func() {
 					},
 					Location: "metadata.labels.mutated",
 					Parameters: gkv1alpha.MetadataParameters{
-						Assign: runtime.RawExtension{
-							Raw: []byte(fmt.Sprintf("{\"value\":\"true\"}")),
+						Assign: gkv1alpha.AssignField{
+							Value: &gktypes.Anything{
+								Value: "true",
+							},
 						},
 					},
 				},
@@ -723,7 +886,7 @@ var _ = Describe("gatekeeper", func() {
 			By("Creating the test namespaces")
 			mutationEnabledNamespace := &corev1.Namespace{
 				ObjectMeta: metav1.ObjectMeta{
-					Name: "test-mutate",
+					Name: mutationEnabledNamespace,
 					Labels: map[string]string{
 						"mutate": "enabled",
 					},
@@ -734,7 +897,7 @@ var _ = Describe("gatekeeper", func() {
 
 			mutationDisabledNamespace := &corev1.Namespace{
 				ObjectMeta: metav1.ObjectMeta{
-					Name: "test-no-mutate",
+					Name: mutationDisabledNamespace,
 				},
 			}
 			err = client.Create(context.Background(), mutationDisabledNamespace)
@@ -760,11 +923,6 @@ var _ = Describe("gatekeeper", func() {
 			Expect(err).ToNot(HaveOccurred())
 			_, ok = mutationDisabledPod.Labels["mutated"]
 			Expect(ok).To(BeFalse())
-
-			gkConfig.Spec.Webhook.NamespaceSelector.MatchLabels = nil
-
-			err = client.Update(context.Background(), gkConfig, &k8sClient.UpdateOptions{})
-			Expect(err).ToNot(HaveOccurred())
 		})
 	})
 })
@@ -788,3 +946,79 @@ func deletePods(namespace string, client *testClient.ClientSet) error {
 
 	return nil
 }
+
+func deleteConstraintTemplates(client *testClient.ClientSet) error {
+	list := &constraints.ConstraintTemplateList{}
+
+	err := client.List(context.Background(), list, &k8sClient.ListOptions{})
+	if err != nil {
+		return err
+	}
+
+	for _, item := range list.Items {
+		err = client.Delete(context.Background(), &item)
+		if err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+func constraintDefinition(name string) *unstructured.Unstructured {
+	return &unstructured.Unstructured{
+		Object: map[string]interface{}{
+			"apiVersion": "constraints.gatekeeper.sh/v1beta1",
+			"kind":       "DenyAll",
+			"metadata": map[string]string{
+				"name": name,
+			},
+			"spec": map[string]interface{}{
+				"match": map[string]interface{}{
+					"kinds": []map[string]interface{}{
+						{
+							"apiGroups": []string{""},
+							"kinds":     []string{"Pod"},
+						},
+					},
+					"namespaceSelector": map[string]interface{}{
+						"matchExpressions": []map[string]interface{}{
+							{
+								"key":      "admission.gatekeeper.sh/denyall",
+								"operator": "In",
+								"values":   []string{"true"},
+							},
+						},
+					},
+				},
+			},
+		},
+	}
+}
+
+func constraintTemplateDef() *constraints.ConstraintTemplate {
+
+	return &constraints.ConstraintTemplate{
+		ObjectMeta: metav1.ObjectMeta{Name: "denyall"},
+		Spec: constraints.ConstraintTemplateSpec{
+			CRD: constraints.CRD{
+				Spec: constraints.CRDSpec{
+					Names: constraints.Names{
+						Kind: "DenyAll",
+					},
+				},
+			},
+			Targets: []constraints.Target{
+				{
+					Target: "admission.k8s.gatekeeper.sh",
+					Rego: `
+package foo
+
+violation[{"msg": "denied!"}] {
+	1 == 1
+}
+`,
+				},
+			},
+		},
+	}
+}
