@@ -70,6 +70,8 @@ const (
 	SriovNetworks = "sriovnetworks.sriovnetwork.openshift.io"
 	// SriovOperatorConfigs contains the name of the sriov Operator config CRD
 	SriovOperatorConfigs = "sriovoperatorconfigs.sriovnetwork.openshift.io"
+	// SriovSupportedNicsCM contains the name of the sriov supported nics ConfigMap
+	SriovSupportedNicsCM = "supported-nic-ids"
 )
 
 const (
@@ -126,7 +128,7 @@ const (
 	// GatekeeperControllerDeploymentName contains the name of the gatekeeper-controller-manager deployment
 	GatekeeperControllerDeploymentName = "gatekeeper-controller-manager"
 	// GatekeeperOperatorDeploymentName contains the name of the gatekeeper-operator-controller-manager deployment
-	GatekeeperOperatorDeploymentName = "gatekeeper-operator-controller-manager"
+	GatekeeperOperatorDeploymentName = "gatekeeper-operator-controller"
 	// GatekeeperTestingNamespace is the namespace for resources in this test
 	GatekeeperTestingNamespace = "gatekeeper-testing"
 	// GatekeeperMutationIncludedNamespace is a test namespace that includes mutation
@@ -139,4 +141,23 @@ const (
 	GatekeeperMutationDisabledNamespace = "mutation-disabled"
 	// GatekeeperTestObjectNamespace is a test namespace used as a mutated runtime object
 	GatekeeperTestObjectNamespace = "gk-test-object"
+	// GatekeeperConstraintValidationNamespace is a test namespace used to test constraints
+	GatekeeperConstraintValidationNamespace = "gk-constraint-validation"
+)
+
+const (
+	// NfdNodeFeatureDiscoveryCRDName node feature discovery crd name
+	NfdNodeFeatureDiscoveryCRDName = "nodefeaturediscoveries.nfd.openshift.io"
+	// NfdNamespace node feature discovery operator namespace
+	NfdNamespace = "openshift-nfd"
+	// NfdOperatorDeploymentName node feature discovery operator deployment name
+	NfdOperatorDeploymentName = "nfd-controller-manager"
+	// NfdMasterNodeDaemonsetName node feature discovery daemonset name for master nodes
+	NfdMasterNodeDaemonsetName = "nfd-master"
+	// NfdWorkerNodeDaemonsetName node feature discovery daemonset name for worker nodes
+	NfdWorkerNodeDaemonsetName = "nfd-worker"
+	// SroSpecialResourceCRDName special resource operator crd name
+	SroSpecialResourceCRDName = "specialresources.sro.openshift.io"
+	// SroOperatorDeploymentName special resource operator deployment name
+	SroOperatorDeploymentName = "special-resource-controller-manager"
 )
