@@ -3,22 +3,25 @@ package utils
 import (
 	"os"
 
+	gkopv1alpha "github.com/gatekeeper/gatekeeper-operator/api/v1alpha1"
+	sriovv1 "github.com/k8snetworkplumbingwg/sriov-network-operator/api/v1"
 	sriovNamespaces "github.com/k8snetworkplumbingwg/sriov-network-operator/test/util/namespaces"
+	gkv1alpha "github.com/open-policy-agent/gatekeeper/apis/mutations/v1alpha1"
+	performancev2 "github.com/openshift-kni/performance-addon-operators/api/v2"
 	perfUtils "github.com/openshift-kni/performance-addon-operators/functests/utils"
+	srov1beta1 "github.com/openshift-psap/special-resource-operator/api/v1beta1"
+	ocpbuildv1 "github.com/openshift/api/build/v1"
+	ocpv1 "github.com/openshift/api/config/v1"
+	nfdv1 "github.com/openshift/cluster-nfd-operator/api/v1"
+	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
+	ptpv1 "github.com/openshift/ptp-operator/api/v1"
 	ptpUtils "github.com/openshift/ptp-operator/test/utils"
 	"k8s.io/apimachinery/pkg/runtime"
 
+	n3000v1 "github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/apis/N3000/api/v1"
+	sriovfecv2 "github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/apis/sriov-fec/api/v2"
 	"github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/k8sreporter"
 	"github.com/openshift-kni/cnf-features-deploy/cnf-tests/testsuites/pkg/namespaces"
-
-	gkopv1alpha "github.com/gatekeeper/gatekeeper-operator/api/v1alpha1"
-	sriovv1 "github.com/k8snetworkplumbingwg/sriov-network-operator/api/v1"
-	gkv1alpha "github.com/open-policy-agent/gatekeeper/apis/mutations/v1alpha1"
-	performancev2 "github.com/openshift-kni/performance-addon-operators/api/v2"
-	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
-	ptpv1 "github.com/openshift/ptp-operator/pkg/apis/ptp/v1"
-	n3000v1 "github.com/smart-edge-open/openshift-operator/N3000/api/v1"
-	sriovfecv2 "github.com/smart-edge-open/openshift-operator/sriov-fec/api/v2"
 )
 
 // NewReporter creates a specific reporter for CNF tests
@@ -30,29 +33,38 @@ func NewReporter(reportPath string) (*k8sreporter.KubernetesReporter, error) {
 		sriovv1.AddToScheme(s)
 		gkv1alpha.AddToScheme(s)
 		gkopv1alpha.AddToScheme(s)
+		nfdv1.AddToScheme(s)
+		srov1beta1.AddToScheme(s)
+		ocpv1.Install(s)
+		ocpbuildv1.Install(s)
 	}
 
 	namespacesToDump := map[string]string{
-		namespaces.PerformanceOperator:      "performance",
-		namespaces.PTPOperator:              "ptp",
-		namespaces.SRIOVOperator:            "sriov",
-		NamespaceTesting:                    "other",
-		perfUtils.NamespaceTesting:          "performance",
-		namespaces.DpdkTest:                 "dpdk",
-		sriovNamespaces.Test:                "sriov",
-		ptpUtils.NamespaceTesting:           "ptp",
-		namespaces.SCTPTest:                 "sctp",
-		namespaces.XTU32Test:                "xt_u32",
-		namespaces.IntelOperator:            "intel",
-		namespaces.OVSQOSTest:               "ovs_qos",
-		GatekeeperNamespace:                 "gatekeeper",
-		OperatorNamespace:                   "gatekeeper",
-		GatekeeperTestingNamespace:          "gatekeeper",
-		GatekeeperMutationIncludedNamespace: "gatekeeper",
-		GatekeeperMutationExcludedNamespace: "gatekeeper",
-		GatekeeperMutationEnabledNamespace:  "gatekeeper",
-		GatekeeperMutationDisabledNamespace: "gatekeeper",
-		GatekeeperTestObjectNamespace:       "gatekeeper",
+		namespaces.PerformanceOperator:          "performance",
+		namespaces.PTPOperator:                  "ptp",
+		namespaces.SRIOVOperator:                "sriov",
+		NamespaceTesting:                        "other",
+		perfUtils.NamespaceTesting:              "performance",
+		namespaces.DpdkTest:                     "dpdk",
+		sriovNamespaces.Test:                    "sriov",
+		ptpUtils.NamespaceTesting:               "ptp",
+		namespaces.SCTPTest:                     "sctp",
+		namespaces.Default:                      "sctp",
+		namespaces.XTU32Test:                    "xt_u32",
+		namespaces.IntelOperator:                "intel",
+		namespaces.OVSQOSTest:                   "ovs_qos",
+		GatekeeperNamespace:                     "gatekeeper",
+		OperatorNamespace:                       "gatekeeper",
+		GatekeeperTestingNamespace:              "gatekeeper",
+		GatekeeperMutationIncludedNamespace:     "gatekeeper",
+		GatekeeperMutationExcludedNamespace:     "gatekeeper",
+		GatekeeperMutationEnabledNamespace:      "gatekeeper",
+		GatekeeperMutationDisabledNamespace:     "gatekeeper",
+		GatekeeperTestObjectNamespace:           "gatekeeper",
+		GatekeeperConstraintValidationNamespace: "gatekeeper",
+		NfdNamespace:                            "sro",
+		namespaces.SpecialResourceOperator:      "sro",
+		namespaces.SroTestNamespace:             "sro",
 	}
 
 	crds := []k8sreporter.CRData{
@@ -73,6 +85,10 @@ func NewReporter(reportPath string) (*k8sreporter.KubernetesReporter, error) {
 		{Cr: &gkv1alpha.AssignList{}},
 		{Cr: &gkv1alpha.AssignMetadataList{}},
 		{Cr: &gkopv1alpha.GatekeeperList{}},
+		{Cr: &srov1beta1.SpecialResourceList{}},
+		{Cr: &nfdv1.NodeFeatureDiscoveryList{}},
+		{Cr: &ocpbuildv1.BuildConfigList{}, Namespace: &namespaces.SroTestNamespace},
+		{Cr: &ocpbuildv1.BuildList{}, Namespace: &namespaces.SroTestNamespace},
 	}
 
 	skipByNamespace := func(ns string) bool {
