@@ -13,7 +13,6 @@ import (
 
 	"github.com/open-policy-agent/opa/types"
 	"github.com/open-policy-agent/opa/util"
-	"github.com/open-policy-agent/opa/util/test"
 )
 
 func TestCheckInference(t *testing.T) {
@@ -284,10 +283,10 @@ func TestCheckInference(t *testing.T) {
 	}
 
 	for _, tc := range tests {
-		test.Subtest(t, tc.note, func(t *testing.T) {
+		t.Run(tc.note, func(t *testing.T) {
 			body := MustParseBody(tc.query)
 			checker := newTypeChecker()
-			env := checker.checkLanguageBuiltins(nil, BuiltinMap)
+			env := checker.Env(BuiltinMap)
 			env, err := checker.CheckBody(env, body)
 			if len(err) != 0 {
 				t.Fatalf("Unexpected error: %v", err)
@@ -469,7 +468,7 @@ func TestCheckInferenceRules(t *testing.T) {
 	}
 
 	for _, tc := range tests {
-		test.Subtest(t, tc.note, func(t *testing.T) {
+		t.Run(tc.note, func(t *testing.T) {
 			var elems []util.T
 
 			// Convert test rules into rule slice for call.
@@ -529,7 +528,7 @@ func TestCheckErrorSuppression(t *testing.T) {
 
 	query = `_ = [true | count(1)]`
 
-	_, errs = newTypeChecker().CheckBody(newTypeChecker().checkLanguageBuiltins(nil, BuiltinMap), MustParseBody(query))
+	_, errs = newTypeChecker().CheckBody(newTypeChecker().Env(BuiltinMap), MustParseBody(query))
 	if len(errs) != 1 {
 		t.Fatalf("Expected exactly one error but got: %v", errs)
 	}
@@ -558,7 +557,7 @@ func TestCheckBadCardinality(t *testing.T) {
 	for _, test := range tests {
 		body := MustParseBody(test.body)
 		tc := newTypeChecker()
-		env := tc.checkLanguageBuiltins(nil, BuiltinMap)
+		env := tc.Env(BuiltinMap)
 		_, err := tc.CheckBody(env, body)
 		if len(err) != 1 || err[0].Code != TypeErr {
 			t.Fatalf("Expected 1 type error from %v but got: %v", body, err)
@@ -598,7 +597,7 @@ func TestCheckMatchErrors(t *testing.T) {
 		{"any", `x = ["str", 1]; x[_] = null`},
 	}
 	for _, tc := range tests {
-		test.Subtest(t, tc.note, func(t *testing.T) {
+		t.Run(tc.note, func(t *testing.T) {
 			body := MustParseBody(tc.query)
 			checker := newTypeChecker()
 			_, err := checker.CheckBody(nil, body)
@@ -652,7 +651,7 @@ func TestCheckBuiltinErrors(t *testing.T) {
 	})
 
 	for _, tc := range tests {
-		test.Subtest(t, tc.note, func(t *testing.T) {
+		t.Run(tc.note, func(t *testing.T) {
 			body := MustParseBody(tc.query)
 			checker := newTypeChecker()
 			_, err := checker.CheckBody(env, body)
@@ -839,7 +838,7 @@ func TestCheckRefErrInvalid(t *testing.T) {
 	}
 
 	for _, tc := range tests {
-		test.Subtest(t, tc.note, func(t *testing.T) {
+		t.Run(tc.note, func(t *testing.T) {
 
 			_, errs := newTypeChecker().CheckBody(env, MustParseBody(tc.query))
 			if len(errs) != 1 {
@@ -966,7 +965,7 @@ func TestFunctionTypeInferenceUnappliedWithObjectVarKey(t *testing.T) {
 		f(x) = y { y = {x: 1} }
 	`)
 
-	env, err := newTypeChecker().CheckTypes(newTypeChecker().checkLanguageBuiltins(nil, BuiltinMap), []util.T{
+	env, err := newTypeChecker().CheckTypes(newTypeChecker().Env(BuiltinMap), []util.T{
 		module.Rules[0],
 	})
 
@@ -1209,10 +1208,816 @@ func newTestEnv(rs []string) *TypeEnv {
 		}
 	}
 
-	env, err := newTypeChecker().CheckTypes(newTypeChecker().checkLanguageBuiltins(nil, BuiltinMap), elems)
+	env, err := newTypeChecker().CheckTypes(newTypeChecker().Env(BuiltinMap), elems)
 	if len(err) > 0 {
 		panic(err)
 	}
 
 	return env
 }
+
+const inputSchema = `{
+	"$schema": "http://json-schema.org/draft-07/schema",
+	"$id": "http://example.com/example.json",
+	"type": "object",
+	"title": "The root schema",
+	"description": "The root schema comprises the entire JSON document.",
+	"default": {},
+	"examples": [
+		{
+			"user": "alice",
+			"operation": "write"
+		}
+	],
+	"required": [
+		"user",
+		"operation"
+	],
+	"properties": {
+		"user": {
+			"$id": "#/properties/user",
+			"type": "string",
+			"title": "The user schema",
+			"description": "An explanation about the purpose of this instance.",
+			"default": "",
+			"examples": [
+				"alice"
+			]
+		},
+		"operation": {
+			"$id": "#/properties/operation",
+			"type": "string",
+			"title": "The operation schema",
+			"description": "An explanation about the purpose of this instance.",
+			"default": "",
+			"examples": [
+				"write"
+			]
+		}
+	},
+	"additionalProperties": true
+}`
+
+const inputSchema2 = `{
+    "$schema": "http://json-schema.org/draft-07/schema",
+    "$id": "http://example.com/example.json",
+    "type": "object",
+    "title": "The root schema",
+    "description": "The root schema comprises the entire JSON document.",
+    "default": {},
+    "examples": [
+        {
+            "operation": "read"
+        }
+    ],
+    "required": [
+        "operation"
+    ],
+    "properties": {
+        "operation": {
+            "$id": "#/properties/operation",
+            "type": "string",
+            "title": "The operation schema",
+            "description": "An explanation about the purpose of this instance.",
+            "default": "",
+            "examples": [
+                "read"
+            ]
+        }
+    },
+    "additionalProperties": true
+}`
+
+const dataSchema = `{
+    "$schema": "http://json-schema.org/draft-07/schema",
+    "$id": "http://example.com/example.json",
+    "type": "object",
+    "title": "The root schema",
+    "description": "The root schema comprises the entire JSON document.",
+    "default": {},
+    "examples": [
+        {
+            "alice": [
+                "read",
+                "write"
+            ],
+            "bob": [
+                "read"
+            ]
+        }
+    ],
+    "required": [
+        "alice",
+        "bob"
+    ],
+    "properties": {
+        "alice": {
+            "$id": "#/properties/alice",
+            "type": "array",
+            "title": "The alice schema",
+            "description": "An explanation about the purpose of this instance.",
+            "default": [],
+            "examples": [
+                [
+                    "read",
+                    "write"
+                ]
+            ],
+            "additionalItems": false,
+            "items": {
+                "$id": "#/properties/alice/items",
+                "type": "string",
+                "title": "The items schema",
+                "description": "An explanation about the purpose of this instance.",
+                "default": "",
+                "examples": [
+                    [
+                        "read",
+                        "write"
+                    ]
+                ]
+            }
+        },
+        "bob": {
+            "$id": "#/properties/bob",
+            "type": "array",
+            "title": "The bob schema",
+            "description": "An explanation about the purpose of this instance.",
+            "default": [],
+            "examples": [
+                [
+                    "read"
+                ]
+            ],
+            "additionalItems": false,
+            "items": {
+                "$id": "#/properties/bob/items",
+                "type": "string",
+                "title": "The items schema",
+                "description": "An explanation about the purpose of this instance.",
+                "default": "",
+                "examples": [
+                    [
+                        "read"
+                    ]
+                ]
+            }
+        }
+    },
+    "additionalProperties": true
+}`
+
+func TestCheckAnnotationRules(t *testing.T) {
+
+	ischema := util.MustUnmarshalJSON([]byte(inputSchema))
+	ischema2 := util.MustUnmarshalJSON([]byte(inputSchema2))
+	dschema := util.MustUnmarshalJSON([]byte(dataSchema))
+	numberSchema := util.MustUnmarshalJSON([]byte(`{"type": "number"}`))
+	stringSchema := util.MustUnmarshalJSON([]byte(`{"type": "string"}`))
+
+	module1 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+}`
+
+	module2 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input: schema["whocan-input-schema"]
+#   - data.acl: schema["acl-schema"]
+whocan[user] {
+		access = acl[user]
+		access[_] == input.operation
+}`
+
+	module3 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input: schema["whocan-input-schema"]
+#   - data.acl: schema["acl-schema"]
+allow {
+	access = data.acl[input.user]
+	access[_] == input.operation
+}`
+
+	module4 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema.missing
+whocan[user] {
+		access = acl[user]
+		access[_] == input.operation
+}`
+
+	module8 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+#   - input.apple.orange: schema["input"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.banana
+}`
+
+	module9 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+#   - input.apple.orange: schema["input"]
+#   - input.apple.orange.banana: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.orange.banana
+}`
+
+	module10 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input.apple.orange: schema["input"]
+#   - input.apple.orange.banana: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.orange.banana.fruit
+}`
+
+	module11 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input.apple.orange: schema["input"]
+#   - input.apple.orange: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.orange.bob
+		input.apple.orange.user
+}`
+
+	module12 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+}`
+
+	module13 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input.apple["orange"]: schema["input"]
+allow {
+		access = data.acl[input.user]
+		input.apple.orange.fruit
+}`
+
+	module14 := `
+package policy
+
+import data.acl
+import input
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input.request.object: schema["acl-schema"]
+deny[msg] {
+	input.request.kind.kind == "Pod"
+	image := input.request.object.spec.typo.containers[_].image
+	not startswith(image, "hooli.com/")
+}`
+
+	module15 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   -  data.acl: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.orange.banana
+}`
+
+	module16 := `
+package policy
+
+import data.acl
+import input
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.acl: schema["acl-schema"]
+deny[msg] {
+	input.request.kind.kinds == "Pod"
+	image := input.request.object.spec.containers[_].image
+	not startswith(image, "hooli.com/")
+	data.blah
+}`
+
+	module17 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["acl-schema"]
+allow {
+		input.alice
+}
+
+deny[msg] {
+	input.foo
+}`
+
+	module18 := `
+package policy
+
+import data.acl
+import input
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - input.apple.banana: schema["input"]
+deny[msg] {
+	input.apple.banana
+}
+
+deny1[msg] {
+	input.apple.banana.foo
+}`
+
+	module19 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.acl: schema["acl-schema"]
+#   - data.acl.foo: schema["input"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		input.apple.orange.banana
+		data.acl.foo.blah
+}`
+
+	module20 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		data.acl.foo
+}`
+
+	module21 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+}
+
+# METADATA for whocan rule
+# scope: rule
+# schemas:
+#   - input: schema["whocan-input-schema"]
+#   - data.acl: schema["acl-schema"]
+whocan[user] {
+		access = acl[user]
+		access[_] == input.operation
+}`
+
+	module22 := `
+package policy
+
+import data.acl
+import input
+
+default allow = false
+
+# METADATA
+# scope: rule
+# schemas:
+#   - input: schema["input"]
+#   - data.acl: schema["acl-schema"]
+allow {
+		access = data.acl[input.user]
+		access[_] == input.operation
+		data.foo
+		data.acl.foo
+}
+
+# METADATA for whocan rule
+# scope: rule
+# schemas:
+#   - input: schema["whocan-input-schema"]
+#   - data.acl: schema["acl-schema"]
+whocan[user] {
+		access = acl[user]
+		access[_] == input.operation.foo
+
+}`
+
+	schemaSet := NewSchemaSet()
+	schemaSet.Put(MustParseRef("schema.number"), numberSchema)
+	schemaSet.Put(MustParseRef("schema.string"), stringSchema)
+	schemaSet.Put(MustParseRef("schema.input"), ischema)
+	schemaSet.Put(MustParseRef(`schema["whocan-input-schema"]`), ischema2)
+	schemaSet.Put(MustParseRef(`schema["acl-schema"]`), dschema)
+
+	tests := []struct {
+		note   string
+		module string
+		err    string
+	}{
+		{note: "data and input annotations", module: module1},
+		{note: "correct data override", module: module2},
+		{note: "incorrect data override", module: module3, err: "undefined ref: input.user"},
+		{note: "missing schema", module: module4, err: "undefined schema: schema.missing"},
+		{note: "overriding ref with length greater than one and not existing", module: module8, err: "undefined ref: input.apple.banana"},
+		{note: "overriding ref with length greater than one and existing prefix", module: module9},
+		{note: "overriding ref with length greater than one and existing prefix with type error", module: module10, err: "undefined ref: input.apple.orange.banana.fruit"},
+		{note: "overriding ref with length greater than one and existing ref", module: module11, err: "undefined ref: input.apple.orange.user"},
+		{note: "overriding ref of size one", module: module12, err: "undefined ref: input.user"},
+		{note: "overriding annotation written with brackets", module: module13, err: "undefined ref: input.apple.orange.fruit"},
+		{note: "overriding strict", module: module14, err: "undefined ref: input.request.object.spec.typo"},
+		{note: "data annotation but no input schema", module: module15},
+		{note: "data schema annotation does not overly restrict data expression", module: module16},
+		{note: "correct defer annotation on another rule has no effect base case", module: module17},
+		{note: "correct defer annotation on another rule has no effect", module: module18},
+		{note: "overriding ref with data prefix", module: module19, err: "data.acl.foo.blah"},
+		{note: "data annotation type error", module: module20, err: "data.acl.foo"},
+		{note: "more than one rule with metadata", module: module21},
+		{note: "more than one rule with metadata with type error", module: module22, err: "undefined ref"},
+		{note: "document scope", err: "test.rego:8: rego_type_error: match error", module: `package test
+# METADATA
+# scope: document
+# schemas:
+# - input.foo: schema.number
+p { input.foo = 7 }
+
+p { input.foo = [] }`},
+
+		{note: "rule scope overrides document scope", module: `package test
+
+# METADATA
+# scope: document
+# schemas:
+# - input.foo: schema.number
+p { input.foo = 7 }
+
+# METADATA
+# scope: rule
+# schemas:
+# - input.foo: schema.string
+p { input.foo = "str" }`},
+
+		{note: "rule scope merges with document scope", err: "test.rego:15: rego_type_error: match error", module: `package test
+
+# METADATA
+# scope: document
+# schemas:
+# - input.bar: schema.number
+p { input.bar = 7 }
+
+# METADATA
+# scope: rule
+# schemas:
+# - input.foo: schema.string
+p {
+	input.foo = "str"
+	input.bar = "str"
+}`},
+
+		{note: "document scope conflict", err: "test.rego:9: rego_type_error: document annotation redeclared: test.rego:3", module: `package test
+
+# METADATA
+# scope: document
+# schemas:
+# - input.foo: schema.number
+p { input.foo = 7 }
+
+# METADATA
+# scope: document
+# schemas:
+# - input.foo: schema.string
+p { input.foo = "str" }`},
+
+		{note: "subpackages scope", err: "test.rego:7: rego_type_error: match error", module: `# METADATA
+# scope: subpackages
+# schemas:
+# - input: schema.number
+package test
+
+p { input = "str" }`},
+
+		{note: "document scope overrides subpackages scope", module: `# METADATA
+# scope: subpackages
+# schemas:
+# - input: schema.number
+package test
+
+# METADATA
+# scope: document
+# schemas:
+# - input: schema.string
+p { input = "str" }`},
+
+		{note: "document scope overrides subpackages scope and finds error", err: "test.rego:11: rego_type_error: match error", module: `# METADATA
+# scope: subpackages
+# schemas:
+# - input: schema.string
+package test
+
+# METADATA
+# scope: rule
+# schemas:
+# - input: schema.number
+p { input = "str" }`},
+
+		{note: "package scope", err: "test.rego:7: rego_type_error: match error", module: `# METADATA
+# scope: package
+# schemas:
+# - input: schema.string
+package test
+
+p { input = 7 }`},
+
+		{note: "rule scope overrides package scope", module: `# METADATA
+# scope: package
+# schemas:
+# - input: schema.string
+package test
+
+# METADATA
+# scope: rule
+# schemas:
+# - input: schema.number
+p { input = 7 }`},
+
+		{note: "inline definition", err: "test.rego:7: rego_type_error: match error", module: `package test
+
+# METADATA
+# scope: rule
+# schemas:
+# - input: {"type": "string"}
+p { input = 7 }`},
+		{note: "document scope is unordered", err: "test.rego:3: rego_type_error: match error", module: `package test
+
+p { input = 7 }
+
+# METADATA
+# scope: document
+# schemas:
+# - input: schema.string
+p { input = "foo" }`},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.note, func(t *testing.T) {
+			mod, err := ParseModuleWithOpts("test.rego", tc.module, ParserOptions{
+				ProcessAnnotation: true,
+			})
+			if err != nil {
+				t.Fatal(err)
+			}
+
+			var elems []util.T
+			for _, rule := range mod.Rules {
+				elems = append(elems, rule)
+				for next := rule.Else; next != nil; next = next.Else {
+					elems = append(elems, next)
+				}
+			}
+
+			oldTypeEnv := newTypeChecker().WithSchemaSet(schemaSet).Env(BuiltinMap)
+			typeenv, errors := newTypeChecker().WithSchemaSet(schemaSet).CheckTypes(oldTypeEnv, elems)
+			if len(errors) > 0 {
+				for _, e := range errors {
+					if tc.err == "" || !strings.Contains(e.Error(), tc.err) {
+						t.Fatalf("Unexpected check rule error when processing annotations: %v", e)
+					}
+				}
+				return
+			} else if tc.err != "" {
+				t.Fatalf("Expected error %q but got success", tc.err)
+			}
+
+			if oldTypeEnv.tree.children != nil && typeenv.next.tree.children != nil && (typeenv.next.tree.children.Len() != oldTypeEnv.tree.children.Len()) {
+				t.Fatalf("Unexpected type env")
+			}
+
+		})
+	}
+}
+
+func TestCheckAnnotationInference(t *testing.T) {
+
+	tests := []struct {
+		note    string
+		modules map[string]string
+		schemas map[string]string
+		exp     map[string]types.Type
+	}{
+		{
+			note: "rule scope",
+			modules: map[string]string{
+				"test.rego": `
+package test
+
+# METADATA
+# scope: rule
+# schemas:
+# - input: schema.foo
+p = x { input = x }
+
+q = p`,
+			},
+			schemas: map[string]string{
+				"schema.foo": `{"type": "number"}`,
+			},
+			exp: map[string]types.Type{
+				"data.test.p": types.N,
+				"data.test.q": types.N,
+			},
+		},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.note, func(t *testing.T) {
+
+			modules := map[string]*Module{}
+			for k, v := range tc.modules {
+				var err error
+				modules[k], err = ParseModuleWithOpts(k, v, ParserOptions{ProcessAnnotation: true})
+				if err != nil {
+					t.Fatal(err)
+				}
+
+				ss := NewSchemaSet()
+				for k, v := range tc.schemas {
+
+					ref := MustParseRef(k)
+					var schema interface{}
+					err = util.Unmarshal([]byte(v), &schema)
+					if err != nil {
+						t.Fatal(err)
+					}
+
+					ss.Put(ref, schema)
+				}
+
+				compiler := NewCompiler().WithSchemas(ss)
+				compiler.Compile(modules)
+				if compiler.Failed() {
+					t.Fatal("unexpected error:", compiler.Errors)
+				}
+
+				for k, v := range tc.exp {
+					ref := MustParseRef(k)
+					result := compiler.TypeEnv.Get(ref)
+					if types.Compare(result, v) != 0 {
+						t.Errorf("expected %v => %v but got %v", ref, v, result)
+					}
+				}
+			}
+
+		})
+	}
+
+}
