@@ -64,6 +64,8 @@ func TestNumberTerms(t *testing.T) {
 		{"-.1", "-0.1"},
 		{"-0.0001", "-0.0001"},
 		{"1e1000", "1e1000"},
+		{"0e1", "0"},
+		{"-0.1", "-0.1"},
 	}
 
 	for _, tc := range tests {
@@ -122,22 +124,37 @@ func TestScalarTerms(t *testing.T) {
 	assertParseOneTerm(t, "exponent", "6e7", FloatNumberTerm(6e7))
 	assertParseOneTerm(t, "string", "\"a string\"", StringTerm("a string"))
 	assertParseOneTerm(t, "string", "\"a string u6abc7def8abc0def with unicode\"", StringTerm("a string u6abc7def8abc0def with unicode"))
-	assertParseError(t, "hex", "6abc")
-	assertParseError(t, "non-terminated", "\"foo")
-	assertParseError(t, "non-terminated-raw", "`foo")
-	assertParseError(t, "non-string", "'a string'")
-	assertParseError(t, "non-number", "6zxy")
-	assertParseError(t, "non-number2", "6d7")
-	assertParseError(t, "non-number3", "6\"foo\"")
-	assertParseError(t, "non-number4", "6true")
-	assertParseError(t, "non-number5", "6false")
-	assertParseError(t, "non-number6", "6[null, null]")
-	assertParseError(t, "non-number7", "6{\"foo\": \"bar\"}")
-	assertParseError(t, "non-number8", ".0.")
-	assertParseError(t, "non-number9", "{0e}")
-	assertParseError(t, "non-number9", "{0e.}")
-	assertParseError(t, "non-number10", "{0F}")
-	assertParseError(t, "number too big", "{7e3000000000}")
+	assertParseErrorContains(t, "hex", "6abc", "illegal number format")
+	assertParseErrorContains(t, "non-terminated", "\"foo", "non-terminated string")
+	assertParseErrorContains(t, "non-terminated-raw", "`foo", "non-terminated string")
+	assertParseErrorContains(t, "non-string", "'a string'", "illegal token")
+	assertParseErrorContains(t, "non-number", "6zxy", "illegal number format")
+	assertParseErrorContains(t, "non-number2", "6d7", "illegal number format")
+	assertParseErrorContains(t, "non-number3", "6\"foo\"", "expected exactly one statement") // ??
+	assertParseErrorContains(t, "non-number4", "6true", "illegal number format")
+	assertParseErrorContains(t, "non-number5", "6false", "illegal number format")
+	assertParseErrorContains(t, "non-number6", "6[null, null]", "illegal ref (head cannot be number)") // ??
+	assertParseErrorContains(t, "non-number7", "6{\"foo\": \"bar\"}", "expected exactly one statement")
+	assertParseErrorContains(t, "non-number8", ".0.", "expected fraction")
+	assertParseErrorContains(t, "non-number9", "0e", "expected exponent")
+	assertParseErrorContains(t, "non-number10", "0e.", "expected exponent")
+	assertParseErrorContains(t, "non-number11", "0F", "illegal number format")
+	assertParseErrorContains(t, "non-number12", "00", "expected number")
+	assertParseErrorContains(t, "non-number13", "00.1", "expected number")
+	assertParseErrorContains(t, "non-number14", "-00", "expected number")
+	assertParseErrorContains(t, "non-number15", "-00.1", "expected number")
+	assertParseErrorContains(t, "non-number16", "-00.01", "expected number")
+	assertParseErrorContains(t, "non-number17", "00e1", "expected number")
+	assertParseErrorContains(t, "non-number18", "-00e1", "expected number")
+	assertParseErrorContains(t, "parsing float fails", "7e3000000000", "invalid float")
+	assertParseErrorContains(t, "float is +inf", "10245423601e680507880", "number too big")
+	assertParseErrorContains(t, "float is -inf", "-10245423601e680507880", "number too big")
+
+	// f := big.NewFloat(1); f.SetMantExp(f, -1e6); f.String() // => 1.010034059e-301030 (this takes ~9s)
+	assertParseErrorContains(t, "float exp < -1e5", "1.010034059e-301030", "number too big")
+
+	// g := big.NewFloat(1); g.SetMantExp(g, 1e6); g.String() // => 9.900656229e+301029
+	assertParseErrorContains(t, "float exp > 1e5", "9.900656229e+301029", "number too big")
 }
 
 func TestVarTerms(t *testing.T) {
@@ -2653,6 +2670,476 @@ else = {
 	`), curElse.Head.Value.Location)
 }
 
+func TestAnnotations(t *testing.T) {
+
+	dataServers := MustParseRef("data.servers")
+	dataNetworks := MustParseRef("data.networks")
+	dataPorts := MustParseRef("data.ports")
+
+	schemaServers := MustParseRef("schema.servers")
+	schemaNetworks := MustParseRef("schema.networks")
+	schemaPorts := MustParseRef("schema.ports")
+
+	stringSchemaAsMap := map[string]interface{}{
+		"type": "string",
+	}
+	var stringSchema interface{} = stringSchemaAsMap
+
+	tests := []struct {
+		note           string
+		module         string
+		expNumComments int
+		expAnnotations []*Annotations
+		expError       string
+	}{
+		{
+			note: "Single valid annotation",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 4,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+		{
+			note: "Multiple annotations on multiple lines",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 6,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+						{Path: dataNetworks, Schema: schemaNetworks},
+						{Path: dataPorts, Schema: schemaPorts},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+		{
+			note: "Comment in between metadata and rule (valid)",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+
+# This is a comment after the metadata YAML
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 7,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+						{Path: dataNetworks, Schema: schemaNetworks},
+						{Path: dataPorts, Schema: schemaPorts},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+		{
+			note: "Empty comment line in between metadata and rule (valid)",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+#
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 7,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+						{Path: dataNetworks, Schema: schemaNetworks},
+						{Path: dataPorts, Schema: schemaPorts},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+		{
+			note: "Ill-structured (invalid) metadata start",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+# METADATA
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expError: "test.rego:14: rego_parse_error: yaml: line 7: could not find expected ':'",
+		},
+		{
+			note: "Ill-structured (invalid) annotation document path",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data/servers: schema.servers
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 4,
+			expError:       "rego_parse_error: invalid document reference",
+		},
+		{
+			note: "Ill-structured (invalid) annotation schema path",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema/servers
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 4,
+			expError:       "rego_parse_error: invalid schema reference",
+		},
+		{
+			note: "Ill-structured (invalid) annotation",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers= schema
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 5,
+			expError:       "rego_parse_error: yaml: unmarshal errors:\n  line 3: cannot unmarshal !!str",
+		},
+		{
+			note: "Indentation error in yaml",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+# - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 6,
+			expError:       "rego_parse_error: yaml: line 3: did not find expected key",
+		},
+		{
+			note: "Multiple rules with and without metadata",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+public_servers[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}
+
+public_servers_1[server] {
+	server = servers[i]; server.ports[j] = ports[k].id
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+	server.typo  # won't catch this type error since rule has no schema metadata
+}`,
+			expNumComments: 7,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+						{Path: dataNetworks, Schema: schemaNetworks},
+						{Path: dataPorts, Schema: schemaPorts},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+		{
+			note: "Multiple rules with metadata",
+			module: `
+package opa.examples
+
+import data.servers
+import data.networks
+import data.ports
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.servers: schema.servers
+public_servers[server] {
+	server = servers[i]
+}
+
+# METADATA
+# scope: rule
+# schemas:
+#   - data.networks: schema.networks
+#   - data.ports: schema.ports
+public_servers_1[server] {
+	ports[k].networks[l] = networks[m].id;
+	networks[m].public = true
+}`,
+			expNumComments: 9,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: dataServers, Schema: schemaServers},
+					},
+					Scope: annotationScopeRule,
+					node:  MustParseRule(`public_servers[server] { server = servers[i] }`),
+				},
+				{
+					Schemas: []*SchemaAnnotation{
+
+						{Path: dataNetworks, Schema: schemaNetworks},
+						{Path: dataPorts, Schema: schemaPorts},
+					},
+					Scope: annotationScopeRule,
+					node:  MustParseRule(`public_servers_1[server] { ports[k].networks[l] = networks[m].id; networks[m].public = true }`),
+				},
+			},
+		},
+		{
+			note: "Empty annotation error due to whitespace following METADATA hint",
+			module: `package test
+
+# METADATA
+
+# scope: rule
+p { input.x > 7 }`,
+			expError: "test.rego:3: rego_parse_error: expected METADATA block, found whitespace",
+		},
+		{
+			note: "Annotation on constant",
+			module: `
+package test
+
+# METADATA
+# scope: rule
+p := 7`,
+			expNumComments: 2,
+			expAnnotations: []*Annotations{
+				{Scope: annotationScopeRule},
+			},
+		},
+		{
+			note: "Default rule scope",
+			module: `
+package test
+
+# METADATA
+# {}
+p := 7`,
+			expNumComments: 2,
+			expAnnotations: []*Annotations{
+				{Scope: annotationScopeRule},
+			},
+		},
+		{
+			note: "Unknown scope",
+			module: `
+package test
+
+# METADATA
+# scope: deadbeef
+p := 7`,
+			expNumComments: 2,
+			expError:       "invalid annotation scope 'deadbeef'",
+		},
+		{
+			note: "Invalid rule scope/attachment",
+			module: `
+# METADATA
+# scope: rule
+package test
+
+p := 7`,
+			expNumComments: 2,
+			expError:       "test.rego:2: rego_parse_error: annotation scope 'rule' must be applied to rule (have package)",
+		},
+		{
+			note: "Scope attachment error: document on import",
+			module: `package test
+# METADATA
+# scope: document
+import data.foo.bar`,
+			expError: "test.rego:2: rego_parse_error: annotation scope 'document' must be applied to rule (have import)",
+		},
+		{
+			note: "Scope attachment error: unattached",
+			module: `package test
+
+# METADATA
+# scope: package`,
+			expError: "test.rego:3: rego_parse_error: annotation scope 'package' must be applied to package",
+		},
+		{
+			note: "Scope attachment error: package on non-package",
+			module: `package test
+# METADATA
+# scope: package
+import data.foo`,
+			expError: "test.rego:2: rego_parse_error: annotation scope 'package' must be applied to package (have import)",
+		},
+		{
+			note: "Inline schema definition",
+			module: `package test
+
+# METADATA
+# schemas:
+# - input: {"type": "string"}
+p { input = "str" }`,
+			expNumComments: 3,
+			expAnnotations: []*Annotations{
+				{
+					Schemas: []*SchemaAnnotation{
+						{Path: InputRootRef, Definition: &stringSchema},
+					},
+					Scope: annotationScopeRule,
+				},
+			},
+		},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.note, func(t *testing.T) {
+			mod, err := ParseModuleWithOpts("test.rego", tc.module, ParserOptions{
+				ProcessAnnotation: true,
+			})
+			if err != nil {
+				if tc.expError == "" || !strings.Contains(err.Error(), tc.expError) {
+					t.Fatalf("Unexpected parse error when getting annotations: %v", err)
+				}
+				return
+			} else if tc.expError != "" {
+				t.Fatalf("Expected err: %v but no error from parse module", tc.expError)
+			}
+
+			if len(mod.Comments) != tc.expNumComments {
+				t.Fatalf("Expected %v comments but got %v", tc.expNumComments, len(mod.Comments))
+			}
+
+			if annotationsCompare(tc.expAnnotations, mod.Annotations) != 0 {
+				t.Fatalf("expected %v but got %v", tc.expAnnotations, mod.Annotations)
+			}
+		})
+	}
+}
+
 func assertLocationText(t *testing.T, expected string, actual *Location) {
 	t.Helper()
 	if actual == nil || actual.Text == nil {
