@@ -2,6 +2,7 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// nolint: goconst // string duplication is for test readability.
 package bundle
 
 import (
@@ -21,7 +22,7 @@ import (
 	"github.com/open-policy-agent/opa/internal/file/archive"
 )
 
-func TestManifestAddroot(t *testing.T) {
+func TestManifestAddRoot(t *testing.T) {
 	var m Manifest
 	m.AddRoot("x/y")
 	m.AddRoot("y/z")
@@ -36,14 +37,62 @@ func TestReadWithBaseDir(t *testing.T) {
 	testReadBundle(t, "/foo/bar")
 }
 
+func TestReadWithSizeLimit(t *testing.T) {
+
+	buf := archive.MustWriteTarGz([][2]string{
+		{"data.json", `"foo"`},
+	})
+
+	loader := NewTarballLoaderWithBaseURL(buf, "")
+	br := NewCustomReader(loader).WithSizeLimitBytes(4)
+
+	_, err := br.Read()
+	if err == nil || err.Error() != "bundle file exceeded max size (4 bytes)" {
+		t.Fatal("expected error but got:", err)
+	}
+
+	buf = archive.MustWriteTarGz([][2]string{
+		{".signatures.json", `"foo"`},
+	})
+
+	loader = NewTarballLoaderWithBaseURL(buf, "")
+	br = NewCustomReader(loader).WithSizeLimitBytes(4)
+
+	_, err = br.Read()
+	if err == nil || err.Error() != "bundle signatures file exceeded max size (4 bytes)" {
+		t.Fatal("expected error but got:", err)
+	}
+
+}
+
 func testReadBundle(t *testing.T, baseDir string) {
+	module := `package example`
+
+	modulePath := "/example/example.rego"
+	if baseDir != "" {
+		modulePath = filepath.Join(baseDir, modulePath)
+	}
+
+	legacyWasmModulePath := "/policy.wasm"
+	if baseDir != "" {
+		legacyWasmModulePath = filepath.Join(baseDir, legacyWasmModulePath)
+	}
+
+	wasmResolverPath := "/authz/allow/policy.wasm"
+	fullWasmResolverPath := wasmResolverPath
+	if baseDir != "" {
+		fullWasmResolverPath = filepath.Join(baseDir, wasmResolverPath)
+	}
+
 	files := [][2]string{
 		{"/a/b/c/data.json", "[1,2,3]"},
 		{"/a/b/d/data.json", "true"},
 		{"/a/b/y/data.yaml", `foo: 1`},
 		{"/example/example.rego", `package example`},
-		{"/policy.wasm", `modules-compiled-as-wasm-binary`},
+		{"/policy.wasm", `legacy-wasm-module`},
+		{wasmResolverPath, `wasm-module`},
 		{"/data.json", `{"x": {"y": true}, "a": {"b": {"z": true}}}}`},
+		{"/.manifest", fmt.Sprintf(`{"wasm":[{"entrypoint": "authz/allow", "module": "%s"}]}`, fullWasmResolverPath)},
 	}
 
 	buf := archive.MustWriteTarGz(files)
@@ -55,14 +104,17 @@ func testReadBundle(t *testing.T, baseDir string) {
 		t.Fatal(err)
 	}
 
-	module := `package example`
-
-	modulePath := "/example/example.rego"
-	if baseDir != "" {
-		modulePath = filepath.Join(baseDir, modulePath)
+	expManifest := Manifest{}
+	expManifest.Init()
+	expManifest.WasmResolvers = []WasmResolver{
+		{
+			Entrypoint: "authz/allow",
+			Module:     fullWasmResolverPath,
+		},
 	}
 
 	exp := Bundle{
+		Manifest: expManifest,
 		Data: map[string]interface{}{
 			"a": map[string]interface{}{
 				"b": map[string]interface{}{
@@ -86,11 +138,23 @@ func testReadBundle(t *testing.T, baseDir string) {
 				Raw:    []byte(module),
 			},
 		},
-		Wasm: []byte("modules-compiled-as-wasm-binary"),
+		WasmModules: []WasmModuleFile{
+			{
+				URL:  legacyWasmModulePath,
+				Path: legacyWasmModulePath,
+				Raw:  []byte(`legacy-wasm-module`),
+			},
+			{
+				URL:         fullWasmResolverPath,
+				Path:        fullWasmResolverPath,
+				Raw:         []byte("wasm-module"),
+				Entrypoints: []ast.Ref{ast.MustParseRef("data.authz.allow")},
+			},
+		},
 	}
 
 	if !exp.Equal(bundle) {
-		t.Fatal("\nExp:", exp, "\n\nGot:", bundle)
+		t.Fatalf("\nExp: %+v\nGot: %+v", exp, bundle)
 	}
 }
 
@@ -108,6 +172,33 @@ func TestReadWithManifest(t *testing.T) {
 	}
 }
 
+func TestManifestMetadata(t *testing.T) {
+	files := [][2]string{
+		{"/.manifest", `{
+			"metadata": {
+				"foo": {
+					"version": "1.0.0"
+				} 
+			}
+		}`},
+	}
+	buf := archive.MustWriteTarGz(files)
+	bundle, err := NewReader(buf).Read()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if bundle.Manifest.Metadata["foo"] == nil {
+		t.Fatal("Unexpected nil metadata key")
+	}
+	data, ok := bundle.Manifest.Metadata["foo"].(map[string]interface{})
+	if !ok {
+		t.Fatal("Unexpected structure in metadata")
+	}
+	if data["version"] != "1.0.0" {
+		t.Fatalf("Unexpected metadata value: %v", data["version"])
+	}
+}
+
 func TestReadWithManifestInData(t *testing.T) {
 	files := [][2]string{
 		{"/.manifest", `{"revision": "quickbrownfaux"}`},
@@ -153,6 +244,14 @@ func TestReadWithSignatures(t *testing.T) {
 
 	signedTokenHS256 := `eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6ImZvbyJ9.eyJmaWxlcyI6W3sibmFtZSI6Ii5tYW5pZmVzdCIsImhhc2giOiI1MDdhMmMzOGExNDQxZGI1OGQyY2I4Nzk4MmM0MmFhOTFhNDM0MmVmNDIyYTZiNTQyZWRkZWJlZWY2ZjA0MTJmIiwiYWxnb3JpdGhtIjoiU0hBLTI1NiJ9LHsibmFtZSI6ImEvYi9jL2RhdGEuanNvbiIsImhhc2giOiI0MmNmZTY3NjhiNTdiYjVmNzUwM2MxNjVjMjhkZDA3YWM1YjgxMzU1NGViYzg1MGYyY2MzNTg0M2U3MTM3YjFkIiwiYWxnb3JpdGhtIjoiU0hBLTI1NiJ9LHsibmFtZSI6Imh0dHAvcG9saWN5L3BvbGljeS5yZWdvIiwiaGFzaCI6ImE2MTVlZWFlZTIxZGU1MTc5ZGUwODBkZThjMzA1MmM4ZGE5MDExMzg0MDZiYTcxYzM4YzAzMjg0NWY3ZDU0ZjQiLCJhbGdvcml0aG0iOiJTSEEtMjU2In1dLCJpYXQiOjE1OTIyNDgwMjcsImlzcyI6IkpXVFNlcnZpY2UiLCJrZXlpZCI6ImZvbyIsInNjb3BlIjoid3JpdGUifQ.grzWHYvyVS6LfWy0oiFTEJThKooOAwic8sexYaflzOM`
 	otherSignedTokenHS256 := `eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6ImZvbyJ9.eyJmaWxlcyI6W3sibmFtZSI6ImEvYi9jL2RhdGEuanNvbiIsImhhc2giOiJmOWNhYzA3MTQ3MDVkMjBkMWEyMDg4MDE4NWNkZWQ2ZTBmNmQwNDA2NjJkMmViYjA5NjFkM2Q5ZjMxN2Q4YWNiIn1dLCJpYXQiOjE1OTIyNDgwMjcsImlzcyI6IkpXVFNlcnZpY2UiLCJzY29wZSI6IndyaXRlIn0.WJhnUjwaVvckSgOd4QcVvKThN6oc99NiPiwHKYnoG7c`
+	defaultSigner, _ := GetSigner(defaultSignerID)
+	defaultVerifier, _ := GetVerifier(defaultVerifierID)
+	if err := RegisterSigner("_bar", defaultSigner); err != nil {
+		t.Fatal(err)
+	}
+	if err := RegisterVerifier("_bar", defaultVerifier); err != nil {
+		t.Fatal(err)
+	}
 
 	tests := map[string]struct {
 		files   [][2]string
@@ -165,9 +264,14 @@ func TestReadWithSignatures(t *testing.T) {
 			nil,
 			true, fmt.Errorf("verification key not provided"),
 		},
-		"no_signatures_file": {
+		"no_signatures_file_no_keyid": {
 			[][2]string{{"/.manifest", `{"revision": "quickbrownfaux"}`}},
 			NewVerificationConfig(map[string]*KeyConfig{}, "", "", nil),
+			false, nil,
+		},
+		"no_signatures_file": {
+			[][2]string{{"/.manifest", `{"revision": "quickbrownfaux"}`}},
+			NewVerificationConfig(map[string]*KeyConfig{}, "somekey", "", nil),
 			true, fmt.Errorf("bundle missing .signatures.json file"),
 		},
 		"no_signatures": {
@@ -202,6 +306,16 @@ func TestReadWithSignatures(t *testing.T) {
 			NewVerificationConfig(map[string]*KeyConfig{"foo": {Key: "secret", Algorithm: "HS256"}}, "", "write", []string{".*", "a/b/c/data.json", "http/policy/policy.rego"}),
 			false, nil,
 		},
+		"customer_signer_verifier": {
+			[][2]string{
+				{"/.signatures.json", fmt.Sprintf(`{"signatures": ["%v"],"plugin":"_bar"}`, signedTokenHS256)},
+				{"/.manifest", `{"revision": "quickbrownfaux"}`},
+				{"/a/b/c/data.json", "[1,2,3]"},
+				{"/http/policy/policy.rego", `package example`},
+			},
+			NewVerificationConfig(map[string]*KeyConfig{"foo": {Key: "secret", Algorithm: "HS256"}}, "", "write", []string{".*", "a/b/c/data.json", "http/policy/policy.rego"}),
+			false, nil,
+		},
 	}
 
 	for name, tc := range tests {
@@ -401,7 +515,7 @@ func TestReadRootValidation(t *testing.T) {
 		err   string
 	}{
 		{
-			note: "default: full extent",
+			note: "default full extent",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd"}`},
 				{"/data.json", `{"a": 1}`},
@@ -410,7 +524,7 @@ func TestReadRootValidation(t *testing.T) {
 			err: "",
 		},
 		{
-			note: "explicit: full extent",
+			note: "explicit full extent",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": [""]}`},
 				{"/data.json", `{"a": 1}`},
@@ -419,7 +533,7 @@ func TestReadRootValidation(t *testing.T) {
 			err: "",
 		},
 		{
-			note: "implicit: prefixed",
+			note: "implicit prefixed",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": ["a/b", "foo"]}`},
 				{"/data.json", `{"a": {"b": 1}}`},
@@ -428,7 +542,7 @@ func TestReadRootValidation(t *testing.T) {
 			err: "",
 		},
 		{
-			note: "err: empty",
+			note: "err empty",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": []}`},
 				{"/x.rego", `package foo`},
@@ -436,21 +550,21 @@ func TestReadRootValidation(t *testing.T) {
 			err: "manifest roots [] do not permit 'package foo' in module '/x.rego'",
 		},
 		{
-			note: "err: overlapped",
+			note: "err overlapped",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": ["a/b", "a"]}`},
 			},
 			err: "manifest has overlapped roots: 'a/b' and 'a'",
 		},
 		{
-			note: "edge: overlapped partial segment",
+			note: "edge overlapped partial segment",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": ["a", "another_root"]}`},
 			},
 			err: "",
 		},
 		{
-			note: "err: package outside scope",
+			note: "err package outside scope",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": ["a", "b", "c/d"]}`},
 				{"/a.rego", `package b.c`},
@@ -459,7 +573,7 @@ func TestReadRootValidation(t *testing.T) {
 			err: "manifest roots [a b c/d] do not permit 'package c.e' in module '/x.rego'",
 		},
 		{
-			note: "err: data outside scope",
+			note: "err data outside scope",
 			files: [][2]string{
 				{"/.manifest", `{"revision": "abcd", "roots": ["a", "b", "c/d"]}`},
 				{"/data.json", `{"a": 1}`},
@@ -575,8 +689,8 @@ func TestReadErrorBadGzip(t *testing.T) {
 func TestReadErrorBadTar(t *testing.T) {
 	var buf bytes.Buffer
 	gw := gzip.NewWriter(&buf)
-	gw.Write([]byte("bad tar bytes"))
-	gw.Close()
+	_, _ = gw.Write([]byte("bad tar bytes"))
+	_ = gw.Close()
 	_, err := NewReader(&buf).Read()
 	if err == nil {
 		t.Fatal("expected error")
@@ -628,7 +742,13 @@ func TestRoundtripDeprecatedWrite(t *testing.T) {
 				Raw:    []byte(`package foo.corge`),
 			},
 		},
-		Wasm: []byte("modules-compiled-as-wasm-binary"),
+		WasmModules: []WasmModuleFile{
+			{
+				Path: "/policy.wasm",
+				URL:  "/policy.wasm",
+				Raw:  []byte("modules-compiled-as-wasm-binary"),
+			},
+		},
 		Manifest: Manifest{
 			Revision: "quickbrownfaux",
 		},
@@ -646,7 +766,7 @@ func TestRoundtripDeprecatedWrite(t *testing.T) {
 	}
 
 	if !bundle2.Equal(bundle) {
-		t.Fatal("Exp:", bundle, "\n\nGot:", bundle2)
+		t.Fatalf("\nExp: %+v\nGot: %+v", bundle, bundle2)
 	}
 
 }
@@ -669,7 +789,13 @@ func TestRoundtrip(t *testing.T) {
 				Raw:    []byte("package foo.corge\n"),
 			},
 		},
-		Wasm: []byte("modules-compiled-as-wasm-binary"),
+		WasmModules: []WasmModuleFile{
+			{
+				Path: "/policy.wasm",
+				URL:  "/policy.wasm",
+				Raw:  []byte("modules-compiled-as-wasm-binary"),
+			},
+		},
 		Manifest: Manifest{
 			Revision: "quickbrownfaux",
 		},
@@ -780,6 +906,62 @@ func TestGenerateSignature(t *testing.T) {
 	}
 }
 
+func TestGenerateSignatureWithPlugin(t *testing.T) {
+	signatures := SignaturesConfig{Signatures: []string{"some_token"}, Plugin: "_foo"}
+
+	bundle := Bundle{
+		Data: map[string]interface{}{
+			"foo": map[string]interface{}{
+				"bar": []interface{}{json.Number("1"), json.Number("2"), json.Number("3")},
+				"baz": true,
+				"qux": "hello",
+			},
+		},
+		Modules: []ModuleFile{
+			{
+				URL:    "/foo/corge/corge.rego",
+				Path:   "/foo/corge/corge.rego",
+				Parsed: ast.MustParseModule(`package foo.corge`),
+				Raw:    []byte("package foo.corge\n"),
+			},
+		},
+		Wasm: []byte("modules-compiled-as-wasm-binary"),
+		Manifest: Manifest{
+			Revision: "quickbrownfaux",
+		},
+		Signatures: signatures,
+	}
+
+	defaultSigner, _ := GetSigner(defaultSignerID)
+	defaultVerifier, _ := GetVerifier(defaultVerifierID)
+	if err := RegisterSigner("_foo", defaultSigner); err != nil {
+		t.Fatal(err)
+	}
+	if err := RegisterVerifier("_foo", defaultVerifier); err != nil {
+		t.Fatal(err)
+	}
+	sc := NewSigningConfig("secret", "HS256", "").WithPlugin("_foo")
+
+	err := bundle.GenerateSignature(sc, "", false)
+	if err != nil {
+		t.Fatal("Unexpected error:", err)
+	}
+
+	if reflect.DeepEqual(signatures, bundle.Signatures) {
+		t.Fatal("Expected signatures to be different")
+	}
+
+	current := bundle.Signatures
+	err = bundle.GenerateSignature(sc, "", false)
+	if err != nil {
+		t.Fatal("Unexpected error:", err)
+	}
+
+	if !reflect.DeepEqual(current, bundle.Signatures) {
+		t.Fatal("Expected signatures to be same")
+	}
+}
+
 func TestFormatModulesRaw(t *testing.T) {
 
 	bundle1 := Bundle{
@@ -882,7 +1064,7 @@ func TestHashBundleFiles(t *testing.T) {
 	for name, tc := range tests {
 		t.Run(name, func(t *testing.T) {
 
-			f, err := hashBundleFiles(h, tc.data, tc.manifest, tc.wasm)
+			f, err := hashBundleFiles(h, &Bundle{Data: tc.data, Manifest: tc.manifest, Wasm: tc.wasm})
 			if err != nil {
 				t.Fatal("Unexpected error:", err)
 			}
@@ -1028,8 +1210,8 @@ func TestParsedModules(t *testing.T) {
 
 func TestMergeCorruptManifest(t *testing.T) {
 	_, err := Merge([]*Bundle{
-		&Bundle{},
-		&Bundle{},
+		{},
+		{},
 	})
 	if err == nil || err.Error() != "bundle manifest not initialized" {
 		t.Fatal("unexpected error:", err)
@@ -1051,7 +1233,7 @@ func TestMerge(t *testing.T) {
 		{
 			note: "no op",
 			bundles: []*Bundle{
-				&Bundle{
+				{
 					Manifest: Manifest{
 						Revision: "abcdef",
 					},
@@ -1079,16 +1261,90 @@ func TestMerge(t *testing.T) {
 			},
 		},
 		{
-			note: "wasm merge error",
+			note: "wasm merge legacy error",
 			bundles: []*Bundle{
 				{
+					Manifest: Manifest{
+						Roots: &[]string{
+							"foo",
+						},
+					},
 					Wasm: []byte("not really wasm, but good enough"),
 				},
 				{
+					Manifest: Manifest{
+						Roots: &[]string{
+							"bar",
+						},
+					},
 					Wasm: []byte("not really wasm, but good enough"),
 				},
 			},
-			wantErr: errors.New("wasm bundles cannot be merged"),
+			wantBundle: &Bundle{
+				Manifest: Manifest{
+					Roots: &[]string{
+						"foo",
+						"bar",
+					},
+				},
+			},
+		},
+		{
+			note: "wasm merge ok",
+			bundles: []*Bundle{
+				{
+					Manifest: Manifest{
+						Roots: &[]string{
+							"logs",
+						},
+					},
+					WasmModules: []WasmModuleFile{
+						{
+							URL:         "logs/mask/policy.wasm",
+							Path:        "logs/mask/policy.wasm",
+							Entrypoints: []ast.Ref{ast.MustParseRef("system.log.mask")},
+							Raw:         []byte("not really wasm, but good enough"),
+						},
+					},
+				},
+				{
+					Manifest: Manifest{
+						Roots: &[]string{
+							"authz",
+						},
+					},
+					WasmModules: []WasmModuleFile{
+						{
+							URL:         "authz/allow/policy.wasm",
+							Path:        "authz/allow/policy.wasm",
+							Entrypoints: []ast.Ref{ast.MustParseRef("authz.allow")},
+							Raw:         []byte("not really wasm, but good enough"),
+						},
+					},
+				},
+			},
+			wantBundle: &Bundle{
+				Manifest: Manifest{
+					Roots: &[]string{
+						"logs",
+						"authz",
+					},
+				},
+				WasmModules: []WasmModuleFile{
+					{
+						URL:         "logs/mask/policy.wasm",
+						Path:        "logs/mask/policy.wasm",
+						Entrypoints: []ast.Ref{ast.MustParseRef("system.log.mask")},
+						Raw:         []byte("not really wasm, but good enough"),
+					},
+					{
+						URL:         "authz/allow/policy.wasm",
+						Path:        "authz/allow/policy.wasm",
+						Entrypoints: []ast.Ref{ast.MustParseRef("authz.allow")},
+						Raw:         []byte("not really wasm, but good enough"),
+					},
+				},
+			},
 		},
 		{
 			note: "merge policy",
