@@ -6,19 +6,25 @@
 package compile
 
 import (
+	"bytes"
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"regexp"
 	"sort"
+	"strings"
 
 	"github.com/pkg/errors"
 
 	"github.com/open-policy-agent/opa/ast"
 	"github.com/open-policy-agent/opa/bundle"
+	"github.com/open-policy-agent/opa/internal/compiler/wasm"
+	"github.com/open-policy-agent/opa/internal/debug"
+	"github.com/open-policy-agent/opa/internal/ir"
+	"github.com/open-policy-agent/opa/internal/planner"
 	"github.com/open-policy-agent/opa/internal/ref"
 	initload "github.com/open-policy-agent/opa/internal/runtime/init"
+	"github.com/open-policy-agent/opa/internal/wasm/encoding"
 	"github.com/open-policy-agent/opa/loader"
 	"github.com/open-policy-agent/opa/rego"
 	"github.com/open-policy-agent/opa/storage"
@@ -39,8 +45,8 @@ const (
 const wasmResultVar = ast.Var("result")
 
 var validTargets = map[string]struct{}{
-	TargetRego: struct{}{},
-	TargetWasm: struct{}{},
+	TargetRego: {},
+	TargetWasm: {},
 }
 
 // Compiler implements bundle compilation and linking.
@@ -54,51 +60,25 @@ type Compiler struct {
 	entrypoints       orderedStringSet           // policy entrypoints required for optimization and certain targets
 	optimizationLevel int                        // how aggressive should optimization be
 	target            string                     // target type (wasm, rego, etc.)
-	output            io.Writer                  // output stream to write bundle to
+	output            *io.Writer                 // output stream to write bundle to
 	entrypointrefs    []*ast.Term                // validated entrypoints computed from default decision or manually supplied entrypoints
 	compiler          *ast.Compiler              // rego ast compiler used for semantic checks and rewriting
-	debug             *debugEvents               // debug information produced during build
+	debug             debug.Debug                // optionally outputs debug information produced during build
 	bvc               *bundle.VerificationConfig // represents the key configuration used to verify a signed bundle
 	bsc               *bundle.SigningConfig      // represents the key configuration used to generate a signed bundle
 	keyID             string                     // represents the name of the default key used to verify a signed bundle
 }
 
-type debugEvents struct {
-	debug []Debug
-}
-
-func (d *debugEvents) Add(info Debug) {
-	if d != nil {
-		d.debug = append(d.debug, info)
-	}
-}
-
-// Debug contains debugging information produced by the build about optimizations and other operations.
-type Debug struct {
-	Location *ast.Location
-	Message  string
-}
-
-func (d Debug) String() string {
-	return fmt.Sprintf("%v: %v", d.Location, d.Message)
-}
-
 // New returns a new compiler instance that can be invoked.
 func New() *Compiler {
 	return &Compiler{
 		asBundle:          false,
 		optimizationLevel: 0,
 		target:            TargetRego,
-		output:            ioutil.Discard,
-		debug:             &debugEvents{},
+		debug:             debug.Discard(),
 	}
 }
 
-// Debug returns a list of debug events produced by the compiler.
-func (c *Compiler) Debug() []Debug {
-	return c.debug.debug
-}
-
 // WithRevision sets the revision to include in the output bundle manifest.
 func (c *Compiler) WithRevision(r string) *Compiler {
 	c.revision = &r
@@ -112,7 +92,7 @@ func (c *Compiler) WithAsBundle(enabled bool) *Compiler {
 }
 
 // WithEntrypoints sets the policy entrypoints on the compiler. Entrypoints tell the
-// compiler what rules to expect and where optimizations can be targetted. The wasm
+// compiler what rules to expect and where optimizations can be targeted. The wasm
 // target requires at least one entrypoint as does optimization.
 func (c *Compiler) WithEntrypoints(e ...string) *Compiler {
 	c.entrypoints = c.entrypoints.Append(e...)
@@ -135,7 +115,15 @@ func (c *Compiler) WithTarget(t string) *Compiler {
 
 // WithOutput sets the output stream to write the bundle to.
 func (c *Compiler) WithOutput(w io.Writer) *Compiler {
-	c.output = w
+	c.output = &w
+	return c
+}
+
+// WithDebug sets the output stream to write debug info to.
+func (c *Compiler) WithDebug(sink io.Writer) *Compiler {
+	if sink != nil {
+		c.debug = debug.New(sink)
+	}
 	return c
 }
 
@@ -151,6 +139,14 @@ func (c *Compiler) WithFilter(filter loader.Filter) *Compiler {
 	return c
 }
 
+// WithBundle sets the input bundle to compile. This should be used as an
+// alternative to reading from paths. This function overrides any file
+// loading options.
+func (c *Compiler) WithBundle(b *bundle.Bundle) *Compiler {
+	c.bundle = b
+	return c
+}
+
 // WithBundleVerificationConfig sets the key configuration to use to verify a signed bundle
 func (c *Compiler) WithBundleVerificationConfig(config *bundle.VerificationConfig) *Compiler {
 	c.bvc = config
@@ -211,7 +207,11 @@ func (c *Compiler) Build(ctx context.Context) error {
 		}
 	}
 
-	return bundle.NewWriter(c.output).Write(*c.bundle)
+	if c.output == nil {
+		return nil
+	}
+
+	return bundle.NewWriter(*c.output).Write(*c.bundle)
 }
 
 func (c *Compiler) init() error {
@@ -231,10 +231,6 @@ func (c *Compiler) init() error {
 			return fmt.Errorf("entrypoint %v not valid: use <package>/<rule>", e)
 		}
 
-		if len(r) <= 2 {
-			return fmt.Errorf("entrypoint %v too short: use <package>/<rule>", e)
-		}
-
 		c.entrypointrefs = append(c.entrypointrefs, ast.NewTerm(r))
 	}
 
@@ -242,15 +238,26 @@ func (c *Compiler) init() error {
 		return errors.New("bundle optimizations require at least one entrypoint")
 	}
 
-	if c.target == TargetWasm && len(c.entrypointrefs) != 1 {
-		return errors.New("wasm compilation requires exactly one entrypoint")
+	if c.target == TargetWasm && len(c.entrypointrefs) == 0 {
+		return errors.New("wasm compilation requires at least one entrypoint")
 	}
 
 	return nil
 }
 
+// Bundle returns the compiled bundle. This function can be called to retrieve the
+// output of the compiler (as an alternative to having the bundle written to a stream.)
+func (c *Compiler) Bundle() *bundle.Bundle {
+	return c.bundle
+}
+
 func (c *Compiler) initBundle() error {
 
+	// If the bundle is already set, skip file loading.
+	if c.bundle != nil {
+		return nil
+	}
+
 	// TODO(tsandall): the metrics object should passed through here so we that
 	// we can track read and parse times.
 
@@ -316,13 +323,13 @@ func (c *Compiler) optimize(ctx context.Context) error {
 
 	if c.optimizationLevel <= 0 {
 		var err error
-		c.compiler, err = compile(c.capabilities, c.bundle)
+		c.compiler, err = compile(c.capabilities, c.bundle, c.debug)
 		return err
 	}
 
 	o := newOptimizer(c.capabilities, c.bundle).
 		WithEntrypoints(c.entrypointrefs).
-		WithDebug(c.debug).
+		WithDebug(c.debug.Writer()).
 		WithShallowInlining(c.optimizationLevel <= 1)
 
 	err := o.Do(ctx)
@@ -341,26 +348,197 @@ func (c *Compiler) compileWasm(ctx context.Context) error {
 	// AST compiler will not be set because the default target does not require it.
 	if c.compiler == nil {
 		var err error
-		c.compiler, err = compile(c.capabilities, c.bundle)
+		c.compiler, err = compile(c.capabilities, c.bundle, c.debug)
 		if err != nil {
 			return err
 		}
 	}
 
-	store := inmem.NewFromObject(c.bundle.Data)
+	// Find transitive dependents of entrypoints and add them to the set to compile.
+	//
+	// NOTE(tsandall): We compile entrypoints because the evaluator does not support
+	// evaluation of wasm-compiled rules when 'with' statements are in-scope. Compiling
+	// out the dependents avoids the need to support that case for now.
+	deps := map[*ast.Rule]struct{}{}
+	for i := range c.entrypointrefs {
+		transitiveDocumentDependents(c.compiler, c.entrypointrefs[i], deps)
+	}
+
+	extras := ast.NewSet()
+	for rule := range deps {
+		extras.Add(ast.NewTerm(rule.Path()))
+	}
+
+	sorted := extras.Sorted()
+
+	for i := 0; i < sorted.Len(); i++ {
+		p, err := sorted.Elem(i).Value.(ast.Ref).Ptr()
+		if err != nil {
+			return err
+		}
+
+		if !c.entrypoints.Contains(p) {
+			c.entrypoints = append(c.entrypoints, p)
+			c.entrypointrefs = append(c.entrypointrefs, sorted.Elem(i))
+		}
+	}
+
+	// Create query sets for each of the entrypoints.
 	resultSym := ast.NewTerm(wasmResultVar)
+	queries := make([]planner.QuerySet, len(c.entrypointrefs))
 
-	cr, err := rego.New(
-		rego.ParsedQuery(ast.NewBody(ast.Equality.Expr(resultSym, c.entrypointrefs[0]))),
-		rego.Compiler(c.compiler),
-		rego.Store(store),
-	).Compile(ctx)
+	for i := range c.entrypointrefs {
 
+		qc := c.compiler.QueryCompiler()
+		query := ast.NewBody(ast.Equality.Expr(resultSym, c.entrypointrefs[i]))
+		compiled, err := qc.Compile(query)
+		if err != nil {
+			return err
+		}
+
+		queries[i] = planner.QuerySet{
+			Name:          c.entrypoints[i],
+			Queries:       []ast.Body{compiled},
+			RewrittenVars: qc.RewrittenVars(),
+		}
+	}
+
+	// Prepare modules and builtins for the planner.
+	modules := []*ast.Module{}
+	for _, module := range c.compiler.Modules {
+		modules = append(modules, module)
+	}
+
+	builtins := make(map[string]*ast.Builtin, len(c.capabilities.Builtins))
+	for _, bi := range c.capabilities.Builtins {
+		builtins[bi.Name] = bi
+	}
+
+	compiler := wasm.New()
+	found := false
+	have := compiler.ABIVersion()
+	if c.capabilities.WasmABIVersions == nil { // discern nil from len=0
+		c.debug.Printf("no wasm ABI versions in capabilities, building for %v", have)
+		found = true
+	}
+	for _, v := range c.capabilities.WasmABIVersions {
+		if v.Version == have.Version && v.Minor <= have.Minor {
+			found = true
+			break
+		}
+	}
+	if !found {
+		return fmt.Errorf("compiler ABI version not in capabilities (have %v, want %d)",
+			c.capabilities.WasmABIVersions,
+			compiler.ABIVersion(),
+		)
+	}
+
+	// Plan the query sets.
+	p := planner.New().
+		WithQueries(queries).
+		WithModules(modules).
+		WithBuiltinDecls(builtins).
+		WithDebug(c.debug.Writer())
+	policy, err := p.Plan()
 	if err != nil {
 		return err
 	}
 
-	c.bundle.Wasm = cr.Bytes
+	// dump policy IR (if "debug" wasn't requested, debug.Witer will discard it)
+	err = ir.Pretty(c.debug.Writer(), policy)
+	if err != nil {
+		return err
+	}
+
+	// Compile the policy into a wasm binary.
+	m, err := compiler.WithPolicy(policy).WithDebug(c.debug.Writer()).Compile()
+	if err != nil {
+		return err
+	}
+
+	var buf bytes.Buffer
+	if err := encoding.WriteModule(&buf, m); err != nil {
+		return err
+	}
+
+	modulePath := bundle.WasmFile
+
+	c.bundle.WasmModules = []bundle.WasmModuleFile{{
+		URL:  modulePath,
+		Path: modulePath,
+		Raw:  buf.Bytes(),
+	}}
+
+	// Each entrypoint needs an entry in the manifest
+	for i := range c.entrypointrefs {
+		entrypointPath := c.entrypoints[i]
+
+		c.bundle.Manifest.WasmResolvers = append(c.bundle.Manifest.WasmResolvers, bundle.WasmResolver{
+			Module:     "/" + strings.TrimLeft(modulePath, "/"),
+			Entrypoint: entrypointPath,
+		})
+	}
+
+	// Remove the entrypoints from remaining source rego files
+	return pruneBundleEntrypoints(c.bundle, c.entrypointrefs)
+}
+
+// pruneBundleEntrypoints will modify modules in the provided bundle to remove
+// rules matching the entrypoints along with injecting import statements to
+// preserve their ability to compile.
+func pruneBundleEntrypoints(b *bundle.Bundle, entrypointrefs []*ast.Term) error {
+
+	// For each package path keep a list of new imports to add.
+	requiredImports := map[string][]*ast.Import{}
+
+	for _, entrypoint := range entrypointrefs {
+		for i := 0; i < len(b.Modules); i++ {
+			mf := &b.Modules[i]
+
+			// Drop any rules that match the entrypoint path.
+			var rules []*ast.Rule
+			for _, rule := range mf.Parsed.Rules {
+				rulePath := rule.Path()
+				if !rulePath.Equal(entrypoint.Value) {
+					rules = append(rules, rule)
+				} else {
+					pkgPath := rule.Module.Package.Path.String()
+					newImport := &ast.Import{Path: ast.NewTerm(rulePath)}
+					shouldAdd := true
+					currentImports := requiredImports[pkgPath]
+					for _, imp := range currentImports {
+						if imp.Equal(newImport) {
+							shouldAdd = false
+							break
+						}
+					}
+					if shouldAdd {
+						requiredImports[pkgPath] = append(currentImports, newImport)
+					}
+				}
+			}
+
+			// If any rules were dropped update the module accordingly
+			if len(rules) != len(mf.Parsed.Rules) {
+				mf.Parsed.Rules = rules
+				// Remove the original raw source, we're editing the AST
+				// directly so it wont be in sync anymore.
+				mf.Raw = nil
+			}
+		}
+	}
+
+	// Any packages which had rules removed need an import injected for the
+	// removed rule to keep the policies valid.
+	for i := 0; i < len(b.Modules); i++ {
+		mf := &b.Modules[i]
+		pkgPath := mf.Parsed.Package.Path.String()
+		if imports, ok := requiredImports[pkgPath]; ok {
+			mf.Raw = nil
+			mf.Parsed.Imports = append(mf.Parsed.Imports, imports...)
+		}
+	}
 
 	return nil
 }
@@ -382,7 +560,7 @@ type optimizer struct {
 	resultsymprefix string
 	outputprefix    string
 	shallow         bool
-	debug           *debugEvents
+	debug           debug.Debug
 }
 
 func newOptimizer(c *ast.Capabilities, b *bundle.Bundle) *optimizer {
@@ -392,11 +570,14 @@ func newOptimizer(c *ast.Capabilities, b *bundle.Bundle) *optimizer {
 		nsprefix:        "partial",
 		resultsymprefix: ast.WildcardPrefix,
 		outputprefix:    "optimized",
+		debug:           debug.Discard(),
 	}
 }
 
-func (o *optimizer) WithDebug(debug *debugEvents) *optimizer {
-	o.debug = debug
+func (o *optimizer) WithDebug(sink io.Writer) *optimizer {
+	if sink != nil {
+		o.debug = debug.New(sink)
+	}
 	return o
 }
 
@@ -439,7 +620,7 @@ func (o *optimizer) Do(ctx context.Context) error {
 	for i, e := range o.entrypoints {
 
 		var err error
-		o.compiler, err = compile(o.capabilities, o.bundle)
+		o.compiler, err = compile(o.capabilities, o.bundle, o.debug)
 		if err != nil {
 			return err
 		}
@@ -448,10 +629,12 @@ func (o *optimizer) Do(ctx context.Context) error {
 			unknowns = o.findUnknowns()
 		}
 
+		required := o.findRequiredDocuments(e)
+
 		r := rego.New(
 			rego.ParsedQuery(ast.NewBody(ast.Equality.Expr(resultsym, e))),
 			rego.PartialNamespace(o.nsprefix),
-			rego.DisableInlining(o.findRequiredDocuments(e)),
+			rego.DisableInlining(required),
 			rego.ShallowInlining(o.shallow),
 			rego.SkipPartialNamespace(true),
 			rego.ParsedUnknowns(unknowns),
@@ -459,6 +642,15 @@ func (o *optimizer) Do(ctx context.Context) error {
 			rego.Store(store),
 		)
 
+		o.debug.Printf("optimizer: entrypoint: %v", e)
+		o.debug.Printf("  partial-namespace: %v", o.nsprefix)
+		o.debug.Printf("  disable-inlining: %v", required)
+		o.debug.Printf("  shallow-inlining: %v", o.shallow)
+
+		for i := range unknowns {
+			o.debug.Printf("  unknown: %v", unknowns[i])
+		}
+
 		pq, err := r.Partial(ctx)
 		if err != nil {
 			return err
@@ -510,9 +702,7 @@ func (o *optimizer) findRequiredDocuments(ref *ast.Term) []string {
 	keep := map[string]*ast.Location{}
 	deps := map[*ast.Rule]struct{}{}
 
-	for _, r := range o.compiler.GetRules(ref.Value.(ast.Ref)) {
-		transitiveDependents(o.compiler, r, deps)
-	}
+	transitiveDocumentDependents(o.compiler, ref, deps)
 
 	for rule := range deps {
 		ast.WalkExprs(rule, func(expr *ast.Expr) bool {
@@ -537,10 +727,7 @@ func (o *optimizer) findRequiredDocuments(ref *ast.Term) []string {
 	sort.Strings(result)
 
 	for _, k := range result {
-		o.debug.Add(Debug{
-			Location: keep[k],
-			Message:  fmt.Sprintf("disables inlining of %v", k),
-		})
+		o.debug.Printf("%s: disables inlining of %v", keep[k], k)
 	}
 
 	return result
@@ -562,10 +749,6 @@ func (o *optimizer) findUnknowns() []*ast.Term {
 				return true
 			}
 			if !refs.ContainsPrefix(prefix) {
-				o.debug.Add(Debug{
-					Location: x[0].Location,
-					Message:  fmt.Sprintf("marking %v as unknown", prefix),
-				})
 				unknowns.AddPrefix(prefix)
 			}
 			return false
@@ -691,15 +874,19 @@ func (o *optimizer) getSupportModuleFilename(used map[string]int, module *ast.Mo
 
 var safePathPattern = regexp.MustCompile(`^[\w-_/]+$`)
 
-func compile(c *ast.Capabilities, b *bundle.Bundle) (*ast.Compiler, error) {
+func compile(c *ast.Capabilities, b *bundle.Bundle, dbg debug.Debug) (*ast.Compiler, error) {
 
 	modules := map[string]*ast.Module{}
 
 	for _, mf := range b.Modules {
+		if _, ok := modules[mf.URL]; ok {
+			return nil, fmt.Errorf("duplicate module URL: %s", mf.URL)
+		}
+
 		modules[mf.URL] = mf.Parsed
 	}
 
-	compiler := ast.NewCompiler().WithCapabilities(c)
+	compiler := ast.NewCompiler().WithCapabilities(c).WithDebug(dbg.Writer())
 	compiler.Compile(modules)
 
 	if compiler.Failed() {
@@ -709,6 +896,12 @@ func compile(c *ast.Capabilities, b *bundle.Bundle) (*ast.Compiler, error) {
 	return compiler, nil
 }
 
+func transitiveDocumentDependents(compiler *ast.Compiler, ref *ast.Term, deps map[*ast.Rule]struct{}) {
+	for _, rule := range compiler.GetRules(ref.Value.(ast.Ref)) {
+		transitiveDependents(compiler, rule, deps)
+	}
+}
+
 func transitiveDependents(compiler *ast.Compiler, rule *ast.Rule, deps map[*ast.Rule]struct{}) {
 	for x := range compiler.Graph.Dependents(rule) {
 		other := x.(*ast.Rule)
@@ -734,6 +927,15 @@ func (ss orderedStringSet) Append(s ...string) orderedStringSet {
 	return ss
 }
 
+func (ss orderedStringSet) Contains(s string) bool {
+	for _, other := range ss {
+		if s == other {
+			return true
+		}
+	}
+	return false
+}
+
 func stringsToRefs(x []string) []ast.Ref {
 	result := make([]ast.Ref, len(x))
 	for i := range result {
