@@ -15,6 +15,7 @@ import (
 	"github.com/open-policy-agent/opa/ast"
 	"github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/format"
+	"github.com/open-policy-agent/opa/internal/ref"
 	"github.com/open-policy-agent/opa/loader"
 	"github.com/open-policy-agent/opa/util"
 	"github.com/open-policy-agent/opa/util/test"
@@ -42,25 +43,15 @@ func TestCompilerInitErrors(t *testing.T) {
 			c:    New().WithTarget("deadbeef"),
 			want: fmt.Errorf("invalid target \"deadbeef\""),
 		},
-		{
-			note: "entrypoint parse error",
-			c:    New().WithEntrypoints("foo%bar"),
-			want: fmt.Errorf("entrypoint foo%%bar not valid: use <package>/<rule>"),
-		},
-		{
-			note: "entrypoint too short error",
-			c:    New().WithEntrypoints("foo"),
-			want: fmt.Errorf("entrypoint foo too short: use <package>/<rule>"),
-		},
 		{
 			note: "optimizations require entrypoint",
 			c:    New().WithOptimizationLevel(1),
 			want: errors.New("bundle optimizations require at least one entrypoint"),
 		},
 		{
-			note: "wasm compilation requires exactly one entrypoint",
+			note: "wasm compilation requires at least one entrypoint",
 			c:    New().WithTarget("wasm"),
-			want: errors.New("wasm compilation requires exactly one entrypoint"),
+			want: errors.New("wasm compilation requires at least one entrypoint"),
 		},
 	}
 
@@ -253,6 +244,60 @@ func TestCompilerLoadHonorsFilter(t *testing.T) {
 	})
 }
 
+func TestCompilerInputBundle(t *testing.T) {
+
+	b := &bundle.Bundle{
+		Modules: []bundle.ModuleFile{
+			{
+				URL:    "/foo.rego",
+				Path:   "/foo.rego",
+				Raw:    []byte("package test\np = 7"),
+				Parsed: ast.MustParseModule("package test\np = 7"),
+			},
+		},
+	}
+
+	compiler := New().WithBundle(b)
+
+	if err := compiler.Build(context.Background()); err != nil {
+		t.Fatal(err)
+	}
+
+	exp := "package test\n\np = 7\n"
+
+	if exp != string(compiler.Bundle().Modules[0].Raw) {
+		t.Fatalf("expected module to have been formatted (output different than expected):\n\ngot: %v\n\nwant: %v", string(compiler.Bundle().Modules[0].Raw), exp)
+	}
+}
+
+func TestCompilerInputInvalidBundle(t *testing.T) {
+
+	b := &bundle.Bundle{
+		Modules: []bundle.ModuleFile{
+			{
+				URL:    "/url",
+				Path:   "/foo.rego",
+				Raw:    []byte("package test\np = 0"),
+				Parsed: ast.MustParseModule("package test\np = 0"),
+			},
+			{
+				URL:    "/url",
+				Path:   "/bar.rego",
+				Raw:    []byte("package test\nq = 1"),
+				Parsed: ast.MustParseModule("package test\nq = 1"),
+			},
+		},
+	}
+
+	compiler := New().WithBundle(b)
+
+	if err := compiler.Build(context.Background()); err == nil {
+		t.Fatal("duplicate module URL not detected")
+	} else if err.Error() != "duplicate module URL: /url" {
+		t.Fatal(err)
+	}
+}
+
 func TestCompilerError(t *testing.T) {
 	files := map[string]string{
 		"test.rego": `
@@ -305,9 +350,9 @@ func TestCompilerOptimizationL1(t *testing.T) {
 		optimizedExp := ast.MustParseModule(`
 			package test
 
-			q { input.x = 1 }
-			p { data.test.q = X; X }
 			default p = false
+			p { data.test.q = X; X }
+			q { input.x = 1 }
 		`)
 
 		// NOTE(tsandall): PE generates vars with wildcard prefix. Instead of
@@ -315,12 +360,15 @@ func TestCompilerOptimizationL1(t *testing.T) {
 		// here. If this becomes a common pattern, we could refactor (e.g.,
 		// allow caller to control var prefix, split into a reusable function,
 		// etc.)
-		ast.TransformVars(optimizedExp, func(x ast.Var) (ast.Value, error) {
+		_, err = ast.TransformVars(optimizedExp, func(x ast.Var) (ast.Value, error) {
 			if x == ast.Var("X") {
 				return ast.Var("$_term_1_01"), nil
 			}
 			return x, nil
 		})
+		if err != nil {
+			t.Fatal(err)
+		}
 
 		if len(compiler.bundle.Modules) != 1 {
 			t.Fatalf("expected 1 module but got: %v", compiler.bundle.Modules)
@@ -364,8 +412,8 @@ func TestCompilerOptimizationL2(t *testing.T) {
 		optimizedExp := ast.MustParseModule(`
 			package test
 
-			p { input.x = 1 }
 			default p = false
+			p { input.x = 1 }
 		`)
 
 		if len(compiler.bundle.Modules) != 2 {
@@ -382,25 +430,189 @@ func TestCompilerOptimizationL2(t *testing.T) {
 	})
 }
 
+// NOTE(sr): we override this to not depend on build tags in tests
+func wasmABIVersions(vs ...ast.WasmABIVersion) *ast.Capabilities {
+	caps := ast.CapabilitiesForThisVersion()
+	caps.WasmABIVersions = vs
+	return caps
+}
+
 func TestCompilerWasmTarget(t *testing.T) {
 	files := map[string]string{
 		"test.rego": `package test
 
-		p = true`,
+		p = 7
+		q = p+1`,
 	}
 
 	test.WithTempFS(files, func(root string) {
 
-		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p")
+		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p", "test/q").
+			WithCapabilities(wasmABIVersions(ast.WasmABIVersion{Version: 1}))
 		err := compiler.Build(context.Background())
 		if err != nil {
 			t.Fatal(err)
 		}
 
-		if len(compiler.bundle.Wasm) == 0 {
+		if len(compiler.bundle.WasmModules) == 0 {
 			t.Fatal("expected to find compiled wasm module")
 		}
 
+		if len(compiler.bundle.Wasm) != 0 {
+			t.Error("expected NOT to find deprecated bundle `Wasm` value")
+		}
+
+		ensureEntrypointRemoved(t, compiler.bundle, "test/p")
+	})
+}
+
+// If we're building a wasm bundle, and the `opa` binary we use to do that
+// does not support wasm _itself_, then it shouldn't bother.
+func TestCompilerWasmTargetWithCapabilitiesUnset(t *testing.T) {
+	files := map[string]string{
+		"test.rego": `package test
+
+		p = 7
+		q = p+1`,
+	}
+
+	test.WithTempFS(files, func(root string) {
+
+		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p", "test/q")
+		err := compiler.Build(context.Background())
+		if err != nil {
+			t.Fatalf("expected no error, got %v", err)
+		}
+	})
+}
+
+func TestCompilerWasmTargetWithCapabilitiesMismatch(t *testing.T) {
+	files := map[string]string{
+		"test.rego": `package test
+
+		p = 7
+		q = p+1`,
+	}
+
+	test.WithTempFS(files, func(root string) {
+
+		for note, wabis := range map[string][]ast.WasmABIVersion{
+			"none":     {},
+			"mismatch": {{Version: 0}, {Version: 1, Minor: 2}},
+		} {
+			t.Run(note, func(t *testing.T) {
+				caps := ast.CapabilitiesForThisVersion()
+				caps.WasmABIVersions = wabis
+				compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p", "test/q").
+					WithCapabilities(caps)
+				err := compiler.Build(context.Background())
+				if err == nil {
+					t.Fatal("expected err, got nil")
+				}
+			})
+		}
+	})
+}
+
+func TestCompilerWasmTargetMultipleEntrypoints(t *testing.T) {
+	files := map[string]string{
+		"test.rego": `package test
+
+		p = true`,
+		"policy.rego": `package policy
+
+		authz = true`,
+		"mask.rego": `package system.log
+
+		mask["/input/password"]`,
+	}
+
+	test.WithTempFS(files, func(root string) {
+
+		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p", "policy/authz").
+			WithCapabilities(wasmABIVersions(ast.WasmABIVersion{Version: 1}))
+		err := compiler.Build(context.Background())
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if len(compiler.bundle.WasmModules) != 1 {
+			t.Fatalf("expected 1 Wasm modules, got: %d", len(compiler.bundle.WasmModules))
+		}
+
+		expManifest := bundle.Manifest{}
+		expManifest.Init()
+		expManifest.WasmResolvers = []bundle.WasmResolver{
+			{
+				Entrypoint: "test/p",
+				Module:     "/policy.wasm",
+			},
+			{
+				Entrypoint: "policy/authz",
+				Module:     "/policy.wasm",
+			},
+		}
+
+		if !compiler.bundle.Manifest.Equal(expManifest) {
+			t.Fatalf("\nExpected manifest: %+v\nGot: %+v\n", expManifest, compiler.bundle.Manifest)
+		}
+
+		ensureEntrypointRemoved(t, compiler.bundle, "test/p")
+		ensureEntrypointRemoved(t, compiler.bundle, "policy/authz")
+	})
+}
+
+func TestCompilerWasmTargetEntrypointDependents(t *testing.T) {
+	files := map[string]string{
+		"test.rego": `package test
+
+		p { q }
+		q { r }
+		r = 1
+		s = 2
+		z { r }`}
+
+	test.WithTempFS(files, func(root string) {
+
+		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/r", "test/z").
+			WithCapabilities(wasmABIVersions(ast.WasmABIVersion{Version: 1}))
+		err := compiler.Build(context.Background())
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if len(compiler.bundle.WasmModules) != 1 {
+			t.Fatalf("expected 1 Wasm modules, got: %d", len(compiler.bundle.WasmModules))
+		}
+
+		expManifest := bundle.Manifest{}
+		expManifest.Init()
+		expManifest.WasmResolvers = []bundle.WasmResolver{
+			{
+				Entrypoint: "test/r",
+				Module:     "/policy.wasm",
+			},
+			{
+				Entrypoint: "test/z",
+				Module:     "/policy.wasm",
+			},
+			{
+				Entrypoint: "test/p",
+				Module:     "/policy.wasm",
+			},
+			{
+				Entrypoint: "test/q",
+				Module:     "/policy.wasm",
+			},
+		}
+
+		if !compiler.bundle.Manifest.Equal(expManifest) {
+			t.Fatalf("\nExpected manifest: %+v\nGot: %+v\n", expManifest, compiler.bundle.Manifest)
+		}
+
+		ensureEntrypointRemoved(t, compiler.bundle, "test/p")
+		ensureEntrypointRemoved(t, compiler.bundle, "test/q")
+		ensureEntrypointRemoved(t, compiler.bundle, "test/r")
 	})
 }
 
@@ -414,22 +626,40 @@ func TestCompilerWasmTargetLazyCompile(t *testing.T) {
 
 	test.WithTempFS(files, func(root string) {
 
-		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p").WithOptimizationLevel(1)
+		compiler := New().WithPaths(root).WithTarget("wasm").WithEntrypoints("test/p").WithOptimizationLevel(1).
+			WithCapabilities(wasmABIVersions(ast.WasmABIVersion{Version: 1}))
 		err := compiler.Build(context.Background())
 		if err != nil {
 			t.Fatal(err)
 		}
 
-		if len(compiler.bundle.Wasm) == 0 {
+		if len(compiler.bundle.WasmModules) == 0 {
 			t.Fatal("expected to find compiled wasm module")
 		}
 
 		if _, exists := compiler.compiler.Modules["optimized/test.rego"]; !exists {
 			t.Fatal("expected to find optimized module on compiler")
 		}
+
+		ensureEntrypointRemoved(t, compiler.bundle, "test/p")
 	})
 }
 
+func ensureEntrypointRemoved(t *testing.T, b *bundle.Bundle, e string) {
+	t.Helper()
+	r, err := ref.ParseDataPath(e)
+	if err != nil {
+		t.Fatalf("unexpected error: %s", err)
+	}
+	for _, mf := range b.Modules {
+		for _, rule := range mf.Parsed.Rules {
+			if rule.Path().Equal(r) {
+				t.Errorf("expected entrypoint to be removed from rego all modules in bundle, found rule: %s in %s", rule.Path(), mf.Path)
+			}
+		}
+	}
+}
+
 func TestCompilerSetRevision(t *testing.T) {
 	files := map[string]string{
 		"test.rego": `package test
@@ -467,6 +697,8 @@ func TestCompilerOutput(t *testing.T) {
 		if err != nil {
 			t.Fatal(err)
 		}
+
+		// Check that the written bundle is expected.
 		result, err := bundle.NewReader(buf).Read()
 		if err != nil {
 			t.Fatal(err)
@@ -478,13 +710,23 @@ func TestCompilerOutput(t *testing.T) {
 		}
 
 		if !exp.Equal(result) {
-			t.Fatalf("expected:\n\n%v\n\ngot:\n\n%v", *exp, result)
+			t.Fatalf("expected:\n\n%+v\n\ngot:\n\n%+v", *exp, result)
 		}
 
 		if !exp.Manifest.Equal(result.Manifest) {
-			t.Fatalf("expected:\n\n%v\n\ngot:\n\n%v", exp.Manifest, result.Manifest)
+			t.Fatalf("expected:\n\n%+v\n\ngot:\n\n%+v", exp.Manifest, result.Manifest)
 		}
 
+		// Check that the returned bundle is the expected.
+		compiled := compiler.Bundle()
+
+		if !exp.Equal(*compiled) {
+			t.Fatalf("expected:\n\n%v\n\ngot:\n\n%v", *exp, *compiled)
+		}
+
+		if !exp.Manifest.Equal(compiled.Manifest) {
+			t.Fatalf("expected:\n\n%v\n\ngot:\n\n%v", exp.Manifest, compiled.Manifest)
+		}
 	})
 }
 
@@ -551,10 +793,10 @@ func TestOptimizerErrors(t *testing.T) {
 			modules: map[string]string{
 				"test.rego": `
 					package test
-					p { div(1, 0, x) }
+					p { {k: v | k = ["a", "a"][_]; v = [0, 1][_] } }
 				`,
 			},
-			wantErr: fmt.Errorf("test.rego:3: eval_builtin_error: div: divide by zero"),
+			wantErr: fmt.Errorf("test.rego:3: eval_conflict_error: object keys must be unique"),
 		},
 	}
 
@@ -633,10 +875,11 @@ func TestOptimizerOutput(t *testing.T) {
 				"optimized/test.rego": `
 					package test
 
+					default p = false
+
 					p = true { 1 = input.x }
 					p = true { 2 = input.x }
 
-					default p = false
 				`,
 				"test.rego": `
 					package test
@@ -840,6 +1083,137 @@ func TestOptimizerOutput(t *testing.T) {
 				`,
 			},
 		},
+		{
+			note:        "generate rules with type violations: complete doc",
+			entrypoints: []string{"data.test.p"},
+			modules: map[string]string{
+				"test.rego": `
+					package test
+
+					p {
+						x := split(input.a, ":")
+						f(x[0])
+					}
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+			roots: []string{"test"},
+			wantModules: map[string]string{
+				"optimized/test.rego": `
+					package test
+
+					p = __result__ { split(input.a, ":", __local3__1); startswith(__local3__1[0], "foo"); __result__ = true }
+				`,
+				"test.rego": `
+					package test
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+		},
+		{
+			note:        "generate rules with type violations: partial set",
+			entrypoints: []string{"data.test.p"},
+			modules: map[string]string{
+				"test.rego": `
+					package test
+
+					p[msg] {
+						x := split(input.a, ":")
+						f(x[0])
+						msg := "test string"
+					}
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+			roots: []string{"test"},
+			wantModules: map[string]string{
+				"optimized/test.rego": `
+					package test
+
+					p["test string"] { split(input.a, ":", __local4__1); startswith(__local4__1[0], "foo") }
+				`,
+				"test.rego": `
+					package test
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+		},
+		{
+			note:        "generate rules with type violations: partial object",
+			entrypoints: []string{"data.test.p"},
+			modules: map[string]string{
+				"test.rego": `
+					package test
+
+					p[k] = value  {
+						x := split(input.a, ":")
+						f(x[0])
+						k := "a"
+						value := 1
+					}
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+			roots: []string{"test"},
+			wantModules: map[string]string{
+				"optimized/test.rego": `
+					package test
+
+					p["a"] = 1 { split(input.a, ":", __local5__1); startswith(__local5__1[0], "foo") }
+				`,
+				"test.rego": `
+					package test
+
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+		},
+		{
+			note:        "generate rules with type violations: negation",
+			entrypoints: []string{"data.test.p"},
+			modules: map[string]string{
+				"test.rego": `
+					package test
+
+					p  { not q }
+					q {
+						x := split(input.a, ":")
+						f(x[0])
+					}
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+			},
+			roots: []string{"test"},
+			wantModules: map[string]string{
+				"optimized/test.rego": `
+					package test
+
+					p = __result__ { not data.partial.__not1_0_2__; __result__ = true }
+				`,
+				"test.rego": `
+					package test
+					q = true { assign(x, split(input.a, ":")); f(x[0]) }
+					f(x) { x == null }
+					f(x) { startswith(x, "foo") }
+				`,
+				"optimized/partial.rego": `
+					package partial
+            		__not1_0_2__ = true { split(input.a, ":", __local3__3); startswith(__local3__3[0], "foo") }
+				`,
+			},
+		},
 	}
 
 	for _, tc := range tests {
