@@ -2,29 +2,27 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
-// Package bundle implements bundle downloading.
+// Package bundle implements bundle loading.
 package bundle
 
 import (
-	"bytes"
 	"context"
 	"errors"
 	"fmt"
-	"io/ioutil"
+	"io"
+	"net/url"
 	"os"
 	"path/filepath"
 	"reflect"
 	"sync"
 	"time"
 
-	"github.com/sirupsen/logrus"
-
-	"github.com/open-policy-agent/opa/metrics"
-
 	"github.com/open-policy-agent/opa/ast"
-
 	"github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/download"
+	bundleUtils "github.com/open-policy-agent/opa/internal/bundle"
+	"github.com/open-policy-agent/opa/logging"
+	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
 	"github.com/open-policy-agent/opa/storage"
 )
@@ -37,12 +35,13 @@ type Plugin struct {
 	etags             map[string]string                        // etag on last successful activation
 	listeners         map[interface{}]func(Status)             // listeners to send status updates to
 	bulkListeners     map[interface{}]func(map[string]*Status) // listeners to send aggregated status updates to
-	downloaders       map[string]*download.Downloader
+	downloaders       map[string]bundleLoader
+	logger            logging.Logger
 	mtx               sync.Mutex
 	cfgMtx            sync.Mutex
-	legacyConfig      bool
 	ready             bool
 	bundlePersistPath string
+	stopped           bool
 }
 
 // New returns a new Plugin with the given config.
@@ -58,9 +57,10 @@ func New(parsedConfig *Config, manager *plugins.Manager) *Plugin {
 		manager:     manager,
 		config:      *parsedConfig,
 		status:      initialStatus,
-		downloaders: make(map[string]*download.Downloader),
+		downloaders: make(map[string]bundleLoader),
 		etags:       make(map[string]string),
 		ready:       false,
+		logger:      manager.Logger(),
 	}
 
 	manager.UpdatePluginStatus(Name, &plugins.Status{State: plugins.StateNotReady})
@@ -87,7 +87,7 @@ func (p *Plugin) Start(ctx context.Context) error {
 
 	var err error
 
-	p.bundlePersistPath, err = getDefaultBundlePersistPath()
+	p.bundlePersistPath, err = p.getBundlePersistPath()
 	if err != nil {
 		return err
 	}
@@ -99,7 +99,7 @@ func (p *Plugin) Start(ctx context.Context) error {
 
 	p.initDownloaders()
 	for name, dl := range p.downloaders {
-		p.logInfo(name, "Starting bundle downloader.")
+		p.log(name).Info("Starting bundle loader.")
 		dl.Start(ctx)
 	}
 	return nil
@@ -108,9 +108,16 @@ func (p *Plugin) Start(ctx context.Context) error {
 // Stop stops the plugin.
 func (p *Plugin) Stop(ctx context.Context) {
 	p.mtx.Lock()
-	defer p.mtx.Unlock()
+	stopDownloaders := map[string]bundleLoader{}
 	for name, dl := range p.downloaders {
-		p.logInfo(name, "Stopping bundle downloader.")
+		stopDownloaders[name] = dl
+	}
+	p.downloaders = nil
+	p.stopped = true
+	p.mtx.Unlock()
+
+	for name, dl := range stopDownloaders {
+		p.log(name).Info("Stopping bundle loader.")
 		dl.Stop(ctx)
 	}
 }
@@ -153,7 +160,7 @@ func (p *Plugin) Reconfigure(ctx context.Context, config interface{}) {
 	// Cleanup existing downloaders that are deleted
 	for name := range p.downloaders {
 		if _, deleted := deletedBundles[name]; deleted {
-			p.logInfo(name, "Bundle downloader configuration removed. Stopping bundle downloader.")
+			p.log(name).Info("Bundle loader configuration removed. Stopping bundle loader.")
 			delete(p.downloaders, name)
 			delete(p.status, name)
 			delete(p.etags, name)
@@ -172,7 +179,7 @@ func (p *Plugin) Reconfigure(ctx context.Context, config interface{}) {
 		}
 		err := bundle.Deactivate(opts)
 		if err != nil {
-			p.logError(fmt.Sprint(deletedBundles), "Failed to deactivate bundles: %s", err)
+			p.manager.Logger().Error(fmt.Sprint(deletedBundles), "Failed to deactivate bundles: %s", err)
 			return err
 		}
 		return nil
@@ -192,9 +199,9 @@ func (p *Plugin) Reconfigure(ctx context.Context, config interface{}) {
 		if isNew || updated {
 			if isNew {
 				p.status[name] = &Status{Name: name}
-				p.logInfo(name, "New bundle downloader configuration added. Starting bundle downloader.")
+				p.log(name).Info("New bundle loader configuration added. Starting bundle loader.")
 			} else {
-				p.logInfo(name, "Bundle downloader configuration changed. Restarting bundle downloader.")
+				p.log(name).Info("Bundle loader configuration changed. Restarting bundle loader.")
 			}
 			p.downloaders[name] = p.newDownloader(name, source)
 			p.downloaders[name].Start(ctx)
@@ -228,7 +235,7 @@ func (p *Plugin) Unregister(name interface{}) {
 	p.mtx.Lock()
 	defer p.mtx.Unlock()
 
-	delete(p.bulkListeners, name)
+	delete(p.listeners, name)
 }
 
 // RegisterBulkListener registers a listener to receive bulk (aggregated) status updates. The name must be comparable.
@@ -264,11 +271,11 @@ func (p *Plugin) initDownloaders() {
 }
 
 func (p *Plugin) loadAndActivateBundlesFromDisk(ctx context.Context) error {
-	for name := range p.config.Bundles {
+	for name, src := range p.config.Bundles {
 		if p.persistBundle(name) {
-			b, err := loadBundleFromDisk(p.bundlePersistPath, name)
+			b, err := loadBundleFromDisk(p.bundlePersistPath, name, src)
 			if err != nil {
-				p.logError(name, "Failed to load bundle from disk: %v", err)
+				p.log(name).Error("Failed to load bundle from disk: %v", err)
 				return err
 			}
 
@@ -280,7 +287,7 @@ func (p *Plugin) loadAndActivateBundlesFromDisk(ctx context.Context) error {
 
 			err = p.activate(ctx, name, b)
 			if err != nil {
-				p.logError(name, "Bundle activation failed: %v", err)
+				p.log(name).Error("Bundle activation failed: %v", err)
 				return err
 			}
 
@@ -289,21 +296,39 @@ func (p *Plugin) loadAndActivateBundlesFromDisk(ctx context.Context) error {
 
 			p.checkPluginReadiness()
 
-			p.logDebug(name, "Bundle loaded from disk and activated successfully.")
+			p.log(name).Debug("Bundle loaded from disk and activated successfully.")
 		}
 	}
 	return nil
 }
 
-func (p *Plugin) newDownloader(name string, source *Source) *download.Downloader {
+func (p *Plugin) newDownloader(name string, source *Source) bundleLoader {
+
+	if u, err := url.Parse(source.Resource); err == nil {
+		switch u.Scheme {
+		case "file":
+			return &fileLoader{
+				name:           name,
+				path:           u.Path,
+				bvc:            source.Signing,
+				sizeLimitBytes: source.SizeLimitBytes,
+				f:              p.oneShot,
+			}
+		}
+	}
+
 	conf := source.Config
 	client := p.manager.Client(source.Service)
 	path := source.Resource
-
-	return download.New(conf, client, path).WithCallback(func(ctx context.Context, u download.Update) {
+	callback := func(ctx context.Context, u download.Update) {
 		// wrap the callback to include the name of the bundle that was updated
 		p.oneShot(ctx, name, u)
-	}).WithBundleVerificationConfig(source.Signing)
+	}
+	return download.New(conf, client, path).
+		WithCallback(callback).
+		WithBundleVerificationConfig(source.Signing).
+		WithSizeLimitBytes(source.SizeLimitBytes).
+		WithBundlePersistence(p.persistBundle(name))
 }
 
 func (p *Plugin) oneShot(ctx context.Context, name string, u download.Update) {
@@ -323,7 +348,8 @@ func (p *Plugin) oneShot(ctx context.Context, name string, u download.Update) {
 		// made to it.
 		statusCpy := map[string]*Status{}
 		for k, v := range p.status {
-			statusCpy[k] = v
+			v := *v
+			statusCpy[k] = &v
 		}
 		listener(statusCpy)
 	}
@@ -340,9 +366,11 @@ func (p *Plugin) process(ctx context.Context, name string, u download.Update) {
 	p.status[name].SetRequest()
 
 	if u.Error != nil {
-		p.logError(name, "Bundle download failed: %v", u.Error)
+		p.log(name).Error("Bundle load failed: %v", u.Error)
 		p.status[name].SetError(u.Error)
-		p.downloaders[name].ClearCache()
+		if !p.stopped {
+			p.downloaders[name].ClearCache()
+		}
 		return
 	}
 
@@ -355,32 +383,36 @@ func (p *Plugin) process(ctx context.Context, name string, u download.Update) {
 		defer p.status[name].Metrics.Timer(metrics.RegoLoadBundles).Stop()
 
 		if err := p.activate(ctx, name, u.Bundle); err != nil {
-			p.logError(name, "Bundle activation failed: %v", err)
+			p.log(name).Error("Bundle activation failed: %v", err)
 			p.status[name].SetError(err)
-			p.downloaders[name].ClearCache()
+			if !p.stopped {
+				p.downloaders[name].ClearCache()
+			}
 			return
 		}
 
 		if p.persistBundle(name) {
-			p.logDebug(name, "Persisting bundle to disk in progress.")
+			p.log(name).Debug("Persisting bundle to disk in progress.")
 
-			err := p.saveBundleToDisk(name, u.Bundle)
+			err := p.saveBundleToDisk(name, u.Raw)
 			if err != nil {
-				p.logError(name, "Persisting bundle to disk failed: %v", err)
+				p.log(name).Error("Persisting bundle to disk failed: %v", err)
 				p.status[name].SetError(err)
-				p.downloaders[name].ClearCache()
+				if !p.stopped {
+					p.downloaders[name].ClearCache()
+				}
 				return
 			}
-			p.logDebug(name, "Bundle persisted to disk successfully at path %v.", filepath.Join(p.bundlePersistPath, name))
+			p.log(name).Debug("Bundle persisted to disk successfully at path %v.", filepath.Join(p.bundlePersistPath, name))
 		}
 
 		p.status[name].SetError(nil)
 		p.status[name].SetActivateSuccess(u.Bundle.Manifest.Revision)
 
 		if u.ETag != "" {
-			p.logInfo(name, "Bundle downloaded and activated successfully. Etag updated to %v.", u.ETag)
+			p.log(name).Info("Bundle loaded and activated successfully. Etag updated to %v.", u.ETag)
 		} else {
-			p.logInfo(name, "Bundle downloaded and activated successfully.")
+			p.log(name).Info("Bundle loaded and activated successfully.")
 		}
 		p.etags[name] = u.ETag
 
@@ -390,7 +422,7 @@ func (p *Plugin) process(ctx context.Context, name string, u download.Update) {
 	}
 
 	if etag, ok := p.etags[name]; ok && u.ETag == etag {
-		p.logDebug(name, "Bundle download skipped, server replied with not modified.")
+		p.log(name).Debug("Bundle load skipped, server replied with not modified.")
 		p.status[name].SetError(nil)
 		return
 	}
@@ -414,14 +446,14 @@ func (p *Plugin) checkPluginReadiness() {
 }
 
 func (p *Plugin) activate(ctx context.Context, name string, b *bundle.Bundle) error {
-	p.logDebug(name, "Bundle activation in progress. Opening storage transaction.")
+	p.log(name).Debug("Bundle activation in progress. Opening storage transaction.")
 
 	params := storage.WriteParams
 	params.Context = storage.NewContext()
 
 	err := storage.Txn(ctx, p.manager.Store, params, func(txn storage.Transaction) error {
-		p.logDebug(name, "Opened storage transaction (%v).", txn.ID())
-		defer p.logDebug(name, "Closing storage transaction (%v).", txn.ID())
+		p.log(name).Debug("Opened storage transaction (%v).", txn.ID())
+		defer p.log(name).Debug("Closing storage transaction (%v).", txn.ID())
 
 		// Compile the bundle modules with a new compiler and set it on the
 		// transaction params for use by onCommit hooks.
@@ -433,6 +465,7 @@ func (p *Plugin) activate(ctx context.Context, name string, b *bundle.Bundle) er
 			Ctx:      ctx,
 			Store:    p.manager.Store,
 			Txn:      txn,
+			TxnCtx:   params.Context,
 			Compiler: compiler,
 			Metrics:  p.status[name].Metrics,
 			Bundles:  map[string]*bundle.Bundle{name: b},
@@ -446,6 +479,13 @@ func (p *Plugin) activate(ctx context.Context, name string, b *bundle.Bundle) er
 
 		plugins.SetCompilerOnContext(params.Context, compiler)
 
+		resolvers, err := bundleUtils.LoadWasmResolversFromStore(ctx, p.manager.Store, txn, nil)
+		if err != nil {
+			return err
+		}
+
+		plugins.SetWasmResolversOnContext(params.Context, resolvers)
+
 		return activateErr
 	})
 
@@ -461,32 +501,6 @@ func (p *Plugin) persistBundle(name string) bool {
 	return bundleSrc.Persist
 }
 
-func (p *Plugin) logError(bundleName string, fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields(bundleName)).Errorf(fmt, a...)
-}
-
-func (p *Plugin) logInfo(bundleName string, fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields(bundleName)).Infof(fmt, a...)
-}
-
-func (p *Plugin) logDebug(bundleName string, fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields(bundleName)).Debugf(fmt, a...)
-}
-
-func (p *Plugin) logWarn(bundleName string, fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields(bundleName)).Warnf(fmt, a...)
-}
-
-func (p *Plugin) logrusFields(bundleName string) logrus.Fields {
-
-	f := logrus.Fields{
-		"plugin": Name,
-		"name":   bundleName,
-	}
-
-	return f
-}
-
 // configDelta will return a map of new bundle sources, updated bundle sources, and a set of deleted bundle names
 func (p *Plugin) configDelta(newConfig *Config) (map[string]*Source, map[string]*Source, map[string]struct{}) {
 	deletedBundles := map[string]struct{}{}
@@ -510,22 +524,22 @@ func (p *Plugin) configDelta(newConfig *Config) (map[string]*Source, map[string]
 	return newBundles, updatedBundles, deletedBundles
 }
 
-func (p *Plugin) saveBundleToDisk(name string, b *bundle.Bundle) error {
+func (p *Plugin) saveBundleToDisk(name string, raw io.Reader) error {
 
 	bundleDir := filepath.Join(p.bundlePersistPath, name)
 	tmpFile := filepath.Join(bundleDir, ".bundle.tar.gz.tmp")
 	bundleFile := filepath.Join(bundleDir, "bundle.tar.gz")
 
-	saveErr := saveCurrentBundleToDisk(bundleDir, ".bundle.tar.gz.tmp", b)
+	saveErr := saveCurrentBundleToDisk(bundleDir, ".bundle.tar.gz.tmp", raw)
 	if saveErr != nil {
-		p.logWarn(name, "Failed to save new bundle to disk: %v", saveErr)
+		p.log(name).Error("Failed to save new bundle to disk: %v", saveErr)
 
 		if err := os.Remove(tmpFile); err != nil {
-			p.logWarn(name, "Failed to remove temp file ('%s'): %v", tmpFile, err)
+			p.log(name).Warn("Failed to remove temp file ('%s'): %v", tmpFile, err)
 		}
 
 		if _, err := os.Stat(bundleFile); err == nil {
-			p.logWarn(name, "Older version of activated bundle persisted, ignoring error")
+			p.log(name).Warn("Older version of activated bundle persisted, ignoring error")
 			return nil
 		}
 		return saveErr
@@ -534,13 +548,7 @@ func (p *Plugin) saveBundleToDisk(name string, b *bundle.Bundle) error {
 	return os.Rename(tmpFile, bundleFile)
 }
 
-func saveCurrentBundleToDisk(path, filename string, b *bundle.Bundle) error {
-	var buf bytes.Buffer
-
-	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(*b); err != nil {
-		return err
-	}
-
+func saveCurrentBundleToDisk(path, filename string, raw io.Reader) error {
 	if _, err := os.Stat(path); os.IsNotExist(err) {
 		err = os.MkdirAll(path, os.ModePerm)
 		if err != nil {
@@ -548,14 +556,21 @@ func saveCurrentBundleToDisk(path, filename string, b *bundle.Bundle) error {
 		}
 	}
 
-	if err := ioutil.WriteFile(filepath.Join(path, filename), buf.Bytes(), 0644); err != nil {
-		return err
+	if raw == nil {
+		return fmt.Errorf("no raw bundle bytes to persist to disk")
 	}
 
-	return nil
+	dest, err := os.OpenFile(filepath.Join(path, filename), os.O_WRONLY|os.O_CREATE|os.O_TRUNC, 0644)
+	if err != nil {
+		return err
+	}
+	defer dest.Close()
+
+	_, err = io.Copy(dest, raw)
+	return err
 }
 
-func loadBundleFromDisk(path, name string) (*bundle.Bundle, error) {
+func loadBundleFromDisk(path, name string, src *Source) (*bundle.Bundle, error) {
 	bundlePath := filepath.Join(path, name, "bundle.tar.gz")
 
 	if _, err := os.Stat(bundlePath); err == nil {
@@ -565,12 +580,17 @@ func loadBundleFromDisk(path, name string) (*bundle.Bundle, error) {
 		}
 		defer f.Close()
 
-		b, err := bundle.NewReader(f).Read()
+		r := bundle.NewReader(f)
+
+		if src != nil {
+			r = r.WithBundleVerificationConfig(src.Signing)
+		}
+
+		b, err := r.Read()
 		if err != nil {
 			return nil, err
 		}
 		return &b, nil
-
 	} else if os.IsNotExist(err) {
 		return nil, nil
 	} else {
@@ -578,10 +598,61 @@ func loadBundleFromDisk(path, name string) (*bundle.Bundle, error) {
 	}
 }
 
-func getDefaultBundlePersistPath() (string, error) {
-	pwd, err := os.Getwd()
+func (p *Plugin) log(name string) logging.Logger {
+	if p.logger == nil {
+		p.logger = logging.NewStandardLogger()
+	}
+	return p.logger.WithFields(map[string]interface{}{"name": name, "plugin": Name})
+}
+
+func (p *Plugin) getBundlePersistPath() (string, error) {
+	persistDir, err := p.manager.Config.GetPersistenceDirectory()
 	if err != nil {
 		return "", err
 	}
-	return filepath.Join(pwd, ".opa", "bundles"), nil
+
+	return filepath.Join(persistDir, "bundles"), nil
+}
+
+type bundleLoader interface {
+	Start(context.Context)
+	Stop(context.Context)
+	ClearCache()
+}
+
+type fileLoader struct {
+	name           string
+	path           string
+	bvc            *bundle.VerificationConfig
+	sizeLimitBytes int64
+	f              func(context.Context, string, download.Update)
+}
+
+func (fl *fileLoader) Start(context.Context) {
+	go func() {
+		var u download.Update
+		u.Metrics = metrics.New()
+		f, err := os.Open(fl.path)
+		u.Error = err
+		if err == nil {
+			defer f.Close()
+			b, err := bundle.NewReader(f).
+				WithMetrics(u.Metrics).
+				WithBundleVerificationConfig(fl.bvc).
+				WithSizeLimitBytes(fl.sizeLimitBytes).Read()
+			u.Error = err
+			if err == nil {
+				u.Bundle = &b
+			}
+		}
+		fl.f(context.Background(), fl.name, u)
+	}()
+}
+
+func (*fileLoader) Stop(context.Context) {
+
+}
+
+func (*fileLoader) ClearCache() {
+
 }
