@@ -2,6 +2,7 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// nolint: goconst // string duplication is for test readability.
 package bundle
 
 import (
@@ -9,10 +10,12 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"io"
 	"io/ioutil"
 	"net/http"
 	"net/http/httptest"
 	"os"
+	"path"
 	"path/filepath"
 	"reflect"
 	"sort"
@@ -20,10 +23,15 @@ import (
 	"testing"
 	"time"
 
+	"github.com/open-policy-agent/opa/internal/file/archive"
+
+	"github.com/open-policy-agent/opa/util/test"
+
 	"github.com/open-policy-agent/opa/ast"
 	"github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/config"
 	"github.com/open-policy-agent/opa/download"
+	"github.com/open-policy-agent/opa/keys"
 	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
 	"github.com/open-policy-agent/opa/storage"
@@ -48,7 +56,7 @@ func TestPluginOneShot(t *testing.T) {
 		Manifest: bundle.Manifest{Revision: "quickbrownfaux"},
 		Data:     util.MustUnmarshalJSON([]byte(`{"foo": {"bar": 1, "baz": "qux"}}`)).(map[string]interface{}),
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "/foo/bar",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -102,6 +110,70 @@ func TestPluginStart(t *testing.T) {
 	}
 }
 
+func TestStop(t *testing.T) {
+	var longPollTimeout int64 = 3
+	done := make(chan struct{})
+	tsURLBase := "/opa-test/"
+
+	ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		if !strings.HasPrefix(r.URL.Path, tsURLBase) {
+			t.Fatalf("Invalid request URL path: %s, expected prefix %s", r.URL.Path, tsURLBase)
+		}
+
+		close(done)
+
+		// simulate long operation
+		time.Sleep(time.Duration(longPollTimeout) * time.Second)
+		fmt.Fprintln(w) // Note: this is an invalid bundle and will fail the download
+	}))
+	defer ts.Close()
+
+	ctx := context.Background()
+	manager := getTestManager()
+
+	serviceName := "test-svc"
+	err := manager.Reconfigure(&config.Config{
+		Services: []byte(fmt.Sprintf("{\"%s\":{ \"url\": \"%s\"}}", serviceName, ts.URL+tsURLBase)),
+	})
+	if err != nil {
+		t.Fatalf("Error configuring plugin manager: %s", err)
+	}
+
+	baseConf := download.Config{Polling: download.PollingConfig{LongPollingTimeoutSeconds: &longPollTimeout}}
+
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
+	bundleName := "test-bundle"
+	plugin.status[bundleName] = &Status{Name: bundleName}
+
+	callback := func(ctx context.Context, u download.Update) {
+		plugin.oneShot(ctx, bundleName, u)
+	}
+	plugin.downloaders[bundleName] = download.New(baseConf, plugin.manager.Client(serviceName), bundleName).WithCallback(callback)
+
+	err = plugin.Start(ctx)
+	if err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	// Give time for a long poll request to be initiated
+	<-done
+
+	plugin.Stop(ctx)
+
+	if plugin.status[bundleName].Code != errCode {
+		t.Fatalf("expected error code %v but got %v", errCode, plugin.status[bundleName].Code)
+	}
+
+	if !strings.Contains(plugin.status[bundleName].Message, "context canceled") {
+		t.Fatalf("unexpected error message %v", plugin.status[bundleName].Message)
+	}
+}
+
 func TestPluginOneShotBundlePersistence(t *testing.T) {
 
 	ctx := context.Background()
@@ -146,7 +218,7 @@ func TestPluginOneShotBundlePersistence(t *testing.T) {
 		Manifest: bundle.Manifest{Revision: "quickbrownfaux"},
 		Data:     util.MustUnmarshalJSON([]byte(`{"foo": {"bar": 1, "baz": "qux"}}`)).(map[string]interface{}),
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				URL:    "/foo/bar.rego",
 				Path:   "/foo/bar.rego",
 				Parsed: ast.MustParseModule(module),
@@ -157,11 +229,16 @@ func TestPluginOneShotBundlePersistence(t *testing.T) {
 
 	b.Manifest.Init()
 
-	plugin.oneShot(ctx, bundleName, download.Update{Bundle: &b, Metrics: metrics.New()})
+	var buf bytes.Buffer
+	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(b); err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	plugin.oneShot(ctx, bundleName, download.Update{Bundle: &b, Metrics: metrics.New(), Raw: &buf})
 
 	ensurePluginState(t, plugin, plugins.StateOK)
 
-	result, err := loadBundleFromDisk(plugin.bundlePersistPath, bundleName)
+	result, err := loadBundleFromDisk(plugin.bundlePersistPath, bundleName, nil)
 	if err != nil {
 		t.Fatal("unexpected error:", err)
 	}
@@ -202,6 +279,105 @@ func TestPluginOneShotBundlePersistence(t *testing.T) {
 	}
 }
 
+func TestPluginOneShotSignedBundlePersistence(t *testing.T) {
+
+	ctx := context.Background()
+	manager := getTestManager()
+
+	dir, err := ioutil.TempDir("", "")
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	defer os.RemoveAll(dir)
+
+	bundleName := "test-bundle"
+	vc := bundle.NewVerificationConfig(map[string]*bundle.KeyConfig{"foo": {Key: "secret", Algorithm: "HS256"}}, "foo", "", nil)
+	bundleSource := Source{
+		Persist: true,
+		Signing: vc,
+	}
+
+	bundles := map[string]*Source{}
+	bundles[bundleName] = &bundleSource
+
+	plugin := New(&Config{Bundles: bundles}, manager)
+
+	plugin.status[bundleName] = &Status{Name: bundleName, Metrics: metrics.New()}
+	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
+	plugin.bundlePersistPath = filepath.Join(dir, ".opa")
+
+	ensurePluginState(t, plugin, plugins.StateNotReady)
+
+	// simulate a bundle download error with no bundle on disk
+	plugin.oneShot(ctx, bundleName, download.Update{Error: fmt.Errorf("unknown error")})
+
+	if plugin.status[bundleName].Message == "" {
+		t.Fatal("expected error but got none")
+	}
+
+	ensurePluginState(t, plugin, plugins.StateNotReady)
+
+	// download a signed bundle and persist to disk. Then verify the bundle persisted to disk
+	signedTokenHS256 := `eyJhbGciOiJIUzI1NiJ9.eyJmaWxlcyI6W3sibmFtZSI6Ii5tYW5pZmVzdCIsImhhc2giOiI1MDdhMmMzOGExNDQxZGI1OGQyY2I4Nzk4MmM0MmFhOTFhNDM0MmVmNDIyYTZiNTQyZWRkZWJlZWY2ZjA0MTJmIiwiYWxnb3JpdGhtIjoiU0hBLTI1NiJ9LHsibmFtZSI6ImV4YW1wbGUxL2RhdGEuanNvbiIsImhhc2giOiI3YTM4YmY4MWYzODNmNjk0MzNhZDZlOTAwZDM1YjNlMjM4NTU5M2Y3NmE3YjdhYjVkNDM1NWI4YmE0MWVlMjRiIiwiYWxnb3JpdGhtIjoiU0hBLTI1NiJ9LHsibmFtZSI6ImV4YW1wbGUyL2RhdGEuanNvbiIsImhhc2giOiI5ZTRmMTg5YmY0MDc5ZDFiNmViNjQ0Njg3OTg2NmNkNWYzOWMyNjg4MGQ0ZmI1MThmNGUwMWNkMWJiZmU1MTNlIiwiYWxnb3JpdGhtIjoiU0hBLTI1NiJ9XX0.jCLRMyys5u8S2sTS2pWWY82IAeKDpLh3S641_BskCtY`
+
+	files := [][2]string{
+		{"/.manifest", `{"revision": "quickbrownfaux"}`},
+		{"/.signatures.json", fmt.Sprintf(`{"signatures": ["%v"]}`, signedTokenHS256)},
+		{"/example1/data.json", `{"foo": "bar"}`},
+		{"/example2/data.json", `{"x": true}`},
+	}
+
+	buf := archive.MustWriteTarGz(files)
+
+	var dup bytes.Buffer
+	tee := io.TeeReader(buf, &dup)
+	reader := bundle.NewReader(tee).WithBundleVerificationConfig(vc)
+	b, err := reader.Read()
+	if err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	plugin.oneShot(ctx, bundleName, download.Update{Bundle: &b, Metrics: metrics.New(), Raw: &dup})
+
+	ensurePluginState(t, plugin, plugins.StateOK)
+
+	// load signed bundle from disk
+	result, err := loadBundleFromDisk(plugin.bundlePersistPath, bundleName, bundles[bundleName])
+	if err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	if !result.Equal(b) {
+		t.Fatal("expected the downloaded bundle to be equal to the one loaded from disk")
+	}
+
+	// simulate a bundle download error and verify that the bundle on disk is activated
+	plugin.oneShot(ctx, bundleName, download.Update{Error: fmt.Errorf("unknown error")})
+
+	ensurePluginState(t, plugin, plugins.StateOK)
+
+	txn := storage.NewTransactionOrDie(ctx, manager.Store)
+	defer manager.Store.Abort(ctx, txn)
+
+	ids, err := manager.Store.ListPolicies(ctx, txn)
+	if err != nil {
+		t.Fatal(err)
+	} else if len(ids) != 0 {
+		t.Fatal("Expected no policy")
+	}
+
+	data, err := manager.Store.Read(ctx, txn, storage.Path{})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	expData := util.MustUnmarshalJSON([]byte(`{"example1": {"foo": "bar"}, "example2": {"x": true}, "system": {"bundles": {"test-bundle": {"manifest": {"revision": "quickbrownfaux", "roots": [""]}}}}}`))
+	if !reflect.DeepEqual(data, expData) {
+		t.Fatalf("Bad data content. Exp:\n%v\n\nGot:\n\n%v", expData, data)
+	}
+}
+
 func TestLoadAndActivateBundlesFromDisk(t *testing.T) {
 
 	ctx := context.Background()
@@ -241,7 +417,7 @@ func TestLoadAndActivateBundlesFromDisk(t *testing.T) {
 		Manifest: bundle.Manifest{Revision: "quickbrownfaux"},
 		Data:     util.MustUnmarshalJSON([]byte(`{"foo": {"bar": 1, "baz": "qux"}}`)).(map[string]interface{}),
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				URL:    "/foo/bar.rego",
 				Path:   "/foo/bar.rego",
 				Parsed: ast.MustParseModule(module),
@@ -252,7 +428,12 @@ func TestLoadAndActivateBundlesFromDisk(t *testing.T) {
 
 	b.Manifest.Init()
 
-	err = plugin.saveBundleToDisk(bundleName, &b)
+	var buf bytes.Buffer
+	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(b); err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	err = plugin.saveBundleToDisk(bundleName, &buf)
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
@@ -524,7 +705,12 @@ func TestPluginOneShotActivationConflictingRoots(t *testing.T) {
 func TestPluginOneShotActivationPrefixMatchingRoots(t *testing.T) {
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleNames := []string{"test-bundle1", "test-bundle2"}
 
 	for _, name := range bundleNames {
@@ -581,10 +767,15 @@ func TestPluginListener(t *testing.T) {
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
 	ch := make(chan Status)
 
-	plugin.Register("test", func(status Status) {
+	listenerName := "test"
+	plugin.Register(listenerName, func(status Status) {
 		ch <- status
 	})
 
+	if len(plugin.listeners) != 1 || plugin.listeners[listenerName] == nil {
+		t.Fatal("Listener not properly registered")
+	}
+
 	module := "package gork\np[x] { x = 1 }"
 
 	b := bundle.Bundle{
@@ -645,6 +836,11 @@ func TestPluginListener(t *testing.T) {
 
 	// Nothing should have changed in the update
 	validateStatus(t, s4, s3.ActiveRevision, false)
+
+	plugin.Unregister(listenerName)
+	if len(plugin.listeners) != 0 {
+		t.Fatal("Listener not properly unregistered")
+	}
 }
 
 func isErrStatus(s Status) bool {
@@ -655,7 +851,7 @@ func validateStatus(t *testing.T, actual Status, expected string, expectStatusEr
 	t.Helper()
 
 	if expectStatusErr && !isErrStatus(actual) {
-		t.Errorf("Expected status to be in an error state, but no error has occured.")
+		t.Errorf("Expected status to be in an error state, but no error has occurred.")
 	} else if !expectStatusErr && isErrStatus(actual) {
 		t.Errorf("Unexpected error status %s", actual)
 	}
@@ -668,7 +864,12 @@ func validateStatus(t *testing.T, actual Status, expected string, expectStatusEr
 func TestPluginListenerErrorClearedOn304(t *testing.T) {
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
@@ -715,7 +916,12 @@ func TestPluginListenerErrorClearedOn304(t *testing.T) {
 func TestPluginBulkListener(t *testing.T) {
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleNames := []string{
 		"b1",
 		"b2",
@@ -727,10 +933,15 @@ func TestPluginBulkListener(t *testing.T) {
 	}
 	bulkChan := make(chan map[string]*Status)
 
-	plugin.RegisterBulkListener("bulk test", func(status map[string]*Status) {
+	listenerName := "bulk test"
+	plugin.RegisterBulkListener(listenerName, func(status map[string]*Status) {
 		bulkChan <- status
 	})
 
+	if len(plugin.bulkListeners) != 1 || plugin.bulkListeners[listenerName] == nil {
+		t.Fatal("Bulk listener not properly registered")
+	}
+
 	module := "package gork\np[x] { x = 1 }"
 
 	b := bundle.Bundle{
@@ -835,8 +1046,9 @@ func TestPluginBulkListener(t *testing.T) {
 	go plugin.oneShot(ctx, bundleNames[0], download.Update{})
 	s4 := <-bulkChan
 
-	if !reflect.DeepEqual(s3, s4) {
-		t.Fatalf("Expected: %v but got: %v", s3, s4)
+	s = s4[bundleNames[0]]
+	if s.ActiveRevision != "fancybluederg" || s.Code != "" || s.Message != "" || len(s.Errors) != 0 {
+		t.Errorf("Unexpected same status update for bundle %q, got: %v", bundleNames[0], s)
 	}
 
 	// Test updates the other bundles
@@ -883,12 +1095,22 @@ func TestPluginBulkListener(t *testing.T) {
 			t.Errorf("Expected bundle %q to have an empty status, got: %+v", name, s5)
 		}
 	}
+
+	plugin.UnregisterBulkListener(listenerName)
+	if len(plugin.bulkListeners) != 0 {
+		t.Fatal("Bulk listener not properly unregistered")
+	}
 }
 
 func TestPluginBulkListenerStatusCopyOnly(t *testing.T) {
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleNames := []string{
 		"b1",
 		"b2",
@@ -940,7 +1162,12 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
@@ -964,10 +1191,7 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 		if err := manager.Store.UpsertPolicy(ctx, txn, "some/id2", []byte(`package a.a4`)); err != nil {
 			return err
 		}
-		if err := manager.Store.UpsertPolicy(ctx, txn, "some/id3", []byte(`package a.a6`)); err != nil {
-			return err
-		}
-		return nil
+		return manager.Store.UpsertPolicy(ctx, txn, "some/id3", []byte(`package a.a6`))
 	}); err != nil {
 		t.Fatal(err)
 	}
@@ -985,7 +1209,7 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "bundle/id1",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1001,7 +1225,7 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 	// that the manifest has been written to storage.
 	expData := util.MustUnmarshalJSON([]byte(`{"a1": "foo", "a3": "x2", "a5": "x3"}`))
 	expIds := []string{filepath.Join(bundleName, "bundle/id1"), "some/id2", "some/id3"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux", nil)
 
 	// Activate a bundle that is scoped to a/a3 ad a/a6. Include a function
 	// inside package a.a4 that we can depend on outside of the bundle scope to
@@ -1009,14 +1233,20 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 	module = "package a.a4\n\nbar=1\n\nfunc(x) = x"
 
 	b = bundle.Bundle{
-		Manifest: bundle.Manifest{Revision: "quickbrownfaux-2", Roots: &[]string{"a/a3", "a/a4"}},
+		Manifest: bundle.Manifest{Revision: "quickbrownfaux-2", Roots: &[]string{"a/a3", "a/a4"},
+			Metadata: map[string]interface{}{
+				"a": map[string]interface{}{
+					"a1": "deadbeef",
+				},
+			},
+		},
 		Data: map[string]interface{}{
 			"a": map[string]interface{}{
 				"a3": "foo",
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "bundle/id2",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1030,7 +1260,10 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 	// Ensure a/a5-a6 are intact. a3 and a4 are overwritten by bundle.
 	expData = util.MustUnmarshalJSON([]byte(`{"a3": "foo", "a5": "x3"}`))
 	expIds = []string{filepath.Join(bundleName, "bundle/id2"), "some/id3"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2",
+		map[string]interface{}{
+			"a": map[string]interface{}{"a1": "deadbeef"},
+		})
 
 	// Upsert policy outside of bundle scope that depends on bundle.
 	if err := storage.Txn(ctx, manager.Store, storage.WriteParams, func(txn storage.Transaction) error {
@@ -1051,14 +1284,22 @@ func TestPluginActivateScopedBundle(t *testing.T) {
 	// Ensure bundle activation failed by checking that previous revision is
 	// still active.
 	expIds = []string{filepath.Join(bundleName, "bundle/id2"), "not_scoped", "some/id3"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2",
+		map[string]interface{}{
+			"a": map[string]interface{}{"a1": "deadbeef"},
+		})
 }
 
 func TestPluginSetCompilerOnContext(t *testing.T) {
 
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
@@ -1073,7 +1314,7 @@ func TestPluginSetCompilerOnContext(t *testing.T) {
 		Manifest: bundle.Manifest{Revision: "quickbrownfaux"},
 		Data:     map[string]interface{}{},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "/test.rego",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1086,12 +1327,12 @@ func TestPluginSetCompilerOnContext(t *testing.T) {
 	events := []storage.TriggerEvent{}
 
 	if err := storage.Txn(ctx, manager.Store, storage.WriteParams, func(txn storage.Transaction) error {
-		manager.Store.Register(ctx, txn, storage.TriggerConfig{
+		_, err := manager.Store.Register(ctx, txn, storage.TriggerConfig{
 			OnCommit: func(ctx context.Context, txn storage.Transaction, event storage.TriggerEvent) {
 				events = append(events, event)
 			},
 		})
-		return nil
+		return err
 	}); err != nil {
 		t.Fatal(err)
 	}
@@ -1310,7 +1551,7 @@ func TestPluginReconfigure(t *testing.T) {
 		})
 	}
 	if len(stages) != updateCount {
-		t.Fatalf("Expected to have recieved %d updates, got %d", len(stages), updateCount)
+		t.Fatalf("Expected to have received %d updates, got %d", len(stages), updateCount)
 	}
 }
 
@@ -1318,7 +1559,12 @@ func TestPluginRequestVsDownloadTimestamp(t *testing.T) {
 
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
@@ -1362,7 +1608,12 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 
 	ctx := context.Background()
 	manager := getTestManager()
-	plugin := Plugin{manager: manager, status: map[string]*Status{}, etags: map[string]string{}, downloaders: map[string]*download.Downloader{}}
+	plugin := Plugin{
+		manager:     manager,
+		status:      map[string]*Status{},
+		etags:       map[string]string{},
+		downloaders: map[string]bundleLoader{},
+	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
 	plugin.downloaders[bundleName] = download.New(download.Config{}, plugin.manager.Client(""), bundleName)
@@ -1370,7 +1621,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 	// Start with a "legacy" style config for a single bundle
 	plugin.config = Config{
 		Bundles: map[string]*Source{
-			bundleName: &Source{
+			bundleName: {
 				Service: "s1",
 			},
 		},
@@ -1389,7 +1640,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "bundle/id1",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1404,7 +1655,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 	// Ensure it has been activated
 	expData := util.MustUnmarshalJSON([]byte(`{"a2": "foo"}`))
 	expIds := []string{"bundle/id1"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux", nil)
 
 	if plugin.config.IsMultiBundle() {
 		t.Fatalf("Expected plugin to be in non-multi bundle config mode")
@@ -1413,7 +1664,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 	// Update to the newer style config with the same bundle
 	multiBundleConf := &Config{
 		Bundles: map[string]*Source{
-			bundleName: &Source{
+			bundleName: {
 				Service: "s1",
 			},
 		},
@@ -1425,7 +1676,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleSameBundle(t *testing.T) {
 
 	// The only thing that should have changed is the store id for the policy
 	expIds = []string{"test-bundle/bundle/id1"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux-2", nil)
 
 	// Make sure the legacy path is gone now that we are in multi-bundle mode
 	var actual string
@@ -1456,7 +1707,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 		manager:     manager,
 		status:      map[string]*Status{},
 		etags:       map[string]string{},
-		downloaders: map[string]*download.Downloader{},
+		downloaders: map[string]bundleLoader{},
 	}
 	bundleName := "test-bundle"
 	plugin.status[bundleName] = &Status{Name: bundleName}
@@ -1485,7 +1736,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 	// Start with a "legacy" style config for a single bundle
 	plugin.config = Config{
 		Bundles: map[string]*Source{
-			bundleName: &Source{
+			bundleName: {
 				Config:  downloadConf,
 				Service: serviceName,
 			},
@@ -1505,7 +1756,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "bundle/id1",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1520,7 +1771,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 	// Ensure it has been activated
 	expData := util.MustUnmarshalJSON([]byte(`{"a2": "foo"}`))
 	expIds := []string{"bundle/id1"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, bundleName, "quickbrownfaux", nil)
 
 	if plugin.config.IsMultiBundle() {
 		t.Fatalf("Expected plugin to be in non-multi bundle config mode")
@@ -1529,7 +1780,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 	// Update to the newer style config with a new bundle
 	multiBundleConf := &Config{
 		Bundles: map[string]*Source{
-			"b2": &Source{
+			"b2": {
 				Config:  downloadConf,
 				Service: serviceName,
 			},
@@ -1542,14 +1793,14 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 
 	module = "package a.c\n\nbar=1"
 	b = bundle.Bundle{
-		Manifest: bundle.Manifest{Revision: fmt.Sprintf("b2-1"), Roots: &[]string{"a/b2", "a/c"}},
+		Manifest: bundle.Manifest{Revision: "b2-1", Roots: &[]string{"a/b2", "a/c"}},
 		Data: map[string]interface{}{
 			"a": map[string]interface{}{
 				"b2": "foo",
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "id1",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1561,7 +1812,7 @@ func TestUpgradeLegacyBundleToMuiltiBundleNewBundles(t *testing.T) {
 
 	expData = util.MustUnmarshalJSON([]byte(`{"b2": "foo"}`))
 	expIds = []string{"b2/id1"}
-	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, "b2", "b2-1")
+	validateStoreState(ctx, t, manager.Store, "/a", expData, expIds, "b2", "b2-1", nil)
 
 	// Make sure the legacy path is gone now that we are in multi-bundle mode
 	var actual string
@@ -1600,14 +1851,41 @@ func TestSaveBundleToDiskNew(t *testing.T) {
 	plugin := New(&Config{Bundles: bundles}, manager)
 	plugin.bundlePersistPath = filepath.Join(dir, ".opa")
 
-	b := getTestBundle(t)
-
-	err = plugin.saveBundleToDisk("foo", &b)
+	err = plugin.saveBundleToDisk("foo", getTestRawBundle(t))
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
 }
 
+func TestSaveBundleToDiskNewConfiguredPersistDir(t *testing.T) {
+	dir, err := ioutil.TempDir("", "")
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+	defer os.RemoveAll(dir)
+
+	manager := getTestManager()
+	manager.Config.PersistenceDirectory = &dir
+	bundles := map[string]*Source{}
+	plugin := New(&Config{Bundles: bundles}, manager)
+
+	err = plugin.Start(context.Background())
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	err = plugin.saveBundleToDisk("foo", getTestRawBundle(t))
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	expectBundlePath := filepath.Join(dir, "bundles", "foo", "bundle.tar.gz")
+	_, err = os.Stat(expectBundlePath)
+	if err != nil {
+		t.Errorf("expected bundle persisted at path %v, %v", expectBundlePath, err)
+	}
+}
+
 func TestSaveBundleToDiskOverWrite(t *testing.T) {
 
 	manager := getTestManager()
@@ -1632,7 +1910,7 @@ func TestSaveBundleToDiskOverWrite(t *testing.T) {
 		t.Fatalf("unexpected error %v", err)
 	}
 
-	b2 := writeTestBundleToDisk(t, bundleDir)
+	b2 := writeTestBundleToDisk(t, bundleDir, false)
 
 	module := "package a.a1\n\nbar=1"
 
@@ -1644,7 +1922,7 @@ func TestSaveBundleToDiskOverWrite(t *testing.T) {
 			},
 		},
 		Modules: []bundle.ModuleFile{
-			bundle.ModuleFile{
+			{
 				Path:   "bundle/id1",
 				Parsed: ast.MustParseModule(module),
 				Raw:    []byte(module),
@@ -1653,12 +1931,17 @@ func TestSaveBundleToDiskOverWrite(t *testing.T) {
 	}
 	newBundle.Manifest.Init()
 
-	err = plugin.saveBundleToDisk("foo", &newBundle)
+	var buf bytes.Buffer
+	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(newBundle); err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	err = plugin.saveBundleToDisk("foo", &buf)
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
 
-	actual, err := loadBundleFromDisk(plugin.bundlePersistPath, "foo")
+	actual, err := loadBundleFromDisk(plugin.bundlePersistPath, "foo", nil)
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
@@ -1669,8 +1952,6 @@ func TestSaveBundleToDiskOverWrite(t *testing.T) {
 }
 
 func TestSaveCurrentBundleToDisk(t *testing.T) {
-	b := getTestBundle(t)
-
 	srcDir, err := ioutil.TempDir("", "")
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
@@ -1678,7 +1959,7 @@ func TestSaveCurrentBundleToDisk(t *testing.T) {
 
 	defer os.RemoveAll(srcDir)
 
-	err = saveCurrentBundleToDisk(srcDir, "bundle.tar.gz", &b)
+	err = saveCurrentBundleToDisk(srcDir, "bundle.tar.gz", getTestRawBundle(t))
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
@@ -1686,12 +1967,22 @@ func TestSaveCurrentBundleToDisk(t *testing.T) {
 	if _, err := os.Stat(filepath.Join(srcDir, "bundle.tar.gz")); err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
+
+	err = saveCurrentBundleToDisk(srcDir, "bundle.tar.gz", nil)
+	if err == nil {
+		t.Fatal("expected error but got nil")
+	}
+
+	expErrMsg := "no raw bundle bytes to persist to disk"
+	if err.Error() != expErrMsg {
+		t.Fatalf("expected error: %v but got: %v", expErrMsg, err)
+	}
 }
 
 func TestLoadBundleFromDisk(t *testing.T) {
 
 	// no bundle on disk
-	_, err := loadBundleFromDisk("foo", "bar")
+	_, err := loadBundleFromDisk("foo", "bar", nil)
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
@@ -1712,9 +2003,9 @@ func TestLoadBundleFromDisk(t *testing.T) {
 		t.Fatalf("unexpected error %v", err)
 	}
 
-	b := writeTestBundleToDisk(t, bundleDir)
+	b := writeTestBundleToDisk(t, bundleDir, false)
 
-	result, err := loadBundleFromDisk(dir, bundleName)
+	result, err := loadBundleFromDisk(dir, bundleName, nil)
 	if err != nil {
 		t.Fatal("unexpected error:", err)
 	}
@@ -1724,8 +2015,53 @@ func TestLoadBundleFromDisk(t *testing.T) {
 	}
 }
 
+func TestLoadSignedBundleFromDisk(t *testing.T) {
+
+	// no bundle on disk
+	_, err := loadBundleFromDisk("foo", "bar", nil)
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	// create a test signed bundle and load it from disk
+	dir, err := ioutil.TempDir("", "")
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	defer os.RemoveAll(dir)
+
+	bundleName := "foo"
+	bundleDir := filepath.Join(dir, bundleName)
+
+	err = os.MkdirAll(bundleDir, os.ModePerm)
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	b := writeTestBundleToDisk(t, bundleDir, true)
+
+	src := Source{
+		Signing: bundle.NewVerificationConfig(map[string]*keys.Config{"foo": {Key: "secret", Algorithm: "HS256"}}, "foo", "", nil),
+	}
+
+	result, err := loadBundleFromDisk(dir, bundleName, &src)
+	if err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	if !result.Equal(b) {
+		t.Fatal("expected the test bundle to be equal to the one loaded from disk")
+	}
+
+	if !reflect.DeepEqual(result.Signatures, b.Signatures) {
+		t.Fatal("Expected signatures to be same")
+	}
+}
+
 func TestGetDefaultBundlePersistPath(t *testing.T) {
-	path, err := getDefaultBundlePersistPath()
+	plugin := New(&Config{}, getTestManager())
+	path, err := plugin.getBundlePersistPath()
 	if err != nil {
 		t.Fatalf("unexpected error %v", err)
 	}
@@ -1735,10 +2071,90 @@ func TestGetDefaultBundlePersistPath(t *testing.T) {
 	}
 }
 
-func writeTestBundleToDisk(t *testing.T, srcDir string) bundle.Bundle {
+func TestConfiguredBundlePersistPath(t *testing.T) {
+	persistPath := "/var/opa"
+	manager := getTestManager()
+	manager.Config.PersistenceDirectory = &persistPath
+	plugin := New(&Config{}, manager)
+
+	path, err := plugin.getBundlePersistPath()
+	if err != nil {
+		t.Fatalf("unexpected error %v", err)
+	}
+
+	if path != "/var/opa/bundles" {
+		t.Errorf("expected configured persist path '/var/opa/bundles'")
+	}
+}
+
+func TestPluginUsingFileLoader(t *testing.T) {
+
+	test.WithTempFS(map[string]string{}, func(dir string) {
+
+		b := bundle.Bundle{
+			Data: map[string]interface{}{},
+			Modules: []bundle.ModuleFile{
+				{
+					URL: "test.rego",
+					Raw: []byte(`package test
+
+					p = 7`),
+				},
+			},
+		}
+
+		name := path.Join(dir, "bundle.tar.gz")
+
+		f, err := os.Create(name)
+		if err != nil {
+			t.Fatal(err)
+		}
+
+		if err := bundle.NewWriter(f).Write(b); err != nil {
+			t.Fatal(err)
+		}
+
+		f.Close()
+
+		mgr := getTestManager()
+		url := "file://" + name
+
+		p := New(&Config{Bundles: map[string]*Source{
+			"test": {
+				SizeLimitBytes: 1e5,
+				Resource:       url,
+			},
+		}}, mgr)
+
+		ch := make(chan Status)
+
+		p.Register("test", func(s Status) {
+			ch <- s
+		})
+
+		if err := p.Start(context.Background()); err != nil {
+			t.Fatal(err)
+		}
+
+		s := <-ch
+
+		if s.LastSuccessfulActivation.IsZero() {
+			t.Fatal("expected successful activation")
+		}
+	})
+
+}
+
+func writeTestBundleToDisk(t *testing.T, srcDir string, signed bool) bundle.Bundle {
 	t.Helper()
 
-	b := getTestBundle(t)
+	var b bundle.Bundle
+
+	if signed {
+		b = getTestSignedBundle(t)
+	} else {
+		b = getTestBundle(t)
+	}
 
 	var buf bytes.Buffer
 	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(b); err != nil {
@@ -1776,7 +2192,31 @@ func getTestBundle(t *testing.T) bundle.Bundle {
 	return b
 }
 
-func validateStoreState(ctx context.Context, t *testing.T, store storage.Store, root string, expData interface{}, expIds []string, expBundleName string, expBundleRev string) {
+func getTestSignedBundle(t *testing.T) bundle.Bundle {
+	t.Helper()
+
+	b := getTestBundle(t)
+
+	if err := b.GenerateSignature(bundle.NewSigningConfig("secret", "HS256", ""), "foo", false); err != nil {
+		t.Fatal("Unexpected error:", err)
+	}
+	return b
+}
+
+func getTestRawBundle(t *testing.T) io.Reader {
+	t.Helper()
+
+	b := getTestBundle(t)
+
+	var buf bytes.Buffer
+	if err := bundle.NewWriter(&buf).UseModulePath(true).Write(b); err != nil {
+		t.Fatal("unexpected error:", err)
+	}
+
+	return &buf
+}
+
+func validateStoreState(ctx context.Context, t *testing.T, store storage.Store, root string, expData interface{}, expIds []string, expBundleName string, expBundleRev string, expMetadata map[string]interface{}) {
 	t.Helper()
 	if err := storage.Txn(ctx, store, storage.TransactionParams{}, func(txn storage.Transaction) error {
 		value, err := store.Read(ctx, txn, storage.MustParsePath(root))
@@ -1809,6 +2249,14 @@ func validateStoreState(ctx context.Context, t *testing.T, store storage.Store,
 			return fmt.Errorf("Unexpected revision found on bundle: %s", rev)
 		}
 
+		metadata, err := bundle.ReadBundleMetadataFromStore(ctx, store, txn, expBundleName)
+		if err != nil {
+			return fmt.Errorf("Unexpected error when reading bundle metadata from store: %s", err)
+		}
+		if !reflect.DeepEqual(expMetadata, metadata) {
+			return fmt.Errorf("Unexpected metadata found on bundle: %v", metadata)
+		}
+
 		return nil
 
 	}); err != nil {
