@@ -2,6 +2,7 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// nolint: goconst // string duplication is for test readability.
 package discovery
 
 import (
@@ -16,13 +17,19 @@ import (
 	"testing"
 	"time"
 
+	"github.com/open-policy-agent/opa/logging/test"
+
 	"github.com/open-policy-agent/opa/ast"
 	bundleApi "github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/download"
+	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
 	"github.com/open-policy-agent/opa/plugins/bundle"
+	"github.com/open-policy-agent/opa/plugins/logs"
 	"github.com/open-policy-agent/opa/plugins/status"
+	"github.com/open-policy-agent/opa/server"
 	"github.com/open-policy-agent/opa/storage/inmem"
+	"github.com/open-policy-agent/opa/topdown/cache"
 	"github.com/open-policy-agent/opa/util"
 	"github.com/open-policy-agent/opa/version"
 )
@@ -169,6 +176,173 @@ func TestProcessBundle(t *testing.T) {
 
 }
 
+func TestProcessBundleWithActiveConfig(t *testing.T) {
+
+	ctx := context.Background()
+
+	manager, err := plugins.New([]byte(`{
+		"labels": {"x": "y"},
+		"services": {
+			"localhost": {
+				"url": "http://localhost:9999",
+				"credentials": {"bearer": {"token": "test"}}
+			}
+		},
+		"keys": {
+			"local_key": {
+				"private_key": "local"
+			}
+		},
+		"discovery": {"name": "config"},
+	}`), "test-id", inmem.New())
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	initialBundle := makeDataBundle(1, `
+		{
+			"config": {
+				"services": {
+					"acmecorp": {
+						"url": "https://example.com/control-plane-api/v1",
+						"credentials": {"bearer": {"token": "test-acmecorp"}}
+					}
+				},
+				"bundles": {"test-bundle": {"service": "localhost"}},
+				"status": {"partition_name": "foo"},
+				"decision_logs": {"partition_name": "bar"},
+				"default_decision": "bar/baz",
+				"default_authorization_decision": "baz/qux",
+				"keys": {
+					"global_key": {
+						"scope": "read",
+						"key": "secret"
+					}
+				}
+			}
+		}
+	`)
+
+	disco, err := New(manager)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	_, err = disco.processBundle(ctx, initialBundle)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	actual, err := manager.Config.ActiveConfig()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	expectedConfig := fmt.Sprintf(`{
+		"services": {
+			"acmecorp": {
+				"url": "https://example.com/control-plane-api/v1"
+			}
+		},
+		"labels": {
+			"id": "test-id",
+			"version": %v,
+			"x": "y"
+		},
+		"keys": {
+			"global_key": {
+				"scope": "read"
+			}
+		},
+		"decision_logs": {
+			"partition_name": "bar"
+		},
+		"status": {
+			"partition_name": "foo"
+		},
+		"bundles": {
+			"test-bundle": {
+				"service": "localhost"
+			}
+		},
+		"default_authorization_decision": "baz/qux",
+		"default_decision": "bar/baz"}`, version.Version)
+
+	var expected map[string]interface{}
+	if err := util.Unmarshal([]byte(expectedConfig), &expected); err != nil {
+		t.Fatal(err)
+	}
+
+	if !reflect.DeepEqual(actual, expected) {
+		t.Fatalf("want %v got %v", expected, actual)
+	}
+
+	initialBundle = makeDataBundle(2, `
+		{
+			"config": {
+				"services": {
+					"opa.example.com": {
+						"url": "https://opa.example.com",
+						"credentials": {"bearer": {"token": "test-opa"}}
+					}
+				},
+				"bundles": {"test-bundle-2": {"service": "opa.example.com"}},
+				"decision_logs": {},
+				"keys": {
+					"global_key_2": {
+						"scope": "write",
+						"key": "secret_2"
+					}
+				}
+			}
+		}
+	`)
+
+	_, err = disco.processBundle(ctx, initialBundle)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	actual, err = manager.Config.ActiveConfig()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	expectedConfig2 := fmt.Sprintf(`{
+		"services": {
+			"opa.example.com": {
+				"url": "https://opa.example.com"
+			}
+		},
+		"labels": {
+			"id": "test-id",
+			"version": %v,
+			"x": "y"
+		},
+		"keys": {
+			"global_key_2": {
+				"scope": "write"
+			}
+		},
+		"decision_logs": {},
+		"bundles": {
+			"test-bundle-2": {
+				"service": "opa.example.com"
+			}
+		},
+		"default_authorization_decision": "/system/authz/allow",
+		"default_decision": "/system/main"}`, version.Version)
+
+	var expected2 map[string]interface{}
+	if err := util.Unmarshal([]byte(expectedConfig2), &expected2); err != nil {
+		t.Fatal(err)
+	}
+
+	if !reflect.DeepEqual(actual, expected2) {
+		t.Fatalf("want %v got %v", expected, actual)
+	}
+}
+
 type testFactory struct {
 	p *reconfigureTestPlugin
 }
@@ -190,7 +364,7 @@ func (r *reconfigureTestPlugin) Start(context.Context) error {
 	return nil
 }
 
-func (r *reconfigureTestPlugin) Stop(context.Context) {
+func (*reconfigureTestPlugin) Stop(context.Context) {
 }
 
 func (r *reconfigureTestPlugin) Reconfigure(_ context.Context, config interface{}) {
@@ -823,6 +997,127 @@ func TestStatusUpdatesTimestamp(t *testing.T) {
 	}
 }
 
+func TestStatusMetricsForLogDrops(t *testing.T) {
+
+	ctx := context.Background()
+
+	testLogger := test.New()
+
+	manager, err := plugins.New([]byte(`{
+		"services": {
+			"localhost": {
+				"url": "http://localhost:9999"
+			}
+		},
+		"discovery": {"name": "config"},
+	}`), "test-id", inmem.New(), plugins.ConsoleLogger(testLogger))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	initialBundle := makeDataBundle(1, `
+		{
+			"config": {
+				"status": {"console": true},
+				"decision_logs": {
+					"service": "localhost",
+					"reporting": {
+						"max_decisions_per_second": 1
+					}
+				}
+			}
+		}
+	`)
+
+	disco, err := New(manager, Metrics(metrics.New()))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	ps, err := disco.processBundle(ctx, initialBundle)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// start the decision log and status plugins
+	for _, p := range ps.Start {
+		if err := p.Start(ctx); err != nil {
+			t.Fatal(err)
+		}
+	}
+
+	plugin := logs.Lookup(manager)
+	if plugin == nil {
+		t.Fatal("Expected decision log plugin registered on manager")
+	}
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	event1 := &server.Info{
+		DecisionID: "abc",
+		Path:       "foo/bar",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-1",
+	}
+
+	event2 := &server.Info{
+		DecisionID: "def",
+		Path:       "foo/baz",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-2",
+	}
+
+	event3 := &server.Info{
+		DecisionID: "ghi",
+		Path:       "foo/aux",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-3",
+	}
+
+	_ = plugin.Log(ctx, event1) // event 1 should be written into the decision log encoder
+	_ = plugin.Log(ctx, event2) // event 2 should not be written into the decision log encoder as rate limit exceeded
+	_ = plugin.Log(ctx, event3) // event 3 should not be written into the decision log encoder as rate limit exceeded
+
+	// trigger a status update
+	disco.oneShot(ctx, download.Update{ETag: "etag-1", Bundle: makeDataBundle(1, `{
+		"config": {
+			"bundles": {"test-bundle": {"service": "localhost"}}
+		}
+	}`)})
+
+	status.Lookup(manager).Stop(ctx)
+
+	entries := testLogger.Entries()
+	if len(entries) == 0 {
+		t.Fatal("Expected log entries but got none")
+	}
+
+	// Pick the last entry as it should have the drop count
+	e := entries[len(entries)-1]
+
+	if _, ok := e.Fields["metrics"]; !ok {
+		t.Fatal("Expected metrics")
+	}
+
+	builtInMet := e.Fields["metrics"].(map[string]interface{})["<built-in>"]
+	dropCount := builtInMet.(map[string]interface{})["counter_decision_logs_dropped"]
+
+	actual, err := dropCount.(json.Number).Int64()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Along with event 2 and event 3, event 1 could also get dropped. This happens when the decision log plugin
+	// tries to requeue event 1 after a failed upload attempt to a non-existent remote endpoint
+	if actual < 2 {
+		t.Fatal("Expected at least 2 events to be dropped")
+	}
+}
+
 func makeDataBundle(n int, s string) *bundleApi.Bundle {
 	return &bundleApi.Bundle{
 		Manifest: bundleApi.Manifest{Revision: fmt.Sprintf("test-revision-%v", n)},
@@ -912,3 +1207,72 @@ bundles:
 		t.Fatalf("Expected the bundle to be configured as bundles[0], got: %+v", bp.Config().Bundles)
 	}
 }
+
+func TestInterQueryBuiltinCacheConfigUpdate(t *testing.T) {
+	var config1 *cache.Config
+	var config2 *cache.Config
+	manager, err := plugins.New([]byte(`{
+		"discovery": {"name": "config"},
+		"services": {
+			"localhost": {
+				"url": "http://localhost:9999"
+			}
+		},
+  }`), "test-id", inmem.New())
+	manager.RegisterCacheTrigger(func(c *cache.Config) {
+		if config1 == nil {
+			config1 = c
+		} else if config2 == nil {
+			config2 = c
+		} else {
+			t.Fatal("Expected cache trigger to only be called twice")
+		}
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	testPlugin := &reconfigureTestPlugin{counts: map[string]int{}}
+	testFactory := testFactory{p: testPlugin}
+
+	disco, err := New(manager, Factories(map[string]plugins.Factory{"test_plugin": testFactory}))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	ctx := context.Background()
+
+	initialBundle := makeDataBundle(1, `{
+    "config": {
+      "caching": {
+        "inter_query_builtin_cache": {
+          "max_size_bytes": 100
+        }
+      }
+    }
+  }`)
+
+	disco.oneShot(ctx, download.Update{Bundle: initialBundle})
+
+	// Verify interQueryBuiltinCacheConfig is triggered with initial config
+	if config1 == nil || *config1.InterQueryBuiltinCache.MaxSizeBytes != int64(100) {
+		t.Fatalf("Expected cache max size bytes to be 100 after initial discovery, got: %v", config1.InterQueryBuiltinCache.MaxSizeBytes)
+	}
+
+	// Verify interQueryBuiltinCache is reconfigured
+	updatedBundle := makeDataBundle(2, `{
+    "config": {
+      "caching": {
+        "inter_query_builtin_cache": {
+          "max_size_bytes": 200
+        }
+      }
+    }
+  }`)
+
+	disco.oneShot(ctx, download.Update{Bundle: updatedBundle})
+
+	if config2 == nil || *config2.InterQueryBuiltinCache.MaxSizeBytes != int64(200) {
+		t.Fatalf("Expected cache max size bytes to be 200 after discovery reconfigure, got: %v", config2.InterQueryBuiltinCache.MaxSizeBytes)
+	}
+}
