@@ -9,6 +9,7 @@ import (
 	"context"
 	"encoding/json"
 	"fmt"
+	"math"
 	"math/rand"
 	"net/http"
 	"reflect"
@@ -17,10 +18,12 @@ import (
 	"time"
 
 	"github.com/pkg/errors"
-	"github.com/sirupsen/logrus"
+	"golang.org/x/time/rate"
 
 	"github.com/open-policy-agent/opa/ast"
 	"github.com/open-policy-agent/opa/internal/ref"
+	"github.com/open-policy-agent/opa/logging"
+	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
 	"github.com/open-policy-agent/opa/plugins/rest"
 	"github.com/open-policy-agent/opa/rego"
@@ -210,14 +213,16 @@ const (
 	defaultUploadSizeLimitBytes = int64(32768) // 32KB limit
 	defaultBufferSizeLimitBytes = int64(0)     // unlimited
 	defaultMaskDecisionPath     = "/system/log/mask"
+	logDropCounterName          = "decision_logs_dropped"
 )
 
 // ReportingConfig represents configuration for the plugin's reporting behaviour.
 type ReportingConfig struct {
-	BufferSizeLimitBytes *int64 `json:"buffer_size_limit_bytes,omitempty"` // max size of in-memory buffer
-	UploadSizeLimitBytes *int64 `json:"upload_size_limit_bytes,omitempty"` // max size of upload payload
-	MinDelaySeconds      *int64 `json:"min_delay_seconds,omitempty"`       // min amount of time to wait between successful poll attempts
-	MaxDelaySeconds      *int64 `json:"max_delay_seconds,omitempty"`       // max amount of time to wait between poll attempts
+	BufferSizeLimitBytes  *int64   `json:"buffer_size_limit_bytes,omitempty"`  // max size of in-memory buffer
+	UploadSizeLimitBytes  *int64   `json:"upload_size_limit_bytes,omitempty"`  // max size of upload payload
+	MinDelaySeconds       *int64   `json:"min_delay_seconds,omitempty"`        // min amount of time to wait between successful poll attempts
+	MaxDelaySeconds       *int64   `json:"max_delay_seconds,omitempty"`        // max amount of time to wait between poll attempts
+	MaxDecisionsPerSecond *float64 `json:"max_decisions_per_second,omitempty"` // max number of decision logs to buffer per second
 }
 
 // Config represents the plugin configuration.
@@ -301,6 +306,10 @@ func (c *Config) validateAndInjectDefaults(services []string, plugins []string)
 
 	c.Reporting.UploadSizeLimitBytes = &uploadLimit
 
+	if c.Reporting.BufferSizeLimitBytes != nil && c.Reporting.MaxDecisionsPerSecond != nil {
+		return fmt.Errorf("invalid decision_log config, specify either 'buffer_size_limit_bytes' or 'max_decisions_per_second'")
+	}
+
 	// default the buffer size limit
 	bufferLimit := defaultBufferSizeLimitBytes
 	if c.Reporting.BufferSizeLimitBytes != nil {
@@ -334,6 +343,9 @@ type Plugin struct {
 	reconfig  chan reconfigure
 	mask      *rego.PreparedEvalQuery
 	maskMutex sync.Mutex
+	limiter   *rate.Limiter
+	metrics   metrics.Metrics
+	logger    logging.Logger
 }
 
 type reconfigure struct {
@@ -370,6 +382,12 @@ func New(parsedConfig *Config, manager *plugins.Manager) *Plugin {
 		buffer:   newLogBuffer(*parsedConfig.Reporting.BufferSizeLimitBytes),
 		enc:      newChunkEncoder(*parsedConfig.Reporting.UploadSizeLimitBytes),
 		reconfig: make(chan reconfigure),
+		logger:   manager.Logger().WithFields(map[string]interface{}{"plugin": Name}),
+	}
+
+	if parsedConfig.Reporting.MaxDecisionsPerSecond != nil {
+		limit := *parsedConfig.Reporting.MaxDecisionsPerSecond
+		plugin.limiter = rate.NewLimiter(rate.Limit(limit), int(math.Max(1, limit)))
 	}
 
 	manager.RegisterCompilerTrigger(plugin.compilerUpdated)
@@ -379,6 +397,12 @@ func New(parsedConfig *Config, manager *plugins.Manager) *Plugin {
 	return plugin
 }
 
+// WithMetrics sets the global metrics provider to be used by the plugin.
+func (p *Plugin) WithMetrics(m metrics.Metrics) *Plugin {
+	p.metrics = m
+	return p
+}
+
 // Name identifies the plugin on manager.
 const Name = "decision_logs"
 
@@ -392,7 +416,7 @@ func Lookup(manager *plugins.Manager) *Plugin {
 
 // Start starts the plugin.
 func (p *Plugin) Start(ctx context.Context) error {
-	p.logInfo("Starting decision logger.")
+	p.logger.Info("Starting decision logger.")
 	go p.loop()
 	p.manager.UpdatePluginStatus(Name, &plugins.Status{State: plugins.StateOK})
 	return nil
@@ -400,13 +424,47 @@ func (p *Plugin) Start(ctx context.Context) error {
 
 // Stop stops the plugin.
 func (p *Plugin) Stop(ctx context.Context) {
-	p.logInfo("Stopping decision logger.")
+	p.logger.Info("Stopping decision logger.")
+
+	if _, ok := ctx.Deadline(); ok && p.config.Service != "" {
+		p.flushDecisions(ctx)
+	}
+
 	done := make(chan struct{})
 	p.stop <- done
-	_ = <-done
+	<-done
 	p.manager.UpdatePluginStatus(Name, &plugins.Status{State: plugins.StateNotReady})
 }
 
+func (p *Plugin) flushDecisions(ctx context.Context) {
+	p.logger.Info("Flushing decision logs.")
+
+	done := make(chan bool)
+
+	go func(ctx context.Context, done chan bool) {
+		for ctx.Err() == nil {
+			if _, err := p.oneShot(ctx); err != nil {
+				p.logger.Error("Error flushing decisions: %s", err)
+				// Wait some before retrying, but skip incrementing interval since we are shutting down
+				time.Sleep(1 * time.Second)
+			} else {
+				done <- true
+				return
+			}
+		}
+	}(ctx, done)
+
+	select {
+	case <-done:
+		p.logger.Info("All decisions in buffer uploaded.")
+	case <-ctx.Done():
+		switch ctx.Err() {
+		case context.DeadlineExceeded, context.Canceled:
+			p.logger.Error("Plugin stopped with decisions possibly still in buffer.")
+		}
+	}
+}
+
 // Log appends a decision log event to the buffer for uploading.
 func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
 
@@ -440,14 +498,14 @@ func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
 	err := p.maskEvent(ctx, decision.Txn, &event)
 	if err != nil {
 		// TODO(tsandall): see note below about error handling.
-		p.logError("Log event masking failed: %v.", err)
+		p.logger.Error("Log event masking failed: %v.", err)
 		return nil
 	}
 
 	if p.config.ConsoleLogs {
 		err := p.logEvent(event)
 		if err != nil {
-			p.logError("Failed to log to console: %v.", err)
+			p.logger.Error("Failed to log to console: %v.", err)
 		}
 	}
 
@@ -462,19 +520,7 @@ func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
 	if p.config.Service != "" {
 		p.mtx.Lock()
 		defer p.mtx.Unlock()
-
-		result, err := p.enc.Write(event)
-		if err != nil {
-			// TODO(tsandall): revisit this now that we have an API that
-			// can return an error. Should the default behaviour be to
-			// fail-closed as we do for plugins?
-			p.logError("Log encoding failed: %v.", err)
-			return nil
-		}
-
-		if result != nil {
-			p.bufferChunk(p.buffer, result)
-		}
+		p.encodeAndBufferEvent(event)
 	}
 
 	return nil
@@ -490,7 +536,7 @@ func (p *Plugin) Reconfigure(_ context.Context, config interface{}) {
 	defer p.maskMutex.Unlock()
 	p.mask = nil
 
-	_ = <-done
+	<-done
 }
 
 // compilerUpdated is called when a compiler trigger on the plugin manager
@@ -516,11 +562,11 @@ func (p *Plugin) loop() {
 			uploaded, err = p.oneShot(ctx)
 
 			if err != nil {
-				p.logError("%v.", err)
+				p.logger.Error("%v.", err)
 			} else if uploaded {
-				p.logInfo("Logs uploaded successfully.")
+				p.logger.Info("Logs uploaded successfully.")
 			} else {
-				p.logInfo("Log upload skipped.")
+				p.logger.Debug("Log upload queue was empty.")
 			}
 		}
 
@@ -535,7 +581,7 @@ func (p *Plugin) loop() {
 		}
 
 		if p.config.Service != "" {
-			p.logDebug("Waiting %v before next upload/retry.", delay)
+			p.logger.Debug("Waiting %v before next upload/retry.", delay)
 		}
 
 		timer := time.NewTimer(delay)
@@ -589,10 +635,24 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 			err = uploadChunk(ctx, p.manager.Client(p.config.Service), p.config.PartitionName, bs)
 		}
 		if err != nil {
-			// requeue the chunk
-			p.mtx.Lock()
-			p.bufferChunk(p.buffer, bs)
-			p.mtx.Unlock()
+			if p.limiter != nil {
+				events, decErr := newChunkDecoder(bs).decode()
+				if decErr != nil {
+					continue
+				}
+
+				p.mtx.Lock()
+				for _, event := range events {
+					p.encodeAndBufferEvent(event)
+				}
+				p.mtx.Unlock()
+
+			} else {
+				// requeue the chunk
+				p.mtx.Lock()
+				p.bufferChunk(p.buffer, bs)
+				p.mtx.Unlock()
+			}
 		}
 	}
 
@@ -604,24 +664,50 @@ func (p *Plugin) reconfigure(config interface{}) {
 	newConfig := config.(*Config)
 
 	if reflect.DeepEqual(p.config, *newConfig) {
-		p.logDebug("Decision log uploader configuration unchanged.")
+		p.logger.Debug("Decision log uploader configuration unchanged.")
 		return
 	}
 
-	p.logInfo("Decision log uploader configuration changed.")
+	p.logger.Info("Decision log uploader configuration changed.")
 	p.config = *newConfig
 }
 
+func (p *Plugin) encodeAndBufferEvent(event EventV1) {
+	if p.limiter != nil {
+		if !p.limiter.Allow() {
+			if p.metrics != nil {
+				p.metrics.Counter(logDropCounterName).Incr()
+			}
+
+			p.logger.Error("Decision log dropped as rate limit exceeded. Reduce reporting interval or increase rate limit.")
+			return
+		}
+	}
+
+	result, err := p.enc.Write(event)
+	if err != nil {
+		// TODO(tsandall): revisit this now that we have an API that
+		// can return an error. Should the default behaviour be to
+		// fail-closed as we do for plugins?
+		p.logger.Error("Log encoding failed: %v.", err)
+		return
+	}
+
+	if result != nil {
+		p.bufferChunk(p.buffer, result)
+	}
+}
+
 func (p *Plugin) bufferChunk(buffer *logBuffer, bs []byte) {
 	dropped := buffer.Push(bs)
 	if dropped > 0 {
-		p.logError("Dropped %v chunks from buffer. Reduce reporting interval or increase buffer size.", dropped)
+		p.logger.Error("Dropped %v chunks from buffer. Reduce reporting interval or increase buffer size.", dropped)
 	}
 }
 
 func (p *Plugin) maskEvent(ctx context.Context, txn storage.Transaction, event *EventV1) error {
 
-	err := func() error {
+	mask, err := func() (rego.PreparedEvalQuery, error) {
 
 		p.maskMutex.Lock()
 		defer p.maskMutex.Unlock()
@@ -640,13 +726,13 @@ func (p *Plugin) maskEvent(ctx context.Context, txn storage.Transaction, event *
 
 			pq, err := r.PrepareForEval(context.Background())
 			if err != nil {
-				return err
+				return rego.PreparedEvalQuery{}, err
 			}
 
 			p.mask = &pq
 		}
 
-		return nil
+		return *p.mask, nil
 	}()
 
 	if err != nil {
@@ -658,7 +744,7 @@ func (p *Plugin) maskEvent(ctx context.Context, txn storage.Transaction, event *
 		return err
 	}
 
-	rs, err := p.mask.Eval(
+	rs, err := mask.Eval(
 		ctx,
 		rego.EvalParsedInput(input),
 		rego.EvalTransaction(txn),
@@ -670,17 +756,17 @@ func (p *Plugin) maskEvent(ctx context.Context, txn storage.Transaction, event *
 		return nil
 	}
 
-	mRules, err := resultValueToMaskRules(rs[0].Expressions[0].Value)
+	mRuleSet, err := newMaskRuleSet(
+		rs[0].Expressions[0].Value,
+		func(mRule *maskRule, err error) {
+			p.logger.Error("mask rule skipped: %s: %s", mRule.String(), err.Error())
+		},
+	)
 	if err != nil {
 		return err
 	}
 
-	for _, mRule := range mRules {
-		err := mRule.Mask(event)
-		if err != nil {
-			p.logError("mask rule skipped: %s: %s", mRule.String(), err.Error())
-		}
-	}
+	mRuleSet.Mask(event)
 
 	return nil
 }
@@ -711,35 +797,17 @@ func uploadChunk(ctx context.Context, client rest.Client, partitionName string,
 	}
 }
 
-func (p *Plugin) logError(fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields()).Errorf(fmt, a...)
-}
-
-func (p *Plugin) logInfo(fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields()).Infof(fmt, a...)
-}
-
-func (p *Plugin) logDebug(fmt string, a ...interface{}) {
-	logrus.WithFields(p.logrusFields()).Debugf(fmt, a...)
-}
-
-func (p *Plugin) logrusFields() logrus.Fields {
-	return logrus.Fields{
-		"plugin": Name,
-	}
-}
-
 func (p *Plugin) logEvent(event EventV1) error {
 	eventBuf, err := json.Marshal(&event)
 	if err != nil {
 		return err
 	}
-	fields := logrus.Fields{}
+	fields := map[string]interface{}{}
 	err = util.UnmarshalJSON(eventBuf, &fields)
 	if err != nil {
 		return err
 	}
-	logrus.WithFields(fields).WithFields(logrus.Fields{
+	p.manager.ConsoleLogger().WithFields(fields).WithFields(map[string]interface{}{
 		"type": "openpolicyagent.org/decision_logs",
 	}).Info("Decision Log")
 	return nil
