@@ -2,9 +2,12 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// +build slow
+
 package logs
 
 import (
+	"bytes"
 	"compress/gzip"
 	"context"
 	"encoding/json"
@@ -17,8 +20,11 @@ import (
 	"time"
 
 	"github.com/open-policy-agent/opa/ast"
+	"github.com/open-policy-agent/opa/logging/test"
 	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
+	"github.com/open-policy-agent/opa/plugins/bundle"
+	"github.com/open-policy-agent/opa/plugins/status"
 	"github.com/open-policy-agent/opa/rego"
 	"github.com/open-policy-agent/opa/server"
 	"github.com/open-policy-agent/opa/storage"
@@ -37,8 +43,6 @@ type testPlugin struct {
 	events []EventV1
 }
 
-type testPluginCustomizer func(c *Config)
-
 func (p *testPlugin) Start(context.Context) error {
 	return nil
 }
@@ -107,10 +111,6 @@ func TestPluginSingleBundle(t *testing.T) {
 	}
 }
 
-func TestPluginMultiBundle(t *testing.T) {
-
-}
-
 func TestPluginErrorNoResult(t *testing.T) {
 	ctx := context.Background()
 	manager, _ := plugins.New(nil, "test-instance-id", inmem.New())
@@ -453,10 +453,7 @@ func logServerInfo(id string, input interface{}, result interface{}) *server.Inf
 func TestPluginRequeBufferPreserved(t *testing.T) {
 	ctx := context.Background()
 
-	fixture := newTestFixture(t, func(c *Config) {
-		limit := int64(300)
-		c.Reporting.UploadSizeLimitBytes = &limit
-	})
+	fixture := newTestFixture(t, testFixtureOptions{ReportingUploadSizeLimitBytes: 300})
 	defer fixture.server.stop()
 
 	fixture.server.ch = make(chan []EventV1, 3)
@@ -479,13 +476,468 @@ func TestPluginRequeBufferPreserved(t *testing.T) {
 		t.Fatal("Expected error")
 	}
 
-	_ = <-fixture.server.ch
+	<-fixture.server.ch
 
 	if fixture.plugin.buffer.Len() < bufLen {
 		t.Fatal("Expected buffer to be preserved")
 	}
 }
 
+func TestPluginRateLimitInt(t *testing.T) {
+	ctx := context.Background()
+
+	ts, err := time.Parse(time.RFC3339Nano, "2018-01-01T12:00:00.123456Z")
+	if err != nil {
+		panic(err)
+	}
+
+	numDecisions := 1 // 1 decision per second
+
+	fixture := newTestFixture(t, testFixtureOptions{
+		ReportingMaxDecisionsPerSecond: float64(numDecisions),
+		ReportingUploadSizeLimitBytes:  300,
+	})
+	defer fixture.server.stop()
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	event1 := &server.Info{
+		DecisionID: "abc",
+		Path:       "foo/bar",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-1",
+		Timestamp:  ts,
+	}
+
+	event2 := &server.Info{
+		DecisionID: "def",
+		Path:       "foo/baz",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-2",
+		Timestamp:  ts,
+	}
+
+	_ = fixture.plugin.Log(ctx, event1) // event 1 should be written into the encoder
+
+	bytesWritten := fixture.plugin.enc.bytesWritten
+	if bytesWritten == 0 {
+		t.Fatal("Expected event to be written into the encoder")
+	}
+
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should not be written into the encoder as rate limit exceeded
+
+	if fixture.plugin.enc.bytesWritten != bytesWritten {
+		t.Fatalf("Expected %v bytes written into the encoder but got %v", bytesWritten, fixture.plugin.enc.bytesWritten)
+	}
+
+	time.Sleep(1 * time.Second)
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should now be written into the encoder
+
+	if fixture.plugin.buffer.Len() != 1 {
+		t.Fatalf("Expected buffer length of 1 but got %v", fixture.plugin.buffer.Len())
+	}
+
+	exp := EventV1{
+		Labels: map[string]string{
+			"id":      "test-instance-id",
+			"app":     "example-app",
+			"version": version.Version,
+		},
+		DecisionID:  "abc",
+		Path:        "foo/bar",
+		Input:       &input,
+		Result:      &result,
+		RequestedBy: "test-1",
+		Timestamp:   ts,
+	}
+	compareLogEvent(t, fixture.plugin.buffer.Pop(), exp)
+
+	chunk, err := fixture.plugin.enc.Flush()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	exp = EventV1{
+		Labels: map[string]string{
+			"id":      "test-instance-id",
+			"app":     "example-app",
+			"version": version.Version,
+		},
+		DecisionID:  "def",
+		Path:        "foo/baz",
+		Input:       &input,
+		Result:      &result,
+		RequestedBy: "test-2",
+		Timestamp:   ts,
+	}
+
+	compareLogEvent(t, chunk, exp)
+}
+
+func TestPluginRateLimitFloat(t *testing.T) {
+	ctx := context.Background()
+
+	ts, err := time.Parse(time.RFC3339Nano, "2018-01-01T12:00:00.123456Z")
+	if err != nil {
+		panic(err)
+	}
+
+	numDecisions := 0.1 // 0.1 decision per second ie. 1 decision per 10 seconds
+	fixture := newTestFixture(t, testFixtureOptions{
+		ReportingMaxDecisionsPerSecond: float64(numDecisions),
+		ReportingUploadSizeLimitBytes:  300,
+	})
+	defer fixture.server.stop()
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	event1 := &server.Info{
+		DecisionID: "abc",
+		Path:       "foo/bar",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-1",
+		Timestamp:  ts,
+	}
+
+	event2 := &server.Info{
+		DecisionID: "def",
+		Path:       "foo/baz",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-2",
+		Timestamp:  ts,
+	}
+
+	_ = fixture.plugin.Log(ctx, event1) // event 1 should be written into the encoder
+
+	bytesWritten := fixture.plugin.enc.bytesWritten
+	if bytesWritten == 0 {
+		t.Fatal("Expected event to be written into the encoder")
+	}
+
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should not be written into the encoder as rate limit exceeded
+
+	if fixture.plugin.enc.bytesWritten != bytesWritten {
+		t.Fatalf("Expected %v bytes written into the encoder but got %v", bytesWritten, fixture.plugin.enc.bytesWritten)
+	}
+
+	time.Sleep(5 * time.Second)
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should not be written into the encoder as rate limit exceeded
+
+	if fixture.plugin.enc.bytesWritten != bytesWritten {
+		t.Fatalf("Expected %v bytes written into the encoder but got %v", bytesWritten, fixture.plugin.enc.bytesWritten)
+	}
+
+	time.Sleep(5 * time.Second)
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should now be written into the encoder
+
+	if fixture.plugin.buffer.Len() != 1 {
+		t.Fatalf("Expected buffer length of 1 but got %v", fixture.plugin.buffer.Len())
+	}
+
+	exp := EventV1{
+		Labels: map[string]string{
+			"id":      "test-instance-id",
+			"app":     "example-app",
+			"version": version.Version,
+		},
+		DecisionID:  "abc",
+		Path:        "foo/bar",
+		Input:       &input,
+		Result:      &result,
+		RequestedBy: "test-1",
+		Timestamp:   ts,
+	}
+
+	compareLogEvent(t, fixture.plugin.buffer.Pop(), exp)
+
+	chunk, err := fixture.plugin.enc.Flush()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	exp = EventV1{
+		Labels: map[string]string{
+			"id":      "test-instance-id",
+			"app":     "example-app",
+			"version": version.Version,
+		},
+		DecisionID:  "def",
+		Path:        "foo/baz",
+		Input:       &input,
+		Result:      &result,
+		RequestedBy: "test-2",
+		Timestamp:   ts,
+	}
+
+	compareLogEvent(t, chunk, exp)
+}
+
+func TestPluginRateLimitRequeue(t *testing.T) {
+	ctx := context.Background()
+
+	numDecisions := 100 // 100 decisions per second
+
+	fixture := newTestFixture(t, testFixtureOptions{
+		ReportingMaxDecisionsPerSecond: float64(numDecisions),
+		ReportingUploadSizeLimitBytes:  300,
+	})
+	defer fixture.server.stop()
+
+	fixture.server.ch = make(chan []EventV1, 3)
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result1 interface{} = false
+
+	_ = fixture.plugin.Log(ctx, logServerInfo("abc", input, result1)) // event 1
+	_ = fixture.plugin.Log(ctx, logServerInfo("def", input, result1)) // event 2
+	_ = fixture.plugin.Log(ctx, logServerInfo("ghi", input, result1)) // event 3
+
+	bufLen := fixture.plugin.buffer.Len()
+	if bufLen < 2 {
+		t.Fatal("Expected buffer length of at least 2")
+	}
+
+	fixture.server.expCode = 500
+	_, err := fixture.plugin.oneShot(ctx)
+	if err == nil {
+		t.Fatal("Expected error")
+	}
+
+	<-fixture.server.ch
+
+	if fixture.plugin.buffer.Len() < bufLen {
+		t.Fatal("Expected buffer to be preserved")
+	}
+
+	chunk, err := fixture.plugin.enc.Flush()
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	events := decodeLogEvent(t, chunk)
+	if len(events) != 1 {
+		t.Fatalf("Expected 1 event but got %v", len(events))
+	}
+}
+
+func TestPluginRateLimitDropCountStatus(t *testing.T) {
+	ctx := context.Background()
+	testLogger := test.New()
+
+	ts, err := time.Parse(time.RFC3339Nano, "2018-01-01T12:00:00.123456Z")
+	if err != nil {
+		panic(err)
+	}
+
+	numDecisions := 1 // 1 decision per second
+	fixture := newTestFixture(t, testFixtureOptions{
+		ConsoleLogger:                  testLogger,
+		ReportingMaxDecisionsPerSecond: float64(numDecisions),
+		ReportingUploadSizeLimitBytes:  300,
+	})
+	defer fixture.server.stop()
+
+	fixture.plugin.metrics = metrics.New()
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	event1 := &server.Info{
+		DecisionID: "abc",
+		Path:       "foo/bar",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-1",
+		Timestamp:  ts,
+	}
+
+	event2 := &server.Info{
+		DecisionID: "def",
+		Path:       "foo/baz",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-2",
+		Timestamp:  ts,
+	}
+
+	event3 := &server.Info{
+		DecisionID: "ghi",
+		Path:       "foo/aux",
+		Input:      &input,
+		Results:    &result,
+		RemoteAddr: "test-3",
+		Timestamp:  ts,
+	}
+
+	_ = fixture.plugin.Log(ctx, event1) // event 1 should be written into the encoder
+
+	fixture.plugin.mtx.Lock()
+	if fixture.plugin.enc.bytesWritten == 0 {
+		t.Fatal("Expected event to be written into the encoder")
+	}
+	fixture.plugin.mtx.Unlock()
+
+	// Create a status plugin that logs to console
+	pluginConfig := []byte(fmt.Sprintf(`{
+			"console": true,
+		}`))
+
+	config, _ := status.ParseConfig(pluginConfig, fixture.manager.Services(), nil)
+	p := status.New(config, fixture.manager).WithMetrics(fixture.plugin.metrics)
+
+	fixture.manager.Register(status.Name, p)
+	if err := fixture.manager.Start(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	_ = fixture.plugin.Log(ctx, event2) // event 2 should not be written into the encoder as rate limit exceeded
+	_ = fixture.plugin.Log(ctx, event3) // event 3 should not be written into the encoder as rate limit exceeded
+
+	// Trigger a status update
+	status := testStatus()
+	p.UpdateDiscoveryStatus(*status)
+
+	// Give the logger / console some time to process and print the events
+	time.Sleep(10 * time.Millisecond)
+	p.Stop(ctx)
+
+	entries := testLogger.Entries()
+	if len(entries) == 0 {
+		t.Fatal("Expected log entries but got none")
+	}
+
+	// Pick the last entry as it should have the drop count
+	e := entries[len(entries)-1]
+
+	if _, ok := e.Fields["metrics"]; !ok {
+		t.Fatal("Expected metrics")
+	}
+
+	exp := map[string]interface{}{"<built-in>": map[string]interface{}{"counter_decision_logs_dropped": json.Number("2")}}
+
+	if !reflect.DeepEqual(e.Fields["metrics"], exp) {
+		t.Fatalf("Expected %v but got %v", exp, e.Fields["metrics"])
+	}
+}
+
+func TestPluginRateLimitBadConfig(t *testing.T) {
+	manager, _ := plugins.New(nil, "test-instance-id", inmem.New())
+
+	bufSize := 40000
+	numDecisions := 10
+
+	pluginConfig := []byte(fmt.Sprintf(`{
+			"console": true,
+			"reporting": {
+				"buffer_size_limit_bytes": %v,
+				"max_decisions_per_second": %v
+			}
+		}`, bufSize, numDecisions))
+
+	_, err := ParseConfig(pluginConfig, manager.Services(), nil)
+	if err == nil {
+		t.Fatal("Expected error but got nil")
+	}
+
+	expected := "invalid decision_log config, specify either 'buffer_size_limit_bytes' or 'max_decisions_per_second'"
+	if err.Error() != expected {
+		t.Fatalf("Expected error message %v but got %v", expected, err.Error())
+	}
+}
+
+func TestPluginGracefulShutdownFlushesDecisions(t *testing.T) {
+	ctx := context.Background()
+
+	fixture := newTestFixture(t)
+	defer fixture.server.stop()
+
+	fixture.server.ch = make(chan []EventV1, 8)
+
+	if err := fixture.plugin.Start(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	logsSent := 200
+	for i := 0; i < logsSent; i++ {
+		input = generateInputMap(i)
+		_ = fixture.plugin.Log(ctx, logServerInfo("abc", input, result))
+	}
+
+	fixture.server.expCode = 200
+
+	timeoutCtx, cancel := context.WithTimeout(ctx, 10*time.Second)
+	defer cancel()
+	fixture.plugin.Stop(timeoutCtx)
+
+	close(fixture.server.ch)
+	logsReceived := 0
+	for element := range fixture.server.ch {
+		logsReceived += len(element)
+	}
+
+	if logsReceived != logsSent {
+		t.Fatalf("Expected %v, got %v", logsSent, logsReceived)
+	}
+}
+
+func TestPluginTerminatesAfterGracefulShutdownPeriod(t *testing.T) {
+	ctx := context.Background()
+
+	fixture := newTestFixture(t)
+	defer fixture.server.stop()
+
+	fixture.server.ch = make(chan []EventV1, 1)
+	fixture.server.expCode = 500
+
+	if err := fixture.plugin.Start(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	var input interface{} = map[string]interface{}{"method": "GET"}
+	var result interface{} = false
+
+	input = generateInputMap(0)
+	_ = fixture.plugin.Log(ctx, logServerInfo("abc", input, result))
+
+	timeoutCtx, cancel := context.WithTimeout(ctx, 1*time.Nanosecond)
+	defer cancel()
+
+	fixture.plugin.Stop(timeoutCtx)
+
+	// Ensure the plugin was stopped without flushing its whole buffer
+	if fixture.plugin.buffer.Len() == 0 && fixture.plugin.enc.buf.Len() == 0 {
+		t.Errorf("Expected the plugin to still have buffered messages")
+	}
+}
+
+func TestPluginTerminatesAfterGracefulShutdownPeriodWithoutLogs(t *testing.T) {
+	ctx := context.Background()
+
+	fixture := newTestFixture(t)
+	defer fixture.server.stop()
+
+	if err := fixture.plugin.Start(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	timeoutCtx, cancel := context.WithTimeout(ctx, time.Second)
+	defer cancel()
+
+	fixture.plugin.Stop(timeoutCtx)
+	if timeoutCtx.Err() != nil {
+		t.Fatal("Stop did not exit before context expiration")
+	}
+}
+
 func TestPluginReconfigure(t *testing.T) {
 
 	ctx := context.Background()
@@ -532,6 +984,52 @@ func TestPluginReconfigure(t *testing.T) {
 	}
 }
 
+func TestPluginReconfigureUploadSizeLimit(t *testing.T) {
+
+	ctx := context.Background()
+	limit := int64(300)
+
+	fixture := newTestFixture(t, testFixtureOptions{
+		ReportingUploadSizeLimitBytes: limit,
+	})
+	defer fixture.server.stop()
+
+	if err := fixture.plugin.Start(ctx); err != nil {
+		t.Fatal(err)
+	}
+
+	ensurePluginState(t, fixture.plugin, plugins.StateOK)
+
+	fixture.plugin.mtx.Lock()
+	if fixture.plugin.enc.limit != limit {
+		t.Fatalf("Expected upload size limit %v but got %v", limit, fixture.plugin.enc.limit)
+	}
+	fixture.plugin.mtx.Unlock()
+
+	newLimit := int64(600)
+
+	pluginConfig := []byte(fmt.Sprintf(`{
+			"service": "example",
+			"reporting": {
+				"upload_size_limit_bytes": %v,
+			}
+		}`, newLimit))
+
+	config, _ := ParseConfig(pluginConfig, fixture.manager.Services(), nil)
+
+	fixture.plugin.Reconfigure(ctx, config)
+	ensurePluginState(t, fixture.plugin, plugins.StateOK)
+
+	fixture.plugin.Stop(ctx)
+	ensurePluginState(t, fixture.plugin, plugins.StateNotReady)
+
+	fixture.plugin.mtx.Lock()
+	if fixture.plugin.enc.limit != newLimit {
+		t.Fatalf("Expected upload size limit %v but got %v", newLimit, fixture.plugin.enc.limit)
+	}
+	fixture.plugin.mtx.Unlock()
+}
+
 func TestPluginMasking(t *testing.T) {
 	tests := []struct {
 		note        string
@@ -834,13 +1332,20 @@ func TestPluginMasking(t *testing.T) {
 	}
 }
 
-type testFixture struct {
-	manager *plugins.Manager
-	plugin  *Plugin
-	server  *testServer
+type testFixtureOptions struct {
+	ConsoleLogger                  *test.Logger
+	ReportingUploadSizeLimitBytes  int64
+	ReportingMaxDecisionsPerSecond float64
 }
 
-func newTestFixture(t *testing.T, options ...testPluginCustomizer) testFixture {
+type testFixture struct {
+	manager       *plugins.Manager
+	consoleLogger *test.Logger
+	plugin        *Plugin
+	server        *testServer
+}
+
+func newTestFixture(t *testing.T, opts ...testFixtureOptions) testFixture {
 
 	ts := testServer{
 		t:       t,
@@ -866,14 +1371,30 @@ func newTestFixture(t *testing.T, options ...testPluginCustomizer) testFixture {
 				}
 			]}`, ts.server.URL))
 
-	manager, err := plugins.New(managerConfig, "test-instance-id", inmem.New())
+	var options testFixtureOptions
+
+	if len(opts) > 0 {
+		options = opts[0]
+	}
+
+	manager, err := plugins.New(
+		managerConfig,
+		"test-instance-id",
+		inmem.New(),
+		plugins.GracefulShutdownPeriod(10),
+		plugins.ConsoleLogger(options.ConsoleLogger))
 	if err != nil {
 		t.Fatal(err)
 	}
 
 	config, _ := ParseConfig([]byte(`{"service": "example"}`), manager.Services(), nil)
-	for _, option := range options {
-		option(config)
+
+	if options.ReportingMaxDecisionsPerSecond != 0 {
+		config.Reporting.MaxDecisionsPerSecond = &options.ReportingMaxDecisionsPerSecond
+	}
+
+	if options.ReportingUploadSizeLimitBytes != 0 {
+		config.Reporting.UploadSizeLimitBytes = &options.ReportingUploadSizeLimitBytes
 	}
 
 	if s, ok := manager.PluginStatus()[Name]; ok {
@@ -885,9 +1406,10 @@ func newTestFixture(t *testing.T, options ...testPluginCustomizer) testFixture {
 	ensurePluginState(t, p, plugins.StateNotReady)
 
 	return testFixture{
-		manager: manager,
-		plugin:  p,
-		server:  &ts,
+		manager:       manager,
+		consoleLogger: options.ConsoleLogger,
+		plugin:        p,
+		server:        &ts,
 	}
 
 }
@@ -1134,7 +1656,12 @@ func (t *testServer) start() {
 	t.server = httptest.NewServer(http.HandlerFunc(t.handle))
 }
 
+// stop the testServer. This should only be done at the end of a test!
 func (t *testServer) stop() {
+	// Drain any pending events to ensure the server can stop
+	for len(t.ch) > 0 {
+		<-t.ch
+	}
 	t.server.Close()
 }
 
@@ -1183,3 +1710,47 @@ func ensurePluginState(t *testing.T, p *Plugin, state plugins.State) {
 		t.Fatalf("Unexpected status state found in plugin manager for %s:\n\n\tFound:%+v\n\n\tExpected: %s", Name, status.State, plugins.StateOK)
 	}
 }
+
+func decodeLogEvent(t *testing.T, bs []byte) []EventV1 {
+	gr, err := gzip.NewReader(bytes.NewReader(bs))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	var events []EventV1
+	if err := json.NewDecoder(gr).Decode(&events); err != nil {
+		t.Fatal(err)
+	}
+
+	if err := gr.Close(); err != nil {
+		t.Fatal(err)
+	}
+
+	return events
+}
+
+func compareLogEvent(t *testing.T, actual []byte, exp EventV1) {
+	events := decodeLogEvent(t, actual)
+	if len(events) != 1 {
+		t.Fatalf("Expected 1 event but got %v", len(events))
+	}
+
+	if !reflect.DeepEqual(events[0], exp) {
+		t.Fatalf("Expected %+v but got %+v", exp, events[0])
+	}
+}
+
+func testStatus() *bundle.Status {
+
+	tDownload, _ := time.Parse(time.RFC3339Nano, "2018-01-01T00:00:00.0000000Z")
+	tActivate, _ := time.Parse(time.RFC3339Nano, "2018-01-01T00:00:01.0000000Z")
+
+	status := bundle.Status{
+		Name:                     "example/authz",
+		ActiveRevision:           "quickbrawnfaux",
+		LastSuccessfulDownload:   tDownload,
+		LastSuccessfulActivation: tActivate,
+	}
+
+	return &status
+}
