@@ -7,15 +7,21 @@ package plugins
 
 import (
 	"context"
+	"fmt"
 	"sync"
+	"time"
 
 	"github.com/open-policy-agent/opa/ast"
 	"github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/config"
+	bundleUtils "github.com/open-policy-agent/opa/internal/bundle"
 	cfg "github.com/open-policy-agent/opa/internal/config"
 	initload "github.com/open-policy-agent/opa/internal/runtime/init"
+	"github.com/open-policy-agent/opa/keys"
 	"github.com/open-policy-agent/opa/loader"
+	"github.com/open-policy-agent/opa/logging"
 	"github.com/open-policy-agent/opa/plugins/rest"
+	"github.com/open-policy-agent/opa/resolver/wasm"
 	"github.com/open-policy-agent/opa/storage"
 	"github.com/open-policy-agent/opa/topdown/cache"
 )
@@ -106,11 +112,21 @@ const (
 	// StateErr indicates that the Plugin is in an error state and should not
 	// be considered as functional.
 	StateErr State = "ERROR"
+
+	// StateWarn indicates the Plugin is operating, but in a potentially dangerous or
+	// degraded state. It may be used to indicate manual remediation is needed, or to
+	// alert admins of some other noteworthy state.
+	StateWarn State = "WARN"
 )
 
 // Status has a Plugin's current status plus an optional Message.
 type Status struct {
-	State State `json:"state"`
+	State   State  `json:"state"`
+	Message string `json:"message,omitempty"`
+}
+
+func (s *Status) String() string {
+	return fmt.Sprintf("{%v %q}", s.State, s.Message)
 }
 
 // StatusListener defines a handler to register for status updates.
@@ -126,8 +142,10 @@ type Manager struct {
 
 	compiler                     *ast.Compiler
 	compilerMux                  sync.RWMutex
+	wasmResolvers                []*wasm.Resolver
+	wasmResolversMtx             sync.RWMutex
 	services                     map[string]rest.Client
-	keys                         map[string]*bundle.KeyConfig
+	keys                         map[string]*keys.Config
 	plugins                      []namedplugin
 	registeredTriggers           []func(txn storage.Transaction)
 	mtx                          sync.Mutex
@@ -138,11 +156,17 @@ type Manager struct {
 	maxErrors                    int
 	initialized                  bool
 	interQueryBuiltinCacheConfig *cache.Config
+	gracefulShutdownPeriod       int
+	registeredCacheTriggers      []func(*cache.Config)
+	logger                       logging.Logger
+	consoleLogger                logging.Logger
 }
 
 type managerContextKey string
+type managerWasmResolverKey string
 
 const managerCompilerContextKey = managerContextKey("compiler")
+const managerWasmResolverContextKey = managerWasmResolverKey("wasmResolvers")
 
 // SetCompilerOnContext puts the compiler into the storage context. Calling this
 // function before committing updated policies to storage allows the manager to
@@ -161,6 +185,23 @@ func GetCompilerOnContext(context *storage.Context) *ast.Compiler {
 	return compiler
 }
 
+// SetWasmResolversOnContext puts a set of Wasm Resolvers into the storage
+// context. Calling this function before committing updated wasm modules to
+// storage allows the manager to skip initializing modules before using them.
+// Instead, the manager will use the compiler that was stored on the context.
+func SetWasmResolversOnContext(context *storage.Context, rs []*wasm.Resolver) {
+	context.Put(managerWasmResolverContextKey, rs)
+}
+
+// getWasmResolversOnContext gets the resolvers cached on the storage context.
+func getWasmResolversOnContext(context *storage.Context) []*wasm.Resolver {
+	resolvers, ok := context.Get(managerWasmResolverContextKey).([]*wasm.Resolver)
+	if !ok {
+		return nil
+	}
+	return resolvers
+}
+
 type namedplugin struct {
 	name   string
 	plugin Plugin
@@ -195,6 +236,29 @@ func MaxErrors(n int) func(*Manager) {
 	}
 }
 
+// GracefulShutdownPeriod passes the configured graceful shutdown period to plugins
+func GracefulShutdownPeriod(gracefulShutdownPeriod int) func(*Manager) {
+	return func(m *Manager) {
+		m.gracefulShutdownPeriod = gracefulShutdownPeriod
+	}
+}
+
+// Logger configures the passed logger on the plugin manager (useful to
+// configure default fields)
+func Logger(logger logging.Logger) func(*Manager) {
+	return func(m *Manager) {
+		m.logger = logger
+	}
+}
+
+// ConsoleLogger sets the passed logger to be used by plugins that are
+// configured with console logging enabled.
+func ConsoleLogger(logger logging.Logger) func(*Manager) {
+	return func(m *Manager) {
+		m.consoleLogger = logger
+	}
+}
+
 // New creates a new Manager using config.
 func New(raw []byte, id string, store storage.Store, opts ...func(*Manager)) (*Manager, error) {
 
@@ -203,12 +267,7 @@ func New(raw []byte, id string, store storage.Store, opts ...func(*Manager)) (*M
 		return nil, err
 	}
 
-	services, err := cfg.ParseServicesConfig(parsedConfig.Services)
-	if err != nil {
-		return nil, err
-	}
-
-	keys, err := bundle.ParseKeysConfig(parsedConfig.Keys)
+	keys, err := keys.ParseKeysConfig(parsedConfig.Keys)
 	if err != nil {
 		return nil, err
 	}
@@ -223,13 +282,33 @@ func New(raw []byte, id string, store storage.Store, opts ...func(*Manager)) (*M
 		Config:                       parsedConfig,
 		ID:                           id,
 		keys:                         keys,
-		services:                     services,
 		pluginStatus:                 map[string]*Status{},
 		pluginStatusListeners:        map[string]StatusListener{},
 		maxErrors:                    -1,
 		interQueryBuiltinCacheConfig: interQueryBuiltinCacheConfig,
 	}
 
+	if m.logger == nil {
+		m.logger = logging.NewStandardLogger()
+	}
+
+	if m.consoleLogger == nil {
+		m.consoleLogger = logging.NewStandardLogger()
+	}
+
+	serviceOpts := cfg.ServiceOptions{
+		Raw:        parsedConfig.Services,
+		AuthPlugin: m.AuthPlugin,
+		Keys:       keys,
+		Logger:     m.logger,
+	}
+	services, err := cfg.ParseServicesConfig(serviceOpts)
+	if err != nil {
+		return nil, err
+	}
+
+	m.services = services
+
 	for _, f := range opts {
 		f(m)
 	}
@@ -265,6 +344,13 @@ func (m *Manager) Init(ctx context.Context) error {
 		}
 
 		SetCompilerOnContext(params.Context, result.Compiler)
+
+		resolvers, err := bundleUtils.LoadWasmResolversFromStore(ctx, m.Store, txn, nil)
+		if err != nil {
+			return err
+		}
+		SetWasmResolversOnContext(params.Context, resolvers)
+
 		_, err = m.Store.Register(ctx, txn, storage.TriggerConfig{OnCommit: m.onCommit})
 		return err
 	})
@@ -301,7 +387,7 @@ func (m *Manager) Register(name string, plugin Plugin) {
 		plugin: plugin,
 	})
 	if _, ok := m.pluginStatus[name]; !ok {
-		m.pluginStatus[name] = nil
+		m.pluginStatus[name] = &Status{State: StateNotReady}
 	}
 }
 
@@ -328,6 +414,18 @@ func (m *Manager) Plugin(name string) Plugin {
 	return nil
 }
 
+// AuthPlugin returns the HTTPAuthPlugin registered with name or nil if name is not found.
+func (m *Manager) AuthPlugin(name string) rest.HTTPAuthPlugin {
+	m.mtx.Lock()
+	defer m.mtx.Unlock()
+	for i := range m.plugins {
+		if m.plugins[i].name == name {
+			return m.plugins[i].plugin.(rest.HTTPAuthPlugin)
+		}
+	}
+	return nil
+}
+
 // GetCompiler returns the manager's compiler.
 func (m *Manager) GetCompiler() *ast.Compiler {
 	m.compilerMux.RLock()
@@ -349,6 +447,19 @@ func (m *Manager) RegisterCompilerTrigger(f func(txn storage.Transaction)) {
 	m.registeredTriggers = append(m.registeredTriggers, f)
 }
 
+// GetWasmResolvers returns the manager's set of Wasm Resolvers.
+func (m *Manager) GetWasmResolvers() []*wasm.Resolver {
+	m.wasmResolversMtx.RLock()
+	defer m.wasmResolversMtx.RUnlock()
+	return m.wasmResolvers
+}
+
+func (m *Manager) setWasmResolvers(rs []*wasm.Resolver) {
+	m.wasmResolversMtx.Lock()
+	defer m.wasmResolversMtx.Unlock()
+	m.wasmResolvers = rs
+}
+
 // Start starts the manager. Init() should be called once before Start().
 func (m *Manager) Start(ctx context.Context) error {
 
@@ -382,7 +493,8 @@ func (m *Manager) Start(ctx context.Context) error {
 	return nil
 }
 
-// Stop stops the manager, stopping all the plugins registered with it
+// Stop stops the manager, stopping all the plugins registered with it. Any plugin that needs to perform cleanup should
+// do so within the duration of the graceful shutdown period passed with the context as a timeout.
 func (m *Manager) Stop(ctx context.Context) {
 	var toStop []Plugin
 
@@ -395,6 +507,8 @@ func (m *Manager) Stop(ctx context.Context) {
 		}
 	}()
 
+	ctx, cancel := context.WithTimeout(ctx, time.Duration(m.gracefulShutdownPeriod)*time.Second)
+	defer cancel()
 	for i := range toStop {
 		toStop[i].Stop(ctx)
 	}
@@ -402,12 +516,24 @@ func (m *Manager) Stop(ctx context.Context) {
 
 // Reconfigure updates the configuration on the manager.
 func (m *Manager) Reconfigure(config *config.Config) error {
-	services, err := cfg.ParseServicesConfig(config.Services)
+	opts := cfg.ServiceOptions{
+		Raw:        config.Services,
+		AuthPlugin: m.AuthPlugin,
+		Logger:     m.logger,
+	}
+
+	keys, err := keys.ParseKeysConfig(config.Keys)
+	if err != nil {
+		return err
+	}
+	opts.Keys = keys
+
+	services, err := cfg.ParseServicesConfig(opts)
 	if err != nil {
 		return err
 	}
 
-	keys, err := bundle.ParseKeysConfig(config.Keys)
+	interQueryBuiltinCacheConfig, err := cache.ParseCachingConfig(config.Caching)
 	if err != nil {
 		return err
 	}
@@ -416,6 +542,7 @@ func (m *Manager) Reconfigure(config *config.Config) error {
 	defer m.mtx.Unlock()
 	config.Labels = m.Config.Labels // don't overwrite labels
 	m.Config = config
+	m.interQueryBuiltinCacheConfig = interQueryBuiltinCacheConfig
 	for name, client := range services {
 		m.services[name] = client
 	}
@@ -423,6 +550,11 @@ func (m *Manager) Reconfigure(config *config.Config) error {
 	for name, key := range keys {
 		m.keys[name] = key
 	}
+
+	for _, trigger := range m.registeredCacheTriggers {
+		trigger(interQueryBuiltinCacheConfig)
+	}
+
 	return nil
 }
 
@@ -482,7 +614,8 @@ func (m *Manager) copyPluginStatus() map[string]*Status {
 		var cpy *Status
 		if v != nil {
 			cpy = &Status{
-				State: v.State,
+				State:   v.State,
+				Message: v.Message,
 			}
 		}
 		statusCpy[k] = cpy
@@ -508,6 +641,27 @@ func (m *Manager) onCommit(ctx context.Context, txn storage.Transaction, event s
 			f(txn)
 		}
 	}
+
+	// Similar to the compiler, look for a set of resolvers on the transaction
+	// context. If they are not set we may need to reload from the store.
+	resolvers := getWasmResolversOnContext(event.Context)
+	if resolvers != nil {
+		m.setWasmResolvers(resolvers)
+
+	} else if event.DataChanged() {
+		if requiresWasmResolverReload(event) {
+			resolvers, err := bundleUtils.LoadWasmResolversFromStore(ctx, m.Store, txn, nil)
+			if err != nil {
+				panic(err)
+			}
+			m.setWasmResolvers(resolvers)
+		} else {
+			err := m.updateWasmResolversData(ctx, event)
+			if err != nil {
+				panic(err)
+			}
+		}
+	}
 }
 
 func loadCompilerFromStore(ctx context.Context, store storage.Store, txn storage.Transaction) (*ast.Compiler, error) {
@@ -534,8 +688,44 @@ func loadCompilerFromStore(ctx context.Context, store storage.Store, txn storage
 	return compiler, nil
 }
 
+func requiresWasmResolverReload(event storage.TriggerEvent) bool {
+	// If the data changes touched the bundle path (which includes
+	// the wasm modules) we will reload them. Otherwise update
+	// data for each module already on the manager.
+	for _, dataEvent := range event.Data {
+		if dataEvent.Path.HasPrefix(bundle.BundlesBasePath) {
+			return true
+		}
+	}
+	return false
+}
+
+func (m *Manager) updateWasmResolversData(ctx context.Context, event storage.TriggerEvent) error {
+	m.wasmResolversMtx.Lock()
+	defer m.wasmResolversMtx.Unlock()
+
+	if len(m.wasmResolvers) == 0 {
+		return nil
+	}
+
+	for _, resolver := range m.wasmResolvers {
+		for _, dataEvent := range event.Data {
+			var err error
+			if dataEvent.Removed {
+				err = resolver.RemoveDataPath(ctx, dataEvent.Path)
+			} else {
+				err = resolver.SetDataPath(ctx, dataEvent.Path, dataEvent.Data)
+			}
+			if err != nil {
+				return fmt.Errorf("failed to update wasm runtime data: %s", err)
+			}
+		}
+	}
+	return nil
+}
+
 // PublicKeys returns a public keys that can be used for verifying signed bundles.
-func (m *Manager) PublicKeys() map[string]*bundle.KeyConfig {
+func (m *Manager) PublicKeys() map[string]*keys.Config {
 	m.mtx.Lock()
 	defer m.mtx.Unlock()
 	return m.keys
@@ -558,3 +748,21 @@ func (m *Manager) Services() []string {
 	}
 	return s
 }
+
+// Logger gets the standard logger for this plugin manager.
+func (m *Manager) Logger() logging.Logger {
+	return m.logger
+}
+
+// ConsoleLogger gets the console logger for this plugin manager.
+func (m *Manager) ConsoleLogger() logging.Logger {
+	return m.consoleLogger
+}
+
+// RegisterCacheTrigger accepts a func that receives new inter-query cache config generated by
+// a reconfigure of the plugin manager, so that it can be propagated to existing inter-query caches.
+func (m *Manager) RegisterCacheTrigger(trigger func(*cache.Config)) {
+	m.mtx.Lock()
+	defer m.mtx.Unlock()
+	m.registeredCacheTriggers = append(m.registeredCacheTriggers, trigger)
+}
