@@ -2,6 +2,7 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// nolint: goconst // string duplication is for test readability.
 package rego
 
 import (
@@ -10,6 +11,7 @@ import (
 	"log"
 	"net/http"
 	"net/http/httptest"
+	"path/filepath"
 	"reflect"
 	"strconv"
 	"strings"
@@ -185,6 +187,7 @@ func TestRegoInputs(t *testing.T) {
 			r := New(
 				Query("input"),
 				Input(tc.input),
+				Schemas(nil),
 			)
 			assertEval(t, r, tc.expected)
 		})
@@ -857,6 +860,9 @@ func TestPrepareAndPartialResult(t *testing.T) {
 	// as expected for PartialResult.
 
 	partial, err := r.PartialResult(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
 
 	r2 := partial.Rego(
 		Input(map[string]int{"y": 7}),
@@ -918,6 +924,9 @@ func TestPrepareAndPartial(t *testing.T) {
 	// as expected for Partial.
 
 	partialQuery, err := r.Partial(ctx)
+	if err != nil {
+		t.Fatal(err)
+	}
 	expectedQuery := "input.x = 1"
 	if len(partialQuery.Queries) != 1 {
 		t.Errorf("expected 1 query but found %d: %+v", len(partialQuery.Queries), pq)
@@ -956,9 +965,9 @@ func TestPartialNamespace(t *testing.T) {
 	expSupport := ast.MustParseModule(`
 		package foo.test
 
-		p { input.x = 1 }
-
 		default p = false
+
+		p { input.x = 1 }
 	`)
 
 	if len(pq.Support) != 1 || !pq.Support[0].Equal(expSupport) {
@@ -1284,7 +1293,7 @@ func TestUnsafeBuiltins(t *testing.T) {
 	t.Run("unsafe query", func(t *testing.T) {
 		r := New(
 			Query(`count([1, 2, 3])`),
-			UnsafeBuiltins(map[string]struct{}{"count": struct{}{}}),
+			UnsafeBuiltins(map[string]struct{}{"count": {}}),
 		)
 		if _, err := r.Eval(ctx); err == nil || !strings.Contains(err.Error(), unsafeCountExpr) {
 			t.Fatalf("Expected unsafe built-in error but got %v", err)
@@ -1299,7 +1308,7 @@ func TestUnsafeBuiltins(t *testing.T) {
 				count(input.requests) > 10
 			}
 			`),
-			UnsafeBuiltins(map[string]struct{}{"count": struct{}{}}),
+			UnsafeBuiltins(map[string]struct{}{"count": {}}),
 		)
 		if _, err := r.Eval(ctx); err == nil || !strings.Contains(err.Error(), unsafeCountExpr) {
 			t.Fatalf("Expected unsafe built-in error but got %v", err)
@@ -1308,7 +1317,7 @@ func TestUnsafeBuiltins(t *testing.T) {
 
 	t.Run("inherit in query", func(t *testing.T) {
 		r := New(
-			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": struct{}{}})),
+			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": {}})),
 			Query("count([])"),
 		)
 		if _, err := r.Eval(ctx); err == nil || !strings.Contains(err.Error(), unsafeCountExpr) {
@@ -1318,7 +1327,7 @@ func TestUnsafeBuiltins(t *testing.T) {
 
 	t.Run("override/disable in query", func(t *testing.T) {
 		r := New(
-			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": struct{}{}})),
+			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": {}})),
 			UnsafeBuiltins(map[string]struct{}{}),
 			Query("count([])"),
 		)
@@ -1329,8 +1338,8 @@ func TestUnsafeBuiltins(t *testing.T) {
 
 	t.Run("override/change in query", func(t *testing.T) {
 		r := New(
-			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": struct{}{}})),
-			UnsafeBuiltins(map[string]struct{}{"max": struct{}{}}),
+			Compiler(ast.NewCompiler().WithUnsafeBuiltins(map[string]struct{}{"count": {}})),
+			UnsafeBuiltins(map[string]struct{}{"max": {}}),
 			Query("count([]); max([1,2])"),
 		)
 
@@ -1343,7 +1352,7 @@ func TestUnsafeBuiltins(t *testing.T) {
 	t.Run("ignore if given compiler", func(t *testing.T) {
 		r := New(
 			Compiler(ast.NewCompiler()),
-			UnsafeBuiltins(map[string]struct{}{"count": struct{}{}}),
+			UnsafeBuiltins(map[string]struct{}{"count": {}}),
 			Query("data.test.p = 0"),
 			Module("test.rego", `package test
 
@@ -1447,10 +1456,47 @@ func TestRegoEvalWithBundle(t *testing.T) {
 		}
 
 		assertResultSet(t, rs, `[["bar"]]`)
+
+		mods := pq.Modules()
+		if exp, act := 1, len(mods); exp != act {
+			t.Fatalf("expected %d modules, found %d", exp, act)
+		}
+		for act := range mods {
+			if exp := filepath.Join(path, "x/x.rego"); exp != act {
+				t.Errorf("expected module name %q, got %q", exp, act)
+			}
+		}
 	})
 }
 
-func TestRegoEvalPoliciesinStore(t *testing.T) {
+func TestRegoEvalWithBundleURL(t *testing.T) {
+	files := map[string]string{
+		"x/x.rego": "package x\np = data.x.b",
+	}
+
+	test.WithTempFS(files, func(path string) {
+		ctx := context.Background()
+		pq, err := New(
+			LoadBundle("file://"+path),
+			Query("data.x.p"),
+		).PrepareForEval(ctx)
+		if err != nil {
+			t.Fatalf("Unexpected error: %s", err)
+		}
+
+		mods := pq.Modules()
+		if exp, act := 1, len(mods); exp != act {
+			t.Fatalf("expected %d modules, found %d", exp, act)
+		}
+		for act := range mods {
+			if exp := filepath.Join(path, "x/x.rego"); exp != act {
+				t.Errorf("expected module name %q, got %q", exp, act)
+			}
+		}
+	})
+}
+
+func TestRegoEvalPoliciesInStore(t *testing.T) {
 	store := mock.New()
 	ctx := context.Background()
 	txn := storage.NewTransactionOrDie(ctx, store, storage.WriteParams)
@@ -1498,6 +1544,7 @@ func TestRegoEvalModulesOnCompiler(t *testing.T) {
 	pq, err := New(
 		Compiler(compiler),
 		Query("data.a.p"),
+		Schemas(nil),
 	).PrepareForEval(ctx)
 
 	if err != nil {
@@ -1689,8 +1736,8 @@ func TestShallowInliningOption(t *testing.T) {
 	exp := ast.MustParseModule(`
 		package partial.test
 
-		q { 7 = input.x }
 		p { data.partial.test.q = true }
+		q { 7 = input.x }
 	`)
 
 	if len(pq.Support) != 1 || !pq.Support[0].Equal(exp) {
@@ -1698,6 +1745,49 @@ func TestShallowInliningOption(t *testing.T) {
 	}
 }
 
+func TestRegoPartialResultSortedRules(t *testing.T) {
+	r := New(Query("data.test.p"), Module("example.rego", `
+		package test
+
+		default p = false
+
+		p {
+			r = (input.d * input.a) + input.c
+			r < s
+		}
+
+		p {
+			r = (input.d * input.b) + input.c
+			r < s
+		}
+
+		s = 100
+
+	`))
+
+	pq, err := r.Partial(context.Background())
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Without sorting of support rules, the output of the above partial evaluation
+	// resulted in a random order of the support rules (in this case two different possible outputs)
+	exp := ast.MustParseModule(
+		`package partial.test
+
+		default p = false
+
+		p = true { lt(plus(mul(input.d, input.a), input.c), 100) }
+		p = true { lt(plus(mul(input.d, input.b), input.c), 100) }
+		`,
+	)
+
+	if len(pq.Support) != 1 || !pq.Support[0].Equal(exp) {
+		t.Fatal("expected module:", exp, "\n\ngot module:", pq.Support[0])
+	}
+
+}
+
 func TestPrepareWithEmptyModule(t *testing.T) {
 	_, err := New(
 		Query("d"),
@@ -1710,6 +1800,29 @@ func TestPrepareWithEmptyModule(t *testing.T) {
 	}
 }
 
+func TestPrepareWithWasmTargetNotSupported(t *testing.T) {
+	files := map[string]string{
+		"x/x.rego":     "package x\np = data.x.b",
+		"x/data.json":  `{"b": "bar"}`,
+		"/policy.wasm": `modules-compiled-as-wasm-binary`,
+	}
+
+	test.WithTempFS(files, func(path string) {
+		ctx := context.Background()
+
+		_, err := New(
+			LoadBundle(path),
+			Query("data.x.p"),
+			Target("wasm"),
+		).PrepareForEval(ctx)
+
+		expected := "wasm target not supported"
+		if err == nil || err.Error() != expected {
+			t.Fatalf("Expected error %s, got %s", expected, err)
+		}
+	})
+}
+
 func TestEvalWithInterQueryCache(t *testing.T) {
 	query := `http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true})`
 	newHeaders := map[string][]string{"Cache-Control": {"max-age=290304000, public"}}
@@ -1724,7 +1837,7 @@ func TestEvalWithInterQueryCache(t *testing.T) {
 		}
 
 		w.WriteHeader(http.StatusOK)
-		w.Write([]byte(`{"x": 1}`))
+		_, _ = w.Write([]byte(`{"x": 1}`))
 	}))
 	defer ts.Close()
 
@@ -1750,6 +1863,25 @@ func TestEvalWithInterQueryCache(t *testing.T) {
 	}
 }
 
+func TestStrictBuiltinErrors(t *testing.T) {
+	_, err := New(Query("1/0"), StrictBuiltinErrors(true)).Eval(context.Background())
+	if err == nil {
+		t.Fatal("expected error")
+	}
+	topdownErr, ok := err.(*topdown.Error)
+	if !ok {
+		t.Fatal("expected topdown error but got:", err)
+	}
+
+	if topdownErr.Code != topdown.BuiltinErr {
+		t.Fatal("expected builtin error code but got:", topdownErr.Code)
+	}
+
+	if topdownErr.Message != "div: divide by zero" {
+		t.Fatal("expected divide by zero error but got:", topdownErr.Message)
+	}
+}
+
 func TestTimeSeedingOptions(t *testing.T) {
 
 	ctx := context.Background()
@@ -1789,3 +1921,62 @@ func TestTimeSeedingOptions(t *testing.T) {
 func int64ToJSONNumber(i int64) json.Number {
 	return json.Number(strconv.FormatInt(i, 10))
 }
+
+func TestPrepareAndCompileWithSchema(t *testing.T) {
+	module := `
+	package test
+	x = input.y
+	`
+
+	schemaBytes := `{
+		"$schema": "http://json-schema.org/draft-07/schema",
+		"$id": "http://example.com/example.json",
+		"type": "object",
+		"title": "The root schema",
+		"description": "The root schema comprises the entire JSON document.",
+		"required": [],
+		"properties": {
+			"y": {
+				"$id": "#/properties/y",
+				"type": "integer",
+				"title": "The y schema",
+				"description": "An explanation about the purpose of this instance."
+			}
+		},
+		"additionalProperties": false
+	}`
+
+	var schema interface{}
+	err := util.Unmarshal([]byte(schemaBytes), &schema)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	schemaSet := ast.NewSchemaSet()
+	schemaSet.Put(ast.InputRootRef, schema)
+
+	r := New(
+		Query("data.test.x"),
+		Module("", module),
+		Package("foo"),
+		Schemas(schemaSet),
+	)
+
+	ctx := context.Background()
+
+	pq, err := r.PrepareForEval(ctx)
+	if err != nil {
+		t.Fatalf("Unexpected error: %s", err.Error())
+	}
+
+	assertPreparedEvalQueryEval(t, pq, []EvalOption{
+		EvalInput(map[string]int{"y": 1}),
+	}, "[[1]]")
+
+	// Ensure that Compile still works after Prepare
+	// and its Eval has been called.
+	_, err = r.Compile(ctx)
+	if err != nil {
+		t.Errorf("Unexpected error when compiling: %s", err.Error())
+	}
+}
