@@ -5,6 +5,7 @@
 // Package repl implements a Read-Eval-Print-Loop (REPL) for interacting with the policy engine.
 //
 // The REPL is typically used from the command line, however, it can also be used as a library.
+// nolint: goconst // String reuse here doesn't make sense to deduplicate.
 package repl
 
 import (
@@ -17,6 +18,8 @@ import (
 	"strings"
 	"sync"
 
+	"github.com/open-policy-agent/opa/compile"
+
 	"github.com/open-policy-agent/opa/version"
 
 	"github.com/peterh/liner"
@@ -38,12 +41,13 @@ type REPL struct {
 	store   storage.Store
 	runtime *ast.Term
 
-	modules         map[string]*ast.Module
-	currentModuleID string
-	buffer          []string
-	txn             storage.Transaction
-	metrics         metrics.Metrics
-	profiler        bool
+	modules             map[string]*ast.Module
+	currentModuleID     string
+	buffer              []string
+	txn                 storage.Transaction
+	metrics             metrics.Metrics
+	profiler            bool
+	strictBuiltinErrors bool
 
 	// TODO(tsandall): replace this state with rule definitions
 	// inside the default module.
@@ -61,6 +65,7 @@ type REPL struct {
 	errLimit          int
 	prettyLimit       int
 	report            [][2]string
+	target            string // target type (wasm, rego, etc.)
 	mtx               sync.Mutex
 }
 
@@ -68,13 +73,15 @@ type explainMode string
 
 const (
 	explainOff   explainMode = "off"
-	explainFull              = "full"
-	explainNotes             = "notes"
-	explainFails             = "fails"
+	explainFull  explainMode = "full"
+	explainNotes explainMode = "notes"
+	explainFails explainMode = "fails"
 )
 
 const defaultPrettyLimit = 80
 
+var allowedTargets = map[string]bool{compile.TargetRego: true, compile.TargetWasm: true}
+
 const exitPromptMessage = "Do you want to exit ([y]/n)? "
 
 // New returns a new instance of the REPL.
@@ -92,6 +99,7 @@ func New(store storage.Store, historyPath string, output io.Writer, outputFormat
 		banner:       banner,
 		errLimit:     errLimit,
 		prettyLimit:  defaultPrettyLimit,
+		target:       compile.TargetRego,
 	}
 }
 
@@ -134,7 +142,7 @@ func (r *REPL) Loop(ctx context.Context) {
 	line.SetCompleter(r.complete)
 
 loop:
-	for true {
+	for {
 
 		input, err := line.Prompt(r.getPrompt())
 
@@ -169,7 +177,7 @@ loop:
 exitPrompt:
 	fmt.Fprintln(r.output)
 
-	for true {
+	for {
 		input, err := line.Prompt(exitPromptMessage)
 
 		// exit on ctrl+d
@@ -249,6 +257,10 @@ func (r *REPL) OneShot(ctx context.Context, line string) error {
 				return r.cmdTypes()
 			case "unknown":
 				return r.cmdUnknown(cmd.args)
+			case "strict-builtin-errors":
+				return r.cmdStrictBuiltinErrors()
+			case "target":
+				return r.cmdTarget(cmd.args)
 			case "help":
 				return r.cmdHelp(cmd.args)
 			case "exit":
@@ -295,12 +307,6 @@ func (r *REPL) SetOPAVersionReport(report [][2]string) {
 	r.report = report
 }
 
-func (r *REPL) getOPAVersionReport() [][2]string {
-	r.mtx.Lock()
-	defer r.mtx.Unlock()
-	return r.report
-}
-
 func (r *REPL) complete(line string) []string {
 	c := []string{}
 	set := map[string]struct{}{}
@@ -385,6 +391,21 @@ func (r *REPL) cmdFormat(s string) error {
 	return nil
 }
 
+func (r *REPL) cmdTarget(t []string) error {
+	if len(t) != 1 {
+		return newBadArgsErr("target <mode>: expects exactly one argument")
+	}
+
+	if _, ok := allowedTargets[t[0]]; !ok {
+		return fmt.Errorf("invalid target \"%v\":must be one of {rego,wasm}", t[0])
+	}
+
+	r.target = t[0]
+
+	r.checkTraceSupported()
+	return nil
+}
+
 func (r *REPL) cmdPrettyLimit(s []string) error {
 	if len(s) != 1 {
 		return fmt.Errorf("usage: pretty-limit <n>")
@@ -425,10 +446,11 @@ func (r *REPL) cmdShow(args []string) error {
 		return nil
 	} else if strings.Compare(args[0], "debug") == 0 {
 		debug := replDebugState{
-			Explain:    r.explain,
-			Metrics:    r.metricsEnabled(),
-			Instrument: r.instrument,
-			Profile:    r.profilerEnabled(),
+			Explain:             r.explain,
+			Metrics:             r.metricsEnabled(),
+			Instrument:          r.instrument,
+			Profile:             r.profilerEnabled(),
+			StrictBuiltinErrors: r.strictBuiltinErrors,
 		}
 		b, err := json.MarshalIndent(debug, "", "\t")
 		if err != nil {
@@ -442,10 +464,11 @@ func (r *REPL) cmdShow(args []string) error {
 }
 
 type replDebugState struct {
-	Explain    explainMode `json:"explain"`
-	Metrics    bool        `json:"metrics"`
-	Instrument bool        `json:"instrument"`
-	Profile    bool        `json:"profile"`
+	Explain             explainMode `json:"explain"`
+	Metrics             bool        `json:"metrics"`
+	Instrument          bool        `json:"instrument"`
+	Profile             bool        `json:"profile"`
+	StrictBuiltinErrors bool        `json:"strict-builtin-errors"`
 }
 
 func (r *REPL) cmdTrace(mode explainMode) error {
@@ -454,14 +477,19 @@ func (r *REPL) cmdTrace(mode explainMode) error {
 	} else {
 		r.explain = mode
 	}
+
+	r.checkTraceSupported()
 	return nil
 }
 
-func (r *REPL) metricsEnabled() bool {
-	if r.metrics != nil {
-		return true
+func (r *REPL) checkTraceSupported() {
+	if r.explain != explainOff && r.target == compile.TargetWasm {
+		fmt.Fprintf(r.output, "warning: trace mode \"%v\" is not supported with wasm target\n", r.explain)
 	}
-	return false
+}
+
+func (r *REPL) metricsEnabled() bool {
+	return r.metrics != nil
 }
 
 func (r *REPL) cmdMetrics() error {
@@ -498,6 +526,11 @@ func (r *REPL) cmdProfile() error {
 	return nil
 }
 
+func (r *REPL) cmdStrictBuiltinErrors() error {
+	r.strictBuiltinErrors = !r.strictBuiltinErrors
+	return nil
+}
+
 func (r *REPL) cmdTypes() error {
 	r.types = !r.types
 	return nil
@@ -905,6 +938,8 @@ func (r *REPL) evalBody(ctx context.Context, compiler *ast.Compiler, input ast.V
 		rego.Metrics(r.metrics),
 		rego.Instrument(r.instrument),
 		rego.Runtime(r.runtime),
+		rego.StrictBuiltinErrors(r.strictBuiltinErrors),
+		rego.Target(r.target),
 	}
 
 	if r.explain != explainOff {
@@ -970,6 +1005,7 @@ func (r *REPL) evalPartial(ctx context.Context, compiler *ast.Compiler, input as
 		rego.Instrument(r.instrument),
 		rego.ParsedUnknowns(r.unknowns),
 		rego.Runtime(r.runtime),
+		rego.StrictBuiltinErrors(r.strictBuiltinErrors),
 	)
 
 	pq, err := eval.Partial(ctx)
@@ -1097,7 +1133,7 @@ func (r *REPL) getPrompt() string {
 
 func (r *REPL) loadHistory(prompt *liner.State) {
 	if f, err := os.Open(r.historyPath); err == nil {
-		prompt.ReadHistory(f)
+		_, _ = prompt.ReadHistory(f) // ignore error
 		f.Close()
 	}
 }
@@ -1148,7 +1184,7 @@ func (r *REPL) printTypes(ctx context.Context, typeEnv *ast.TypeEnv, body ast.Bo
 
 func (r *REPL) saveHistory(prompt *liner.State) {
 	if f, err := os.Create(r.historyPath); err == nil {
-		prompt.WriteHistory(f)
+		_, _ = prompt.WriteHistory(f) // ignore error
 		f.Close()
 	}
 }
@@ -1199,8 +1235,10 @@ var builtin = [...]commandDesc{
 	{"profile", []string{}, "toggle profiler and turns off trace"},
 	{"types", []string{}, "toggle type information"},
 	{"unknown", []string{"[ref-1 [ref-2 [...]]]"}, "toggle partial evaluation mode"},
+	{"strict-builtin-errors", []string{}, "toggle strict built-in error mode"},
 	{"dump", []string{"[path]"}, "dump raw data in storage"},
 	{"help", []string{"[topic]"}, "print this message"},
+	{"target", []string{"[mode]"}, "set the runtime to exercise {rego,wasm} (default rego)"},
 	{"exit", []string{}, "exit out of shell (or ctrl+d)"},
 	{"ctrl+l", []string{}, "clear the screen"},
 }
