@@ -20,13 +20,14 @@ import (
 	"syscall"
 	"time"
 
-	"github.com/open-policy-agent/opa/bundle"
-
+	"github.com/fsnotify/fsnotify"
+	"github.com/gorilla/mux"
 	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
-	"gopkg.in/fsnotify.v1"
+	"go.uber.org/automaxprocs/maxprocs"
 
 	"github.com/open-policy-agent/opa/ast"
+	"github.com/open-policy-agent/opa/bundle"
 	"github.com/open-policy-agent/opa/internal/config"
 	"github.com/open-policy-agent/opa/internal/prometheus"
 	"github.com/open-policy-agent/opa/internal/report"
@@ -34,6 +35,7 @@ import (
 	initload "github.com/open-policy-agent/opa/internal/runtime/init"
 	"github.com/open-policy-agent/opa/internal/uuid"
 	"github.com/open-policy-agent/opa/loader"
+	"github.com/open-policy-agent/opa/logging"
 	"github.com/open-policy-agent/opa/metrics"
 	"github.com/open-policy-agent/opa/plugins"
 	"github.com/open-policy-agent/opa/plugins/discovery"
@@ -80,10 +82,6 @@ type Params struct {
 	// for read-only diagnostic API's (/health, /metrics, etc)
 	DiagnosticAddrs *[]string
 
-	// InsecureAddr is the listening address that the OPA server will bind to
-	// in addition to Addr if TLS is enabled.
-	InsecureAddr string
-
 	// H2CEnabled flag controls whether OPA will allow H2C (HTTP/2 cleartext) on
 	// HTTP listeners.
 	H2CEnabled bool
@@ -145,6 +143,9 @@ type Params struct {
 	// Logging configures the logging behaviour.
 	Logging LoggingConfig
 
+	// ConsoleLogger sets the logger implementation to use for console logs.
+	ConsoleLogger logging.Logger
+
 	// ConfigFile refers to the OPA configuration to load on startup.
 	ConfigFile string
 
@@ -165,6 +166,9 @@ type Params struct {
 	// server to shutdown gracefully.
 	GracefulShutdownPeriod int
 
+	// ShutdownWaitPeriod is the time (in seconds) to wait before initiating shutdown.
+	ShutdownWaitPeriod int
+
 	// EnableVersionCheck flag controls whether OPA will report its version to an external service.
 	// If this flag is true, OPA will report its version to the external service
 	EnableVersionCheck bool
@@ -179,6 +183,11 @@ type Params struct {
 	// configured bundles and plugins to be activated/ready before listening for traffic.
 	// A value of 0 or less means no wait is exercised.
 	ReadyTimeout int
+
+	// Router is the router to which handlers for the REST API are added.
+	// Router uses a first-matching-route-wins strategy, so no existing routes are overridden
+	// If it is nil, a new mux.Router will be created
+	Router *mux.Router
 }
 
 // LoggingConfig stores the configuration for OPA's logging behaviour.
@@ -247,7 +256,25 @@ func NewRuntime(ctx context.Context, params Params) (*Runtime, error) {
 		return nil, err
 	}
 
-	manager, err := plugins.New(config, params.ID, inmem.New(), plugins.Info(info), plugins.InitBundles(loaded.Bundles), plugins.InitFiles(loaded.Files), plugins.MaxErrors(params.ErrorLimit))
+	var consoleLogger logging.Logger
+
+	if params.ConsoleLogger == nil {
+		stdLogger := logging.NewStandardLogger()
+		stdLogger.SetFormatter(getFormatter(params.Logging.Format))
+		consoleLogger = stdLogger
+	} else {
+		consoleLogger = params.ConsoleLogger
+	}
+
+	manager, err := plugins.New(config,
+		params.ID,
+		inmem.New(),
+		plugins.Info(info),
+		plugins.InitBundles(loaded.Bundles),
+		plugins.InitFiles(loaded.Files),
+		plugins.MaxErrors(params.ErrorLimit),
+		plugins.GracefulShutdownPeriod(params.GracefulShutdownPeriod),
+		plugins.ConsoleLogger(consoleLogger))
 	if err != nil {
 		return nil, errors.Wrap(err, "config error")
 	}
@@ -290,6 +317,7 @@ func (rt *Runtime) StartServer(ctx context.Context) {
 // will block until either: an error occurs, the context is canceled, or
 // a SIGTERM or SIGKILL signal is sent.
 func (rt *Runtime) Serve(ctx context.Context) error {
+
 	if rt.Params.Addrs == nil {
 		return fmt.Errorf("at least one address must be configured in runtime parameters")
 	}
@@ -303,9 +331,25 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 	logrus.WithFields(logrus.Fields{
 		"addrs":            *rt.Params.Addrs,
 		"diagnostic-addrs": *rt.Params.DiagnosticAddrs,
-		"insecure_addr":    rt.Params.InsecureAddr,
 	}).Info("Initializing server.")
 
+	if rt.Params.Authorization == server.AuthorizationOff && rt.Params.Authentication == server.AuthenticationToken {
+		logrus.Error("Token authentication enabled without authorization. Authentication will be ineffective. See https://www.openpolicyagent.org/docs/latest/security/#authentication-and-authorization for more information.")
+	}
+
+	// NOTE(tsandall): at some point, hopefully we can remove this because the
+	// Go runtime will just do the right thing. Until then, try to set
+	// GOMAXPROCS based on the CPU quota applied to the process.
+	undo, err := maxprocs.Set(maxprocs.Logger(func(f string, a ...interface{}) {
+		logrus.Debugf(f, a...)
+	}))
+
+	if err != nil {
+		logrus.WithFields(logrus.Fields{"err": err}).Debug("Failed to set GOMAXPROCS from CPU quota.")
+	}
+
+	defer undo()
+
 	if err := rt.Manager.Start(ctx); err != nil {
 		logrus.WithField("err", err).Error("Failed to start plugins.")
 		return err
@@ -313,14 +357,13 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 
 	defer rt.Manager.Stop(ctx)
 
-	var err error
 	rt.server = server.New().
+		WithRouter(rt.Params.Router).
 		WithStore(rt.Store).
 		WithManager(rt.Manager).
 		WithCompilerErrorLimit(rt.Params.ErrorLimit).
 		WithPprofEnabled(rt.Params.PprofEnabled).
 		WithAddresses(*rt.Params.Addrs).
-		WithInsecureAddress(rt.Params.InsecureAddr).
 		WithH2CEnabled(rt.Params.H2CEnabled).
 		WithCertificate(rt.Params.Certificate).
 		WithCertPool(rt.Params.CertPool).
@@ -383,7 +426,11 @@ func (rt *Runtime) Serve(ctx context.Context) error {
 		}(loop)
 	}
 
-	signalc := make(chan os.Signal)
+	// Buffer one element as os/signal uses non-blocking channel sends.
+	// This prevents potentially dropping the first element and failing to shut
+	// down gracefully. A buffer of 1 is sufficient as we're just looking for a
+	// one-time shutdown signal.
+	signalc := make(chan os.Signal, 1)
 	signal.Notify(signalc, syscall.SIGINT, syscall.SIGTERM)
 
 	rt.serverInitMtx.Lock()
@@ -439,7 +486,8 @@ func (rt *Runtime) StartREPL(ctx context.Context) {
 	defer rt.Manager.Stop(ctx)
 
 	banner := rt.getBanner()
-	repl := repl.New(rt.Store, rt.Params.HistoryPath, rt.Params.Output, rt.Params.OutputFormat, rt.Params.ErrorLimit, banner).WithRuntime(rt.Manager.Info)
+	repl := repl.New(rt.Store, rt.Params.HistoryPath, rt.Params.Output, rt.Params.OutputFormat, rt.Params.ErrorLimit, banner).
+		WithRuntime(rt.Manager.Info)
 
 	if rt.Params.Watch {
 		if err := rt.startWatcher(ctx, rt.Params.Paths, onReloadPrinter(rt.Params.Output)); err != nil {
@@ -530,24 +578,20 @@ func (rt *Runtime) startWatcher(ctx context.Context, paths []string, onReload fu
 }
 
 func (rt *Runtime) readWatcher(ctx context.Context, watcher *fsnotify.Watcher, paths []string, onReload func(time.Duration, error)) {
-	for {
-		select {
-		case evt := <-watcher.Events:
-
-			removalMask := (fsnotify.Remove | fsnotify.Rename)
-			mask := (fsnotify.Create | fsnotify.Write | removalMask)
-			if (evt.Op & mask) != 0 {
-				logrus.WithFields(logrus.Fields{
-					"event": evt.String(),
-				}).Debugf("registered file event")
-				t0 := time.Now()
-				removed := ""
-				if (evt.Op & removalMask) != 0 {
-					removed = evt.Name
-				}
-				err := rt.processWatcherUpdate(ctx, paths, removed)
-				onReload(time.Since(t0), err)
+	for evt := range watcher.Events {
+		removalMask := fsnotify.Remove | fsnotify.Rename
+		mask := fsnotify.Create | fsnotify.Write | removalMask
+		if (evt.Op & mask) != 0 {
+			logrus.WithFields(logrus.Fields{
+				"event": evt.String(),
+			}).Debugf("registered file event")
+			t0 := time.Now()
+			removed := ""
+			if (evt.Op & removalMask) != 0 {
+				removed = evt.Name
 			}
+			err := rt.processWatcherUpdate(ctx, paths, removed)
+			onReload(time.Since(t0), err)
 		}
 	}
 }
@@ -620,6 +664,11 @@ func (rt *Runtime) getBanner() string {
 }
 
 func (rt *Runtime) gracefulServerShutdown(s *server.Server) error {
+	if rt.Params.ShutdownWaitPeriod > 0 {
+		logrus.Infof("Waiting %vs before initiating shutdown...", rt.Params.ShutdownWaitPeriod)
+		time.Sleep(time.Duration(rt.Params.ShutdownWaitPeriod) * time.Second)
+	}
+
 	logrus.Info("Shutting down...")
 	ctx, cancel := context.WithTimeout(context.Background(), time.Duration(rt.Params.GracefulShutdownPeriod)*time.Second)
 	defer cancel()
@@ -708,18 +757,22 @@ func onReloadPrinter(output io.Writer) func(time.Duration, error) {
 	}
 }
 
-func setupLogging(config LoggingConfig) {
-	switch config.Format {
+func getFormatter(format string) logrus.Formatter {
+	switch format {
 	case "text":
-		logrus.SetFormatter(&prettyFormatter{})
+		return &prettyFormatter{}
 	case "json-pretty":
-		logrus.SetFormatter(&logrus.JSONFormatter{PrettyPrint: true})
+		return &logrus.JSONFormatter{PrettyPrint: true}
 	case "json":
 		fallthrough
 	default:
-		logrus.SetFormatter(&logrus.JSONFormatter{})
+		return &logrus.JSONFormatter{}
 	}
+}
 
+func setupLogging(config LoggingConfig) {
+	formatter := getFormatter(config.Format)
+	logrus.SetFormatter(formatter)
 	lvl := logrus.InfoLevel
 
 	if config.Level != "" {
