@@ -26,6 +26,7 @@ import (
 
 	"github.com/gorilla/mux"
 	"github.com/pkg/errors"
+
 	"golang.org/x/net/http2"
 	"golang.org/x/net/http2/h2c"
 
@@ -75,6 +76,7 @@ const (
 	PromHandlerV1Query    = "v1/query"
 	PromHandlerV1Policies = "v1/policies"
 	PromHandlerV1Compile  = "v1/compile"
+	PromHandlerV1Config   = "v1/config"
 	PromHandlerIndex      = "index"
 	PromHandlerCatch      = "catchall"
 	PromHandlerHealth     = "health"
@@ -93,7 +95,6 @@ type Server struct {
 	router                 *mux.Router
 	addrs                  []string
 	diagAddrs              []string
-	insecureAddr           string
 	h2cEnabled             bool
 	authentication         AuthenticationScheme
 	authorization          AuthorizationScheme
@@ -105,8 +106,6 @@ type Server struct {
 	store                  storage.Store
 	manager                *plugins.Manager
 	decisionIDFactory      func() string
-	revisions              map[string]string
-	legacyRevision         string
 	buffer                 Buffer
 	logger                 func(context.Context, *Info) error
 	errLimit               int
@@ -158,16 +157,9 @@ func (s *Server) Init(ctx context.Context) (*Server, error) {
 
 	s.partials = map[string]rego.PartialResult{}
 	s.preparedEvalQueries = newCache(pqMaxCacheSize)
-
-	// Check if there is a bundle revision available at the legacy storage path
-	rev, err := bundle.LegacyReadRevisionFromStore(ctx, s.store, txn)
-	if err == nil && rev != "" {
-		s.legacyRevision = rev
-	}
-
 	s.defaultDecisionPath = s.generateDefaultDecisionPath()
-
 	s.interQueryBuiltinCache = iCache.NewInterQueryCache(s.manager.InterQueryBuiltinCacheConfig())
+	s.manager.RegisterCacheTrigger(s.updateCacheConfig)
 
 	return s, s.store.Commit(ctx, txn)
 }
@@ -214,12 +206,6 @@ func (s *Server) WithDiagnosticAddresses(addrs []string) *Server {
 	return s
 }
 
-// WithInsecureAddress sets the listening address that the server will bind to.
-func (s *Server) WithInsecureAddress(addr string) *Server {
-	s.insecureAddr = addr
-	return s
-}
-
 // WithAuthentication sets authentication scheme to use on the server.
 func (s *Server) WithAuthentication(scheme AuthenticationScheme) *Server {
 	s.authentication = scheme
@@ -339,19 +325,6 @@ func (s *Server) Listeners() ([]Loop, error) {
 		}
 	}
 
-	if s.insecureAddr != "" {
-		parsedURL, err := parseURL(s.insecureAddr, false)
-		if err != nil {
-			return nil, err
-		}
-		loop, httpListener, err := s.getListenerForHTTPServer(parsedURL, s.Handler, defaultListenerType)
-		if err != nil {
-			return nil, err
-		}
-		s.httpListeners = append(s.httpListeners, httpListener)
-		loops = append(loops, loop)
-	}
-
 	return loops, nil
 }
 
@@ -388,8 +361,14 @@ func (ln tcpKeepAliveListener) Accept() (net.Conn, error) {
 	if err != nil {
 		return nil, err
 	}
-	tc.SetKeepAlive(true)
-	tc.SetKeepAlivePeriod(3 * time.Minute)
+	err = tc.SetKeepAlive(true)
+	if err != nil {
+		return nil, err
+	}
+	err = tc.SetKeepAlivePeriod(3 * time.Minute)
+	if err != nil {
+		return nil, err
+	}
 	return tc, nil
 }
 
@@ -640,6 +619,7 @@ func (s *Server) initRouters() {
 	s.registerHandler(mainRouter, 1, "/query", http.MethodGet, s.instrumentHandler(s.v1QueryGet, PromHandlerV1Query))
 	s.registerHandler(mainRouter, 1, "/query", http.MethodPost, s.instrumentHandler(s.v1QueryPost, PromHandlerV1Query))
 	s.registerHandler(mainRouter, 1, "/compile", http.MethodPost, s.instrumentHandler(s.v1CompilePost, PromHandlerV1Compile))
+	s.registerHandler(mainRouter, 1, "/config", http.MethodGet, s.instrumentHandler(s.v1ConfigGet, PromHandlerV1Config))
 	mainRouter.Handle("/", s.instrumentHandler(s.unversionedPost, PromHandlerIndex)).Methods(http.MethodPost)
 	mainRouter.Handle("/", s.instrumentHandler(s.indexGet, PromHandlerIndex)).Methods(http.MethodGet)
 
@@ -678,9 +658,9 @@ func (s *Server) instrumentHandler(handler func(http.ResponseWriter, *http.Reque
 	return http.HandlerFunc(handler)
 }
 
-func (s *Server) execQuery(ctx context.Context, r *http.Request, txn storage.Transaction, decisionID string, parsedQuery ast.Body, input ast.Value, m metrics.Metrics, explainMode types.ExplainModeV1, includeMetrics, includeInstrumentation, pretty bool) (results types.QueryResponseV1, err error) {
+func (s *Server) execQuery(ctx context.Context, r *http.Request, br bundleRevisions, txn storage.Transaction, decisionID string, parsedQuery ast.Body, input ast.Value, m metrics.Metrics, explainMode types.ExplainModeV1, includeMetrics, includeInstrumentation, pretty bool) (results types.QueryResponseV1, err error) {
 
-	logger := s.getDecisionLogger()
+	logger := s.getDecisionLogger(br)
 
 	var buf *topdown.BufferTracer
 	if explainMode != types.ExplainOffV1 {
@@ -696,12 +676,10 @@ func (s *Server) execQuery(ctx context.Context, r *http.Request, txn storage.Tra
 		rawInput = &x
 	}
 
-	compiler := s.getCompiler()
-
-	rego := rego.New(
+	opts := []func(*rego.Rego){
 		rego.Store(s.store),
 		rego.Transaction(txn),
-		rego.Compiler(compiler),
+		rego.Compiler(s.getCompiler()),
 		rego.ParsedQuery(parsedQuery),
 		rego.ParsedInput(input),
 		rego.Metrics(m),
@@ -710,7 +688,15 @@ func (s *Server) execQuery(ctx context.Context, r *http.Request, txn storage.Tra
 		rego.Runtime(s.runtime),
 		rego.UnsafeBuiltins(unsafeBuiltinsMap),
 		rego.InterQueryBuiltinCache(s.interQueryBuiltinCache),
-	)
+	}
+
+	for _, r := range s.manager.GetWasmResolvers() {
+		for _, entrypoint := range r.Entrypoints() {
+			opts = append(opts, rego.Resolver(entrypoint, r))
+		}
+	}
+
+	rego := rego.New(opts...)
 
 	output, err := rego.Eval(ctx)
 	if err != nil {
@@ -780,7 +766,13 @@ func (s *Server) indexGet(w http.ResponseWriter, r *http.Request) {
 
 	defer s.store.Abort(ctx, txn)
 
-	results, err := s.execQuery(ctx, r, txn, decisionID, parsedQuery, input, nil, explainMode, false, false, true)
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
+	results, err := s.execQuery(ctx, r, br, txn, decisionID, parsedQuery, input, nil, explainMode, false, false, true)
 	if err != nil {
 		renderQueryResult(w, nil, err, t0)
 		return
@@ -794,6 +786,40 @@ func (s *Server) registerHandler(router *mux.Router, version int, path string, m
 	router.Handle(prefix+path, h).Methods(method)
 }
 
+type bundleRevisions struct {
+	LegacyRevision string
+	Revisions      map[string]string
+}
+
+func getRevisions(ctx context.Context, store storage.Store, txn storage.Transaction) (bundleRevisions, error) {
+
+	var err error
+	var br bundleRevisions
+	br.Revisions = map[string]string{}
+
+	// Check if we still have a legacy bundle manifest in the store
+	br.LegacyRevision, err = bundle.LegacyReadRevisionFromStore(ctx, store, txn)
+	if err != nil && !storage.IsNotFound(err) {
+		return br, err
+	}
+
+	// read all bundle revisions from storage (if any exist)
+	names, err := bundle.ReadBundleNamesFromStore(ctx, store, txn)
+	if err != nil && !storage.IsNotFound(err) {
+		return br, err
+	}
+
+	for _, name := range names {
+		r, err := bundle.ReadBundleRevisionFromStore(ctx, store, txn, name)
+		if err != nil && !storage.IsNotFound(err) {
+			return br, err
+		}
+		br.Revisions[name] = r
+	}
+
+	return br, nil
+}
+
 func (s *Server) reload(ctx context.Context, txn storage.Transaction, event storage.TriggerEvent) {
 
 	// NOTE(tsandall): We currently rely on the storage txn to provide
@@ -805,29 +831,8 @@ func (s *Server) reload(ctx context.Context, txn storage.Transaction, event stor
 
 	// reset some cached info
 	s.partials = map[string]rego.PartialResult{}
-	s.revisions = map[string]string{}
 	s.preparedEvalQueries = newCache(pqMaxCacheSize)
 	s.defaultDecisionPath = s.generateDefaultDecisionPath()
-
-	// read all bundle revisions from storage (if any exist)
-	names, err := bundle.ReadBundleNamesFromStore(ctx, s.store, txn)
-	if err != nil && !storage.IsNotFound(err) {
-		panic(err)
-	}
-
-	for _, name := range names {
-		r, err := bundle.ReadBundleRevisionFromStore(ctx, s.store, txn, name)
-		if err != nil && !storage.IsNotFound(err) {
-			panic(err)
-		}
-		s.revisions[name] = r
-	}
-
-	// Check if we still have a legacy bundle manifest in the store
-	s.legacyRevision, err = bundle.LegacyReadRevisionFromStore(ctx, s.store, txn)
-	if err != nil && !storage.IsNotFound(err) {
-		panic(err)
-	}
 }
 
 func (s *Server) unversionedPost(w http.ResponseWriter, r *http.Request) {
@@ -870,18 +875,24 @@ func (s *Server) v0QueryPath(w http.ResponseWriter, r *http.Request, urlPath str
 
 	defer s.store.Abort(ctx, txn)
 
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
 	if useDefaultDecisionPath {
 		urlPath = s.defaultDecisionPath
 	}
 
-	logger := s.getDecisionLogger()
+	logger := s.getDecisionLogger(br)
 
 	pqID := "v0QueryPath::" + urlPath
 	preparedQuery, ok := s.getCachedPreparedEvalQuery(pqID, m)
 	if !ok {
 		path := stringPathToDataRef(urlPath)
 
-		rego := rego.New(
+		opts := []func(*rego.Rego){
 			rego.Compiler(s.getCompiler()),
 			rego.Store(s.store),
 			rego.Transaction(txn),
@@ -889,8 +900,17 @@ func (s *Server) v0QueryPath(w http.ResponseWriter, r *http.Request, urlPath str
 			rego.Metrics(m),
 			rego.Runtime(s.runtime),
 			rego.UnsafeBuiltins(unsafeBuiltinsMap),
-		)
-		pq, err := rego.PrepareForEval(ctx)
+		}
+
+		// Set resolvers on the base Rego object to avoid having them get
+		// re-initialized, and to propagate them to the prepared query.
+		for _, r := range s.manager.GetWasmResolvers() {
+			for _, entrypoint := range r.Entrypoints() {
+				opts = append(opts, rego.Resolver(entrypoint, r))
+			}
+		}
+
+		pq, err := rego.New(opts...).PrepareForEval(ctx)
 		if err != nil {
 			_ = logger.Log(ctx, txn, decisionID, r.RemoteAddr, urlPath, "", goInput, input, nil, err, m)
 			writer.ErrorAuto(w, err)
@@ -900,11 +920,16 @@ func (s *Server) v0QueryPath(w http.ResponseWriter, r *http.Request, urlPath str
 		s.preparedEvalQueries.Insert(pqID, preparedQuery)
 	}
 
-	rs, err := preparedQuery.Eval(
-		ctx,
+	evalOpts := []rego.EvalOption{
 		rego.EvalTransaction(txn),
 		rego.EvalParsedInput(input),
 		rego.EvalMetrics(m),
+		rego.EvalInterQueryBuiltinCache(s.interQueryBuiltinCache),
+	}
+
+	rs, err := preparedQuery.Eval(
+		ctx,
+		evalOpts...,
 	)
 
 	m.Timer(metrics.ServerHandler).Stop()
@@ -932,7 +957,8 @@ func (s *Server) v0QueryPath(w http.ResponseWriter, r *http.Request, urlPath str
 		return
 	}
 
-	writer.JSON(w, 200, rs[0].Expressions[0].Value, false)
+	pretty := getBoolParam(r.URL, types.ParamPrettyV1, true)
+	writer.JSON(w, 200, rs[0].Expressions[0].Value, pretty)
 }
 
 func (s *Server) getCachedPreparedEvalQuery(key string, m metrics.Metrics) (*rego.PreparedEvalQuery, bool) {
@@ -947,8 +973,19 @@ func (s *Server) getCachedPreparedEvalQuery(key string, m metrics.Metrics) (*reg
 
 func (s *Server) canEval(ctx context.Context) bool {
 	// Create very simple query that binds a single variable.
-	eval := rego.New(rego.Compiler(s.getCompiler()),
-		rego.Store(s.store), rego.Query("x = 1"))
+	opts := []func(*rego.Rego){
+		rego.Compiler(s.getCompiler()),
+		rego.Store(s.store),
+		rego.Query("x = 1"),
+	}
+
+	for _, r := range s.manager.GetWasmResolvers() {
+		for _, ep := range r.Entrypoints() {
+			opts = append(opts, rego.Resolver(ep, r))
+		}
+	}
+
+	eval := rego.New(opts...)
 	// Run evaluation.
 	rs, err := eval.Eval(ctx)
 	if err != nil {
@@ -1043,7 +1080,7 @@ func (s *Server) v1CompilePost(w http.ResponseWriter, r *http.Request) {
 	includeInstrumentation := getBoolParam(r.URL, types.ParamInstrumentV1, true)
 
 	m := metrics.New()
-
+	m.Timer(metrics.ServerHandler).Start()
 	m.Timer(metrics.RegoQueryParse).Start()
 
 	request, reqErr := readInputCompilePostV1(r.Body)
@@ -1093,6 +1130,8 @@ func (s *Server) v1CompilePost(w http.ResponseWriter, r *http.Request) {
 		return
 	}
 
+	m.Timer(metrics.ServerHandler).Stop()
+
 	result := types.CompileResponseV1{}
 
 	if includeMetrics || includeInstrumentation {
@@ -1128,6 +1167,7 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 	includeMetrics := getBoolParam(r.URL, types.ParamMetricsV1, true)
 	includeInstrumentation := getBoolParam(r.URL, types.ParamInstrumentV1, true)
 	provenance := getBoolParam(r.URL, types.ParamProvenanceV1, true)
+	strictBuiltinErrors := getBoolParam(r.URL, types.ParamStrictBuiltinErrors, true)
 
 	m.Timer(metrics.RegoInputParse).Start()
 
@@ -1164,7 +1204,13 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 	}
 	defer s.store.Abort(ctx, txn)
 
-	logger := s.getDecisionLogger()
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
+	logger := s.getDecisionLogger(br)
 
 	var buf *topdown.BufferTracer
 
@@ -1172,10 +1218,14 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 		buf = topdown.NewBufferTracer()
 	}
 
-	pqID := "v1DataGet::" + urlPath
+	pqID := "v1DataGet::"
+	if strictBuiltinErrors {
+		pqID += "strict-builtin-errors::"
+	}
+	pqID += urlPath
 	preparedQuery, ok := s.getCachedPreparedEvalQuery(pqID, m)
 	if !ok {
-		rego := rego.New(
+		opts := []func(*rego.Rego){
 			rego.Compiler(s.getCompiler()),
 			rego.Store(s.store),
 			rego.Transaction(txn),
@@ -1186,9 +1236,16 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 			rego.Instrument(includeInstrumentation),
 			rego.Runtime(s.runtime),
 			rego.UnsafeBuiltins(unsafeBuiltinsMap),
-		)
+			rego.StrictBuiltinErrors(strictBuiltinErrors),
+		}
 
-		pq, err := rego.PrepareForEval(ctx)
+		for _, r := range s.manager.GetWasmResolvers() {
+			for _, entrypoint := range r.Entrypoints() {
+				opts = append(opts, rego.Resolver(entrypoint, r))
+			}
+		}
+
+		pq, err := rego.New(opts...).PrepareForEval(ctx)
 		if err != nil {
 			_ = logger.Log(ctx, txn, decisionID, r.RemoteAddr, urlPath, "", goInput, input, nil, err, m)
 			writer.ErrorAuto(w, err)
@@ -1198,13 +1255,18 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 		s.preparedEvalQueries.Insert(pqID, preparedQuery)
 	}
 
-	rs, err := preparedQuery.Eval(
-		ctx,
+	evalOpts := []rego.EvalOption{
 		rego.EvalTransaction(txn),
 		rego.EvalParsedInput(input),
 		rego.EvalMetrics(m),
 		rego.EvalQueryTracer(buf),
 		rego.EvalInterQueryBuiltinCache(s.interQueryBuiltinCache),
+		rego.EvalInstrument(includeInstrumentation),
+	}
+
+	rs, err := preparedQuery.Eval(
+		ctx,
+		evalOpts...,
 	)
 
 	m.Timer(metrics.ServerHandler).Stop()
@@ -1225,7 +1287,7 @@ func (s *Server) v1DataGet(w http.ResponseWriter, r *http.Request) {
 	}
 
 	if provenance {
-		result.Provenance = s.getProvenance()
+		result.Provenance = s.getProvenance(br)
 	}
 
 	if len(rs) == 0 {
@@ -1321,6 +1383,7 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 	includeInstrumentation := getBoolParam(r.URL, types.ParamInstrumentV1, true)
 	partial := getBoolParam(r.URL, types.ParamPartialV1, true)
 	provenance := getBoolParam(r.URL, types.ParamProvenanceV1, true)
+	strictBuiltinErrors := getBoolParam(r.URL, types.ParamStrictBuiltinErrors, true)
 
 	m.Timer(metrics.RegoInputParse).Start()
 
@@ -1347,15 +1410,17 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 		writer.ErrorAuto(w, err)
 		return
 	}
+
 	defer s.store.Abort(ctx, txn)
 
-	logger := s.getDecisionLogger()
-
-	opts := []func(*rego.Rego){
-		rego.Compiler(s.getCompiler()),
-		rego.Store(s.store),
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
 	}
 
+	logger := s.getDecisionLogger(br)
+
 	var buf *topdown.BufferTracer
 
 	if explainMode != types.ExplainOffV1 {
@@ -1366,9 +1431,26 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 	if partial {
 		pqID += "partial::"
 	}
+	if strictBuiltinErrors {
+		pqID += "strict-builtin-errors::"
+	}
 	pqID += urlPath
 	preparedQuery, ok := s.getCachedPreparedEvalQuery(pqID, m)
 	if !ok {
+		opts := []func(*rego.Rego){
+			rego.Compiler(s.getCompiler()),
+			rego.Store(s.store),
+			rego.StrictBuiltinErrors(strictBuiltinErrors),
+		}
+
+		// Set resolvers on the base Rego object to avoid having them get
+		// re-initialized, and to propagate them to the prepared query.
+		for _, r := range s.manager.GetWasmResolvers() {
+			for _, entrypoint := range r.Entrypoints() {
+				opts = append(opts, rego.Resolver(entrypoint, r))
+			}
+		}
+
 		rego, err := s.makeRego(ctx, partial, txn, input, urlPath, m, includeInstrumentation, buf, opts)
 
 		if err != nil {
@@ -1387,13 +1469,18 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 		s.preparedEvalQueries.Insert(pqID, preparedQuery)
 	}
 
-	rs, err := preparedQuery.Eval(
-		ctx,
+	evalOpts := []rego.EvalOption{
 		rego.EvalTransaction(txn),
 		rego.EvalParsedInput(input),
 		rego.EvalMetrics(m),
 		rego.EvalQueryTracer(buf),
 		rego.EvalInterQueryBuiltinCache(s.interQueryBuiltinCache),
+		rego.EvalInstrument(includeInstrumentation),
+	}
+
+	rs, err := preparedQuery.Eval(
+		ctx,
+		evalOpts...,
 	)
 
 	m.Timer(metrics.ServerHandler).Stop()
@@ -1414,7 +1501,7 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 	}
 
 	if provenance {
-		result.Provenance = s.getProvenance()
+		result.Provenance = s.getProvenance(br)
 	}
 
 	if len(rs) == 0 {
@@ -1422,6 +1509,7 @@ func (s *Server) v1DataPost(w http.ResponseWriter, r *http.Request) {
 			result.Explanation, err = types.NewTraceV1(*buf, pretty)
 			if err != nil {
 				writer.ErrorAuto(w, err)
+				return
 			}
 		}
 		err = logger.Log(ctx, txn, decisionID, r.RemoteAddr, urlPath, "", goInput, input, nil, nil, m)
@@ -1553,7 +1641,13 @@ func (s *Server) v1PoliciesDelete(w http.ResponseWriter, r *http.Request) {
 	vars := mux.Vars(r)
 	pretty := getBoolParam(r.URL, types.ParamPrettyV1, true)
 	includeMetrics := getBoolParam(r.URL, types.ParamPrettyV1, true)
-	id := vars["path"]
+
+	id, err := url.PathUnescape(vars["path"])
+	if err != nil {
+		writer.ErrorString(w, http.StatusBadRequest, types.CodeInvalidParameter, err)
+		return
+	}
+
 	m := metrics.New()
 
 	txn, err := s.store.NewTransaction(ctx, storage.WriteParams)
@@ -1610,7 +1704,13 @@ func (s *Server) v1PoliciesDelete(w http.ResponseWriter, r *http.Request) {
 func (s *Server) v1PoliciesGet(w http.ResponseWriter, r *http.Request) {
 	ctx := r.Context()
 	vars := mux.Vars(r)
-	path := vars["path"]
+
+	path, err := url.PathUnescape(vars["path"])
+	if err != nil {
+		writer.ErrorString(w, http.StatusBadRequest, types.CodeInvalidParameter, err)
+		return
+	}
+
 	pretty := getBoolParam(r.URL, types.ParamPrettyV1, true)
 
 	txn, err := s.store.NewTransaction(ctx)
@@ -1687,7 +1787,13 @@ func (s *Server) v1PoliciesList(w http.ResponseWriter, r *http.Request) {
 func (s *Server) v1PoliciesPut(w http.ResponseWriter, r *http.Request) {
 	ctx := r.Context()
 	vars := mux.Vars(r)
-	path := vars["path"]
+
+	path, err := url.PathUnescape(vars["path"])
+	if err != nil {
+		writer.ErrorString(w, http.StatusBadRequest, types.CodeInvalidParameter, err)
+		return
+	}
+
 	includeMetrics := getBoolParam(r.URL, types.ParamMetricsV1, true)
 	pretty := getBoolParam(r.URL, types.ParamPrettyV1, true)
 	m := metrics.New()
@@ -1829,7 +1935,13 @@ func (s *Server) v1QueryGet(w http.ResponseWriter, r *http.Request) {
 
 	defer s.store.Abort(ctx, txn)
 
-	results, err := s.execQuery(ctx, r, txn, decisionID, parsedQuery, nil, m, explainMode, includeMetrics, includeInstrumentation, pretty)
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
+	results, err := s.execQuery(ctx, r, br, txn, decisionID, parsedQuery, nil, m, explainMode, includeMetrics, includeInstrumentation, pretty)
 	if err != nil {
 		switch err := err.(type) {
 		case ast.Errors:
@@ -1882,7 +1994,13 @@ func (s *Server) v1QueryPost(w http.ResponseWriter, r *http.Request) {
 
 	defer s.store.Abort(ctx, txn)
 
-	results, err := s.execQuery(ctx, r, txn, decisionID, parsedQuery, nil, m, explainMode, includeMetrics, includeInstrumentation, pretty)
+	br, err := getRevisions(ctx, s.store, txn)
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
+	results, err := s.execQuery(ctx, r, br, txn, decisionID, parsedQuery, nil, m, explainMode, includeMetrics, includeInstrumentation, pretty)
 	if err != nil {
 		switch err := err.(type) {
 		case ast.Errors:
@@ -1896,6 +2014,20 @@ func (s *Server) v1QueryPost(w http.ResponseWriter, r *http.Request) {
 	writer.JSON(w, 200, results, pretty)
 }
 
+func (s *Server) v1ConfigGet(w http.ResponseWriter, r *http.Request) {
+	pretty := getBoolParam(r.URL, types.ParamPrettyV1, true)
+	result, err := s.manager.Config.ActiveConfig()
+	if err != nil {
+		writer.ErrorAuto(w, err)
+		return
+	}
+
+	var resp types.ConfigResponseV1
+	resp.Result = &result
+
+	writer.JSON(w, http.StatusOK, resp, pretty)
+}
+
 func (s *Server) checkPolicyIDScope(ctx context.Context, txn storage.Transaction, id string) error {
 
 	bs, err := s.store.GetPolicy(ctx, txn, id)
@@ -1947,9 +2079,15 @@ func (s *Server) checkPathScope(ctx context.Context, txn storage.Transaction, pa
 
 	spath := strings.Trim(path.String(), "/")
 
+	if spath == "" && len(bundleRoots) > 0 {
+		return types.BadRequestErr("can't write to document root with bundle roots configured")
+	}
+
+	spathParts := strings.Split(spath, "/")
+
 	for name, roots := range bundleRoots {
 		for _, root := range roots {
-			if strings.HasPrefix(spath, root) || strings.HasPrefix(root, spath) {
+			if isPathOwned(spathParts, strings.Split(root, "/")) {
 				return types.BadRequestErr(fmt.Sprintf("path %v is owned by bundle %q", spath, name))
 			}
 		}
@@ -1958,12 +2096,12 @@ func (s *Server) checkPathScope(ctx context.Context, txn storage.Transaction, pa
 	return nil
 }
 
-func (s *Server) getDecisionLogger() (logger decisionLogger) {
+func (s *Server) getDecisionLogger(br bundleRevisions) (logger decisionLogger) {
 	// For backwards compatibility use `revision` as needed.
-	if s.hasLegacyBundle() {
-		logger.revision = s.legacyRevision
+	if s.hasLegacyBundle(br) {
+		logger.revision = br.LegacyRevision
 	} else {
-		logger.revisions = s.revisions
+		logger.revisions = br.Revisions
 	}
 	logger.logger = s.logger
 	logger.buffer = s.buffer
@@ -2127,7 +2265,7 @@ func (s *Server) generateDecisionID() string {
 	return ""
 }
 
-func (s *Server) getProvenance() *types.ProvenanceV1 {
+func (s *Server) getProvenance(br bundleRevisions) *types.ProvenanceV1 {
 
 	p := &types.ProvenanceV1{
 		Version:   version.Version,
@@ -2139,11 +2277,11 @@ func (s *Server) getProvenance() *types.ProvenanceV1 {
 	// For backwards compatibility, if the bundles are using the old
 	// style config we need to fill in the older `Revision` field.
 	// Otherwise use the newer `Bundles` keyword.
-	if s.hasLegacyBundle() {
-		p.Revision = s.legacyRevision
+	if s.hasLegacyBundle(br) {
+		p.Revision = br.LegacyRevision
 	} else {
 		p.Bundles = map[string]types.ProvenanceBundleV1{}
-		for name, revision := range s.revisions {
+		for name, revision := range br.Revisions {
 			p.Bundles[name] = types.ProvenanceBundleV1{Revision: revision}
 		}
 	}
@@ -2151,9 +2289,9 @@ func (s *Server) getProvenance() *types.ProvenanceV1 {
 	return p
 }
 
-func (s *Server) hasLegacyBundle() bool {
+func (s *Server) hasLegacyBundle(br bundleRevisions) bool {
 	bp := bundlePlugin.Lookup(s.manager)
-	return s.legacyRevision != "" || (bp != nil && !bp.Config().IsMultiBundle())
+	return br.LegacyRevision != "" || (bp != nil && !bp.Config().IsMultiBundle())
 }
 
 func (s *Server) generateDefaultDecisionPath() string {
@@ -2162,6 +2300,19 @@ func (s *Server) generateDefaultDecisionPath() string {
 	return p
 }
 
+func isPathOwned(path, root []string) bool {
+	for i := 0; i < len(path) && i < len(root); i++ {
+		if path[i] != root[i] {
+			return false
+		}
+	}
+	return true
+}
+
+func (s *Server) updateCacheConfig(cacheConfig *iCache.Config) {
+	s.interQueryBuiltinCache.UpdateConfig(cacheConfig)
+}
+
 // parsePatchPathEscaped returns a new path for the given escaped str.
 // This is based on storage.ParsePathEscaped so will do URL unescaping of
 // the provided str for backwards compatibility, but also handles the
@@ -2268,14 +2419,22 @@ func getExplain(p []string, zero types.ExplainModeV1) types.ExplainModeV1 {
 }
 
 func readInputV0(r *http.Request) (ast.Value, error) {
+
+	parsed, ok := authorizer.GetBodyOnContext(r.Context())
+	if ok {
+		return ast.InterfaceToValue(parsed)
+	}
+
 	bs, err := ioutil.ReadAll(r.Body)
 	if err != nil {
 		return nil, err
 	}
+
 	bs = bytes.TrimSpace(bs)
 	if len(bs) == 0 {
 		return nil, nil
 	}
+
 	var x interface{}
 
 	if strings.Contains(r.Header.Get("Content-Type"), "yaml") {
@@ -2299,6 +2458,16 @@ func readInputGetV1(str string) (ast.Value, error) {
 
 func readInputPostV1(r *http.Request) (ast.Value, error) {
 
+	parsed, ok := authorizer.GetBodyOnContext(r.Context())
+	if ok {
+		if obj, ok := parsed.(map[string]interface{}); ok {
+			if input, ok := obj["input"]; ok {
+				return ast.InterfaceToValue(input)
+			}
+		}
+		return nil, nil
+	}
+
 	bs, err := ioutil.ReadAll(r.Body)
 
 	if err != nil {
