@@ -2,6 +2,7 @@
 // Use of this source code is governed by an Apache2
 // license that can be found in the LICENSE file.
 
+// nolint: goconst // string duplication is for test readability.
 package topdown
 
 import (
@@ -17,6 +18,7 @@ import (
 	"net/http"
 	"net/http/httptest"
 	"os"
+	"reflect"
 	"strconv"
 	"strings"
 	"testing"
@@ -48,7 +50,7 @@ func TestHTTPGetRequest(t *testing.T) {
 		headers := w.Header()
 		headers["test-header"] = []string{"test-value"}
 		w.WriteHeader(http.StatusOK)
-		json.NewEncoder(w).Encode(people)
+		_ = json.NewEncoder(w).Encode(people)
 	}))
 
 	defer ts.Close()
@@ -104,7 +106,7 @@ func TestHTTPGetRequestTlsInsecureSkipVerify(t *testing.T) {
 	// test server
 	ts := httptest.NewTLSServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		w.WriteHeader(http.StatusOK)
-		json.NewEncoder(w).Encode(people)
+		_ = json.NewEncoder(w).Encode(people)
 	}))
 	defer ts.Close()
 
@@ -205,8 +207,7 @@ func echoCustomHeaders(w http.ResponseWriter, r *http.Request) {
 			headers[k] = v
 		}
 	}
-	json.NewEncoder(w).Encode(headers)
-	return
+	_ = json.NewEncoder(w).Encode(headers)
 }
 
 // TestHTTPSendCustomRequestHeaders adds custom headers to request
@@ -268,7 +269,7 @@ func TestHTTPHostHeader(t *testing.T) {
 	// test server
 	ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 		w.Header().Set("Content-Type", "application/json")
-		json.NewEncoder(w).Encode(r.Host)
+		_ = json.NewEncoder(w).Encode(r.Host)
 	}))
 
 	defer ts.Close()
@@ -314,7 +315,10 @@ func TestHTTPPostRequest(t *testing.T) {
 
 		w.Header().Set("Content-Type", contentType)
 		w.WriteHeader(http.StatusOK)
-		w.Write(bs)
+		_, err = w.Write(bs)
+		if err != nil {
+			t.Fatal(err)
+		}
 	}))
 
 	defer ts.Close()
@@ -433,7 +437,7 @@ func TestHTTPDeleteRequest(t *testing.T) {
 
 		w.Header().Set("Content-Type", "application/json")
 		w.WriteHeader(http.StatusOK)
-		json.NewEncoder(w).Encode(people)
+		_ = json.NewEncoder(w).Encode(people)
 	}))
 
 	defer ts.Close()
@@ -461,7 +465,7 @@ func TestHTTPDeleteRequest(t *testing.T) {
 	// delete a new person
 	personToDelete := Person{ID: "2", Firstname: "Joe"}
 	b := new(bytes.Buffer)
-	json.NewEncoder(b).Encode(personToDelete)
+	_ = json.NewEncoder(b).Encode(personToDelete)
 
 	// run the test
 	tests := []struct {
@@ -752,11 +756,6 @@ func TestHTTPSendRaiseError(t *testing.T) {
 }
 
 func TestHTTPSendCaching(t *testing.T) {
-	// expected result
-	var body []interface{}
-	bodyMap := map[string]string{"id": "1", "firstname": "John"}
-	body = append(body, bodyMap)
-
 	// run the test
 	tests := []struct {
 		note             string
@@ -857,7 +856,10 @@ func TestHTTPSendCaching(t *testing.T) {
 			ts := httptest.NewServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
 				requests = append(requests, r)
 				w.WriteHeader(http.StatusOK)
-				w.Write([]byte(tc.response))
+				_, err := w.Write([]byte(tc.response))
+				if err != nil {
+					t.Fatal(err)
+				}
 			}))
 			defer ts.Close()
 
@@ -917,6 +919,20 @@ func TestHTTPSendInterQueryCaching(t *testing.T) {
 			response:         `{"x": 1}`,
 			expectedReqCount: 1,
 		},
+		{
+			note: "http.send GET (expires_header_invalid_value)",
+			ruleTemplate: `p = x {
+									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true})
+									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true})  # not cached
+									r3 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true})  # not cached
+									r1 == r2
+									r2 == r3
+									x = r1.body
+								}`,
+			headers:          map[string][]string{"Expires": {"0"}},
+			response:         `{"x": 1}`,
+			expectedReqCount: 3,
+		},
 		{
 			note: "http.send GET no-store cache",
 			ruleTemplate: `p = x {
@@ -977,6 +993,9 @@ func TestHTTPSendInterQueryCaching(t *testing.T) {
 
 	data := loadSmallTestData()
 
+	t0 := time.Now()
+	opts := setTime(t0)
+
 	for _, tc := range tests {
 		t.Run(tc.note, func(t *testing.T) {
 
@@ -989,6 +1008,8 @@ func TestHTTPSendInterQueryCaching(t *testing.T) {
 					headers[k] = v
 				}
 
+				headers.Set("Date", t0.Format(time.RFC850))
+
 				etag := w.Header().Get("etag")
 				lm := w.Header().Get("last-modified")
 
@@ -1003,11 +1024,11 @@ func TestHTTPSendInterQueryCaching(t *testing.T) {
 				} else {
 					w.WriteHeader(http.StatusOK)
 				}
-				w.Write([]byte(tc.response))
+				_, _ = w.Write([]byte(tc.response)) // ignore error
 			}))
 			defer ts.Close()
 
-			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response)
+			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response, opts)
 
 			// Note: The runTopDownTestCase ends up evaluating twice (once with and once without partial
 			// eval first), so expect 2x the total request count the test case specified.
@@ -1056,10 +1077,64 @@ func TestHTTPSendInterQueryForceCaching(t *testing.T) {
 			response:         `{"x": 1}`,
 			expectedReqCount: 1,
 		},
+		{
+			note: "http.send GET cache hit (force_cache_only_no_store_override)",
+			ruleTemplate: `p = x {
+									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "force_cache": true, "force_cache_duration_seconds": 300})
+									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r3 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r1 == r2
+									r2 == r3
+									x = r1.body
+								}`,
+			headers: map[string][]string{
+				"Expires":       {"Wed, 31 Dec 2005 07:28:00 GMT"},
+				"Cache-Control": {"no-store"},
+			},
+			response:         `{"x": 1}`,
+			expectedReqCount: 1,
+		},
+		{
+			note: "http.send GET cache hit (cache_param_override_no_store_override)",
+			ruleTemplate: `p = x {
+									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})
+									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r3 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r1 == r2
+									r2 == r3
+									x = r1.body
+								}`,
+			headers: map[string][]string{
+				"Expires":       {"Wed, 31 Dec 2005 07:28:00 GMT"},
+				"Cache-Control": {"no-store", "no-cache", "max-age=0"},
+			},
+			response:         `{"x": 1}`,
+			expectedReqCount: 1,
+		},
+		{
+			note: "http.send GET cache hit (cache_param_override_no_store_override_invalid_expires_header_value)",
+			ruleTemplate: `p = x {
+									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})
+									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r3 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true, "force_cache": true, "force_cache_duration_seconds": 300})  # cached and fresh
+									r1 == r2
+									r2 == r3
+									x = r1.body
+								}`,
+			headers: map[string][]string{
+				"Expires":       {"0"},
+				"Cache-Control": {"no-store", "no-cache", "max-age=0"},
+			},
+			response:         `{"x": 1}`,
+			expectedReqCount: 1,
+		},
 	}
 
 	data := loadSmallTestData()
 
+	t0 := time.Now()
+	opts := setTime(t0)
+
 	for _, tc := range tests {
 		t.Run(tc.note, func(t *testing.T) {
 
@@ -1072,12 +1147,17 @@ func TestHTTPSendInterQueryForceCaching(t *testing.T) {
 					headers[k] = v
 				}
 
+				headers.Set("Date", t0.Format(time.RFC850))
+
 				w.WriteHeader(http.StatusOK)
-				w.Write([]byte(tc.response))
+				_, err := w.Write([]byte(tc.response))
+				if err != nil {
+					t.Fatal(err)
+				}
 			}))
 			defer ts.Close()
 
-			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response)
+			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response, opts)
 
 			// Note: The runTopDownTestCase ends up evaluating twice (once with and once without partial
 			// eval first), so expect 2x the total request count the test case specified.
@@ -1104,7 +1184,6 @@ func TestHTTPSendInterQueryCachingModifiedResp(t *testing.T) {
 									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true})
 									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true}) # stale
 									r3 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true}) # cached and fresh
-									r1 == r2
 									r2 == r3
 									x = r1.body
 								}`,
@@ -1128,6 +1207,9 @@ func TestHTTPSendInterQueryCachingModifiedResp(t *testing.T) {
 
 	data := loadSmallTestData()
 
+	t0 := time.Now()
+	opts := setTime(t0)
+
 	for _, tc := range tests {
 		t.Run(tc.note, func(t *testing.T) {
 
@@ -1140,6 +1222,8 @@ func TestHTTPSendInterQueryCachingModifiedResp(t *testing.T) {
 					headers[k] = v
 				}
 
+				headers.Set("Date", t0.Format(time.RFC850))
+
 				etag := w.Header().Get("etag")
 
 				if r.Header.Get("if-none-match") != "" {
@@ -1152,11 +1236,11 @@ func TestHTTPSendInterQueryCachingModifiedResp(t *testing.T) {
 				} else {
 					w.WriteHeader(http.StatusOK)
 				}
-				w.Write([]byte(tc.response))
+				_, _ = w.Write([]byte(tc.response)) // ignore error
 			}))
 			defer ts.Close()
 
-			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response)
+			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response, opts)
 
 			// Note: The runTopDownTestCase ends up evaluating twice (once with and once without partial
 			// eval first), so expect 2x the total request count the test case specified.
@@ -1193,6 +1277,9 @@ func TestHTTPSendInterQueryCachingNewResp(t *testing.T) {
 
 	data := loadSmallTestData()
 
+	t0 := time.Now()
+	opts := setTime(t0)
+
 	for _, tc := range tests {
 		t.Run(tc.note, func(t *testing.T) {
 
@@ -1205,6 +1292,8 @@ func TestHTTPSendInterQueryCachingNewResp(t *testing.T) {
 					headers[k] = v
 				}
 
+				headers.Set("Date", t0.Format(time.RFC850))
+
 				etag := w.Header().Get("etag")
 
 				if r.Header.Get("if-none-match") != "" {
@@ -1213,11 +1302,14 @@ func TestHTTPSendInterQueryCachingNewResp(t *testing.T) {
 					}
 				}
 				w.WriteHeader(http.StatusOK)
-				w.Write([]byte(tc.response))
+				_, err := w.Write([]byte(tc.response))
+				if err != nil {
+					t.Fatal(err)
+				}
 			}))
 			defer ts.Close()
 
-			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response)
+			runTopDownTestCase(t, data, tc.note, []string{strings.ReplaceAll(tc.ruleTemplate, "%URL%", ts.URL)}, tc.response, opts)
 
 			// Note: The runTopDownTestCase ends up evaluating twice (once with and once without partial
 			// eval first), so expect 2x the total request count the test case specified.
@@ -1238,18 +1330,6 @@ func TestInsertIntoHTTPSendInterQueryCacheError(t *testing.T) {
 		response         string
 		expectedReqCount int
 	}{
-		{
-			note: "http.send GET (bad_expires_header_value)",
-			ruleTemplate: `p = x {
-									r1 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true}) # fallback to normal cache
-									r2 = http.send({"method": "get", "url": "%URL%", "force_json_decode": true, "cache": true}) # retrieved from normal cache
-									r1 == r2
-									x = r1.body
-								}`,
-			headers:          map[string][]string{"Cache-Control": {"max-age=0, public"}, "Expires": {"Wed, 32 Dec 2115 07:28:00 GMT"}},
-			response:         `{"x": 1}`,
-			expectedReqCount: 1,
-		},
 		{
 			note: "http.send GET (bad_date_header_value)",
 			ruleTemplate: `p = x {
@@ -1291,7 +1371,10 @@ func TestInsertIntoHTTPSendInterQueryCacheError(t *testing.T) {
 				}
 
 				w.WriteHeader(http.StatusOK)
-				w.Write([]byte(tc.response))
+				_, err := w.Write([]byte(tc.response))
+				if err != nil {
+					t.Fatal(err)
+				}
 			}))
 			defer ts.Close()
 
@@ -1321,12 +1404,12 @@ func TestGetResponseHeaderDateEmpty(t *testing.T) {
 
 func TestIsCachedResponseFreshZeroTime(t *testing.T) {
 	zeroTime := new(time.Time)
-	result := isCachedResponseFresh(BuiltinContext{}, &interQueryCacheValue{date: *zeroTime}, nil)
+	result := isCachedResponseFresh(BuiltinContext{}, &responseHeaders{date: *zeroTime}, nil)
 	if result {
 		t.Fatal("Expected stale cache response")
 	}
 
-	result = isCachedResponseFresh(BuiltinContext{Time: ast.NullTerm()}, &interQueryCacheValue{date: time.Now()}, nil)
+	result = isCachedResponseFresh(BuiltinContext{Time: ast.NullTerm()}, &responseHeaders{date: time.Now()}, nil)
 	if result {
 		t.Fatal("Expected stale cache response")
 	}
@@ -1549,6 +1632,49 @@ func TestInterQueryCheckCacheError(t *testing.T) {
 	}
 }
 
+func TestNewInterQueryCacheValue(t *testing.T) {
+	headers := make(http.Header)
+	headers.Set("test-header", "test-value")
+	headers.Set("Cache-Control", "max-age=290304000, public")
+	headers.Set("Date", "Wed, 31 Dec 2115 07:28:00 GMT")
+
+	// test data
+	var b = []byte(`[{"ID": "1", "Firstname": "John"}]`)
+
+	response := &http.Response{
+		Status:     "200 OK",
+		StatusCode: http.StatusOK,
+		Header:     headers,
+		Request:    &http.Request{Method: "Get"},
+		Body:       ioutil.NopCloser(bytes.NewBuffer(b)),
+	}
+
+	result, err := newInterQueryCacheValue(response, b)
+	if err != nil {
+		t.Fatalf("Unexpected error %v", err)
+	}
+
+	cvd := interQueryCacheData{RespBody: b,
+		Status:     "200 OK",
+		StatusCode: http.StatusOK,
+		Headers:    headers}
+
+	cvdBytes, err := json.Marshal(cvd)
+	if err != nil {
+		t.Fatalf("Unexpected error %v", err)
+	}
+
+	expectedResult := &interQueryCacheValue{Data: cvdBytes}
+
+	if !reflect.DeepEqual(result, expectedResult) {
+		t.Fatalf("Expected result %v but got %v", expectedResult, result)
+	}
+
+	if int64(len(cvdBytes)) != result.SizeInBytes() {
+		t.Fatalf("Expected cache item size %v but got %v", len(cvdBytes), result.SizeInBytes())
+	}
+}
+
 func getTestServer() (baseURL string, teardownFn func()) {
 	mux := http.NewServeMux()
 	ts := httptest.NewServer(mux)
@@ -1582,7 +1708,7 @@ func getTLSTestServer() (ts *httptest.Server) {
 			return
 		}
 		w.Header().Set("Content-Type", "application/json")
-		w.Write(js)
+		_, _ = w.Write(js)
 	})
 
 	mux.HandleFunc("/", func(w http.ResponseWriter, req *http.Request) {
@@ -2107,6 +2233,139 @@ func TestHTTPSNoClientCerts(t *testing.T) {
 	})
 }
 
+func TestCertSelectionLogic(t *testing.T) {
+	const (
+		localCaFile = "testdata/ca.pem"
+	)
+
+	// Set up Environment
+	caCertPEM, err := ioutil.ReadFile(localCaFile)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	caPool := x509.NewCertPool()
+	if ok := caPool.AppendCertsFromPEM(caCertPEM); !ok {
+		t.Fatal("failed to parse CA cert")
+	}
+
+	ca, err := ioutil.ReadFile(localCaFile)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	err = os.Setenv("CLIENT_CA_ENV", string(caCertPEM))
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	getClientTLSConfig := func(obj ast.Object) *tls.Config {
+		_, client, err := createHTTPRequest(BuiltinContext{Context: context.Background()}, obj)
+		if err != nil {
+			t.Fatalf("Unexpected error creating HTTP request %v", err)
+		}
+		if client.Transport == nil {
+			return nil
+		}
+		return client.Transport.(*http.Transport).TLSClientConfig
+	}
+
+	systemCertsPool, err := x509.SystemCertPool()
+	if err != nil {
+		t.Fatalf("Unexpected error reading system certs %v", err)
+	}
+
+	tempSystemCertsPool, err := x509.SystemCertPool()
+	if err != nil {
+		t.Fatalf("Unexpected error reading system certs %v", err)
+	}
+	systemCertsAndCaPool, err := addCACertsFromBytes(tempSystemCertsPool, ca)
+	if err != nil {
+		t.Fatalf("Unexpected error merging system certs and ca certs %v", err)
+	}
+
+	tests := []struct {
+		note     string
+		input    map[*ast.Term]*ast.Term
+		expected [][]byte
+		msg      string
+	}{
+		{
+			note:     "tls_use_system_certs set to true",
+			input:    map[*ast.Term]*ast.Term{ast.StringTerm("tls_use_system_certs"): ast.BooleanTerm(true)},
+			expected: systemCertsPool.Subjects(),
+			msg:      "Expected TLS config to use system certs",
+		},
+		{
+			note:     "tls_use_system_certs set to nil",
+			input:    map[*ast.Term]*ast.Term{ast.StringTerm("tls_use_system_certs"): ast.BooleanTerm(false)},
+			expected: nil,
+			msg:      "Expected no TLS config",
+		},
+		{
+			note:     "no CAs specified",
+			input:    nil,
+			expected: systemCertsPool.Subjects(),
+			msg:      "Expected TLS config to use system certs",
+		},
+		{
+			note:     "CA cert provided directly",
+			input:    map[*ast.Term]*ast.Term{ast.StringTerm("tls_ca_cert"): ast.StringTerm(string(ca))},
+			expected: caPool.Subjects(),
+			msg:      "Expected TLS config to use provided CA certs",
+		},
+		{
+			note:     "CA cert file path provided",
+			input:    map[*ast.Term]*ast.Term{ast.StringTerm("tls_ca_cert_file"): ast.StringTerm(localCaFile)},
+			expected: caPool.Subjects(),
+			msg:      "Expected TLS config to use provided CA certs in file",
+		},
+		{
+			note:     "CA cert provided in env variable",
+			input:    map[*ast.Term]*ast.Term{ast.StringTerm("tls_ca_cert_env_variable"): ast.StringTerm("CLIENT_CA_ENV")},
+			expected: caPool.Subjects(),
+			msg:      "Expected TLS config to use provided CA certs in env variable",
+		},
+		{
+			note: "CA cert provided directly and tls_use_system_certs parameter set to false",
+			input: map[*ast.Term]*ast.Term{
+				ast.StringTerm("tls_ca_cert"):          ast.StringTerm(string(ca)),
+				ast.StringTerm("tls_use_system_certs"): ast.BooleanTerm(false),
+			},
+			expected: caPool.Subjects(),
+			msg:      "Expected TLS config to use provided CA certs only",
+		},
+		{
+			note: "CA cert provided directly and tls_use_system_certs parameter set to true",
+			input: map[*ast.Term]*ast.Term{
+				ast.StringTerm("tls_ca_cert"):          ast.StringTerm(string(ca)),
+				ast.StringTerm("tls_use_system_certs"): ast.BooleanTerm(true),
+			},
+			expected: systemCertsAndCaPool.Subjects(),
+			msg:      "Expected TLS config to use provided CA certs and system certs",
+		},
+	}
+
+	for _, tc := range tests {
+		t.Run(tc.note, func(t *testing.T) {
+			obj := ast.NewObject()
+			for key, value := range tc.input {
+				obj.Insert(key, value)
+			}
+			tlsConfig := getClientTLSConfig(obj)
+			if tc.expected == nil {
+				if tlsConfig != nil {
+					t.Fatalf(tc.msg)
+				}
+			} else {
+				if !reflect.DeepEqual(tlsConfig.RootCAs.Subjects(), tc.expected) {
+					t.Fatal(tc.msg)
+				}
+			}
+		})
+	}
+}
+
 func TestHTTPSendMetrics(t *testing.T) {
 
 	// run test server
