@@ -15,24 +15,24 @@ import (
 	"github.com/open-policy-agent/opa/storage"
 	"github.com/open-policy-agent/opa/storage/inmem"
 	"github.com/open-policy-agent/opa/util"
-	"github.com/open-policy-agent/opa/util/test"
 )
 
 func TestTopDownPartialEval(t *testing.T) {
 
 	tests := []struct {
-		note            string
-		unknowns        []string
-		disableInlining []string
-		shallow         bool
-		query           string
-		modules         []string
-		data            string
-		input           string
-		wantQueries     []string
-		wantQueryASTs   []ast.Body
-		wantSupport     []string
-		ignoreOrder     bool
+		note                 string
+		unknowns             []string
+		disableInlining      []string
+		shallow              bool
+		skipPartialNamespace bool
+		query                string
+		modules              []string
+		data                 string
+		input                string
+		wantQueries          []string
+		wantQueryASTs        []ast.Body
+		wantSupport          []string
+		ignoreOrder          bool
 	}{
 		{
 			note:        "empty",
@@ -2270,6 +2270,47 @@ func TestTopDownPartialEval(t *testing.T) {
 			},
 			wantQueries: []string{`input.foo`, `input.foo`},
 		},
+		{
+			note:  "package path copied when skip partial namespace enabled (bug 3302)",
+			query: "data.test.p = x",
+			modules: []string{`
+				package test
+				pkg = "foo" { input.x = "foo" }
+				pkg = "bar" { input.x = "bar" }
+				p = x { k = pkg; x = data.other[k].p }
+			`, `
+				package other.foo
+				p = 1 { input = a }
+			`, `
+				package other.bar
+				p = 2 { input = a }
+			`},
+			wantQueries: []string{"data.test.p = x"},
+			wantSupport: []string{
+				`
+					package other.foo
+
+					p = 1 { input = a5 }
+				`,
+				`
+					package other.bar
+
+					p = 2 { input = a4 }
+				`,
+				`
+					package test
+
+					pkg = "foo" { input.x = "foo" }
+
+					pkg = "bar" { input.x = "bar" }
+
+					p = x1 { data.test.pkg = k1; "bar" = k1; data.other[k1].p = x1 }
+					p = x1 { data.test.pkg = k1; "foo" = k1; data.other[k1].p = x1 }
+				`,
+			},
+			shallow:              true,
+			skipPartialNamespace: true,
+		},
 	}
 
 	ctx := context.Background()
@@ -2312,6 +2353,7 @@ func TestTopDownPartialEval(t *testing.T) {
 				WithTracer(&buf).
 				WithUnknowns(unknowns).
 				WithDisableInlining(disableInlining).
+				WithSkipPartialNamespace(tc.skipPartialNamespace).
 				WithShallowInlining(tc.shallow)
 
 			// Set genvarprefix so that tests can refer to vars in generated
@@ -2377,7 +2419,7 @@ type fixture struct {
 
 func prepareTest(ctx context.Context, t *testing.T, params fixtureParams, f func(context.Context, *testing.T, fixture)) {
 
-	test.Subtest(t, params.note, func(t *testing.T) {
+	t.Run(params.note, func(t *testing.T) {
 
 		var store storage.Store
 
@@ -2388,7 +2430,7 @@ func prepareTest(ctx context.Context, t *testing.T, params fixtureParams, f func
 			store = inmem.New()
 		}
 
-		storage.Txn(ctx, store, storage.TransactionParams{}, func(txn storage.Transaction) error {
+		err := storage.Txn(ctx, store, storage.TransactionParams{}, func(txn storage.Transaction) error {
 
 			compiler := ast.NewCompiler()
 			modules := map[string]*ast.Module{}
@@ -2425,6 +2467,9 @@ func prepareTest(ctx context.Context, t *testing.T, params fixtureParams, f func
 
 			return nil
 		})
+		if err != nil {
+			t.Fatal(err)
+		}
 	})
 }
 
