@@ -31,7 +31,6 @@ import (
 	"github.com/open-policy-agent/opa/storage/inmem"
 	"github.com/open-policy-agent/opa/types"
 	"github.com/open-policy-agent/opa/util"
-	testutil "github.com/open-policy-agent/opa/util/test"
 )
 
 func TestTopDownQueryIDsUnique(t *testing.T) {
@@ -406,10 +405,16 @@ func loadSmallTestData() map[string]interface{} {
 	return data
 }
 
-func runTopDownTestCase(t *testing.T, data map[string]interface{}, note string, rules []string, expected interface{}) {
+func setTime(t time.Time) func(*Query) *Query {
+	return func(q *Query) *Query {
+		return q.WithTime(t)
+	}
+}
+
+func runTopDownTestCase(t *testing.T, data map[string]interface{}, note string, rules []string, expected interface{}, options ...func(*Query) *Query) {
 	t.Helper()
 
-	runTopDownTestCaseWithContext(context.Background(), t, data, note, rules, nil, "", expected)
+	runTopDownTestCaseWithContext(context.Background(), t, data, note, rules, nil, "", expected, options...)
 }
 
 func runTopDownTestCaseWithModules(t *testing.T, data map[string]interface{}, note string, rules []string, modules []string, input string, expected interface{}) {
@@ -418,7 +423,8 @@ func runTopDownTestCaseWithModules(t *testing.T, data map[string]interface{}, no
 	runTopDownTestCaseWithContext(context.Background(), t, data, note, rules, modules, input, expected)
 }
 
-func runTopDownTestCaseWithContext(ctx context.Context, t *testing.T, data map[string]interface{}, note string, rules []string, modules []string, input string, expected interface{}) {
+func runTopDownTestCaseWithContext(ctx context.Context, t *testing.T, data map[string]interface{}, note string, rules []string, modules []string, input string, expected interface{},
+	options ...func(*Query) *Query) {
 	t.Helper()
 
 	imports := []string{}
@@ -438,16 +444,11 @@ func runTopDownTestCaseWithContext(ctx context.Context, t *testing.T, data map[s
 
 	store := inmem.NewFromObject(data)
 
-	assertTopDownWithPathAndContext(ctx, t, compiler, store, note, []string{"generated", "p"}, input, expected)
+	assertTopDownWithPathAndContext(ctx, t, compiler, store, note, []string{"generated", "p"}, input, expected, options...)
 }
 
-func assertTopDownWithPath(t *testing.T, compiler *ast.Compiler, store storage.Store, note string, path []string, input string, expected interface{}) {
-	t.Helper()
-
-	assertTopDownWithPathAndContext(context.Background(), t, compiler, store, note, path, input, expected)
-}
-
-func assertTopDownWithPathAndContext(ctx context.Context, t *testing.T, compiler *ast.Compiler, store storage.Store, note string, path []string, input string, expected interface{}) {
+func assertTopDownWithPathAndContext(ctx context.Context, t *testing.T, compiler *ast.Compiler, store storage.Store, note string, path []string, input string, expected interface{},
+	options ...func(*Query) *Query) {
 	t.Helper()
 
 	var inputTerm *ast.Term
@@ -490,12 +491,20 @@ func assertTopDownWithPathAndContext(ctx context.Context, t *testing.T, compiler
 	config, _ := iCache.ParseCachingConfig(nil)
 	interQueryCache := iCache.NewInterQueryCache(config)
 
+	var strictBuiltinErrors bool
+
+	switch expected.(type) {
+	case *Error, error:
+		strictBuiltinErrors = true
+	}
+
 	query := NewQuery(body).
 		WithCompiler(compiler).
 		WithStore(store).
 		WithTransaction(txn).
 		WithInput(inputTerm).
-		WithInterQueryBuiltinCache(interQueryCache)
+		WithInterQueryBuiltinCache(interQueryCache).
+		WithStrictBuiltinErrors(strictBuiltinErrors)
 
 	var tracer BufferTracer
 
@@ -503,7 +512,11 @@ func assertTopDownWithPathAndContext(ctx context.Context, t *testing.T, compiler
 		query = query.WithTracer(&tracer)
 	}
 
-	testutil.Subtest(t, note, func(t *testing.T) {
+	for _, opt := range options {
+		query = opt(query)
+	}
+
+	t.Run(note, func(t *testing.T) {
 		t.Helper()
 
 		switch e := expected.(type) {
