@@ -1,7 +1,11 @@
 #ifndef OPA_MALLOC_H
 #define OPA_MALLOC_H
 
-#include <stdint.h>
+#include <stddef.h>
+
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 void *opa_malloc(size_t size);
 void opa_free(void *ptr);
@@ -12,6 +16,14 @@ unsigned int opa_heap_top_get(void);
 void opa_heap_ptr_set(unsigned int);
 void opa_heap_top_set(unsigned int);
 
+void *opa_builtin_cache_get(size_t i);
+void opa_builtin_cache_set(size_t i, void *p);
+
+void opa_heap_compact(void);
 size_t opa_heap_free_blocks(void);
 
+#ifdef __cplusplus
+}
+#endif
+
 #endif
