@@ -1,10 +1,12 @@
 #include "malloc.h"
 #include "mpd.h"
+#include "std.h"
 #include "str.h"
 #include "string.h"
 #include "strings.h"
 #include "unicode.h"
 
+OPA_BUILTIN
 opa_value *opa_strings_concat(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING)
@@ -129,6 +131,7 @@ static int strings_indexof(opa_string_t *s, int pos, opa_string_t *substr)
     return -1;
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_contains(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -139,9 +142,10 @@ opa_value *opa_strings_contains(opa_value *a, opa_value *b)
     opa_string_t *s = opa_cast_string(a);
     opa_string_t *substr = opa_cast_string(b);
 
-    return opa_boolean(strings_indexof(s, 0, substr) >= 0 ? TRUE : FALSE);
+    return opa_boolean(strings_indexof(s, 0, substr) >= 0);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_endswith(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -154,20 +158,21 @@ opa_value *opa_strings_endswith(opa_value *a, opa_value *b)
 
     if (s->len < suffix->len)
     {
-        return opa_boolean(FALSE);
+        return opa_boolean(false);
     }
 
     for (int i = 0; i < suffix->len; i++)
     {
         if (s->v[s->len - suffix->len + i] != suffix->v[i])
         {
-            return opa_boolean(FALSE);
+            return opa_boolean(false);
         }
     }
 
-    return opa_boolean(TRUE);
+    return opa_boolean(true);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_format_int(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_NUMBER || opa_value_type(b) != OPA_NUMBER)
@@ -209,29 +214,32 @@ opa_value *opa_strings_format_int(opa_value *a, opa_value *b)
 
     mpd_t *i = mpd_qnew();
     uint32_t status = 0;
-
-    mpd_qround_to_intx(i, input, mpd_max_ctx(), &status);
-    if (status & ~MPD_Rounded)
+    mpd_qtrunc(i, input, mpd_max_ctx(), &status);
+    if (status != 0)
     {
-        opa_abort("strings: invalid rounding");
+        opa_abort("strings: truncate failed");
     }
 
-    opa_value *n = opa_bf_to_number(i);
-    opa_number_try_int(opa_cast_number(n), &v);
+    int32_t w = mpd_qget_i32(i, &status);
+    if (status != 0)
+    {
+        opa_abort("strings: get uint failed");
+    }
 
     char *str = opa_malloc(21); // enough for int_t (with sign).
 
-    if (v < 0)
+    if (w < 0)
     {
         str[0] = '-';
-        snprintf(&str[1], 21, format, -v);
+        snprintf(&str[1], 21, format, -w);
     } else {
-        snprintf(str, 21, format, v);
+        snprintf(str, 21, format, w);
     }
 
     return opa_string_allocated(str, opa_strlen(str));
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_indexof(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -241,10 +249,26 @@ opa_value *opa_strings_indexof(opa_value *a, opa_value *b)
 
     opa_string_t *s = opa_cast_string(a);
     opa_string_t *substr = opa_cast_string(b);
+    int n = strings_indexof(s, 0, substr);
 
-    return opa_number_int(strings_indexof(s, 0, substr));
+    if (n < 0)
+    {
+        return opa_number_int(n);
+    }
+
+    int units = 0;
+    for (int i = 0, len = 0; i < n; units++, i += len)
+    {
+        if (opa_unicode_decode_utf8(s->v, i, s->len, &len) == -1)
+        {
+            opa_abort("string: invalid unicode");
+        }
+    }
+
+    return opa_number_int(units);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_replace(opa_value *a, opa_value *b, opa_value *c)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING || opa_value_type(c) != OPA_STRING)
@@ -287,6 +311,7 @@ opa_value *opa_strings_replace(opa_value *a, opa_value *b, opa_value *c)
     return opa_string_allocated(r, j);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_replace_n(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_OBJECT || opa_value_type(b) != OPA_STRING)
@@ -326,6 +351,7 @@ opa_value *opa_strings_replace_n(opa_value *a, opa_value *b)
     return result;
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_split(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -361,7 +387,7 @@ opa_value *opa_strings_split(opa_value *a, opa_value *b)
     }
 
     int j = 0;
-    for (int i = 0; i <= (s->len - d->len); )
+    for (int i = 0; s->len >= d->len && i <= (s->len - d->len); )
     {
         if (opa_strncmp(&s->v[i], d->v, d->len) == 0)
         {
@@ -389,6 +415,7 @@ opa_value *opa_strings_split(opa_value *a, opa_value *b)
     return &arr->hdr;
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_startswith(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -401,12 +428,13 @@ opa_value *opa_strings_startswith(opa_value *a, opa_value *b)
 
     if (s->len < prefix->len)
     {
-        return opa_boolean(FALSE);
+        return opa_boolean(false);
     }
 
-    return opa_strncmp(s->v, prefix->v, prefix->len) == 0 ? opa_boolean(TRUE) : opa_boolean(FALSE);
+    return opa_boolean(opa_strncmp(s->v, prefix->v, prefix->len) == 0);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_substring(opa_value *a, opa_value *b, opa_value *c)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_NUMBER || opa_value_type(c) != OPA_NUMBER)
@@ -427,31 +455,55 @@ opa_value *opa_strings_substring(opa_value *a, opa_value *b, opa_value *c)
         return NULL;
     }
 
-    if (start >= base->len)
-    {
-        return opa_string_terminated("");
-    }
-
     if (start < 0)
     {
         return NULL;
     }
 
-    if (length < 0)
+    if (length == 0)
     {
-        length = base->len - start;
+        return opa_string_terminated("");
     }
 
-    if (base->len < (start + length))
+    size_t spos = base->len, epos = base->len;
+    for (int i = 0, units = 0, len = 0; i < base->len; units++, i += len)
     {
-        length = base->len - start;
+        if (units == start)
+        {
+            spos = i;
+        }
+
+        if (opa_unicode_decode_utf8(base->v, i, base->len, &len) == -1)
+        {
+            opa_abort("string: invalid unicode");
+        }
+
+        if (units < start)
+        {
+            // Start index not reached yet.
+            continue;
+        }
+
+        if (length < 0)
+        {
+            // Everything from start to end.
+            break;
+        }
+
+        if (length == (units - start))
+        {
+            epos = i;
+            break;
+        }
     }
 
-    char *str = opa_malloc(length + 1);
-    memcpy(str, &base->v[start], length + 1);
-    return opa_string_allocated(str, length);
+    char *str = opa_malloc(epos - spos + 1);
+    memcpy(str, &base->v[spos], epos - spos);
+    str[epos - spos] = 0;
+    return opa_string_allocated(str, epos - spos);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -465,6 +517,7 @@ opa_value *opa_strings_trim(opa_value *a, opa_value *b)
     return r;
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim_left(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -507,6 +560,7 @@ opa_value *opa_strings_trim_left(opa_value *a, opa_value *b)
     return opa_string_allocated(str, s->len - j);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim_prefix(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -529,6 +583,7 @@ opa_value *opa_strings_trim_prefix(opa_value *a, opa_value *b)
     return opa_string_allocated(str, len);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim_right(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -578,6 +633,7 @@ opa_value *opa_strings_trim_right(opa_value *a, opa_value *b)
     return opa_string_allocated(str, j);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim_suffix(opa_value *a, opa_value *b)
 {
     if (opa_value_type(a) != OPA_STRING || opa_value_type(b) != OPA_STRING)
@@ -648,6 +704,7 @@ static opa_value *trim_space(const char *s, int start, int end)
     return opa_string_allocated(str, end - start);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_trim_space(opa_value *a)
 {
     if (opa_value_type(a) != OPA_STRING)
@@ -693,6 +750,7 @@ opa_value *opa_strings_trim_space(opa_value *a)
     return opa_string_allocated(str, stop - start);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_lower(opa_value *a)
 {
     if (opa_value_type(a) != OPA_STRING)
@@ -701,7 +759,7 @@ opa_value *opa_strings_lower(opa_value *a)
     }
 
     opa_string_t *s = opa_cast_string(a);
-    int is_ascii = TRUE;
+    int is_ascii = true;
 
     for (int i = 0; i < s->len && is_ascii; i++)
     {
@@ -767,6 +825,7 @@ opa_value *opa_strings_lower(opa_value *a)
     return opa_string_allocated(out, j);
 }
 
+OPA_BUILTIN
 opa_value *opa_strings_upper(opa_value *a)
 {
     if (opa_value_type(a) != OPA_STRING)
@@ -775,7 +834,7 @@ opa_value *opa_strings_upper(opa_value *a)
     }
 
     opa_string_t *s = opa_cast_string(a);
-    int is_ascii = TRUE;
+    int is_ascii = true;
 
     for (int i = 0; i < s->len && is_ascii; i++)
     {
