@@ -1,5 +1,6 @@
 #include <mpdecimal.h>
 
+#include "json.h"
 #include "malloc.h"
 #include "mpd.h"
 #include "str.h"
@@ -17,43 +18,39 @@ static opa_array_t *__opa_set_values(opa_set_t *set);
 static void __opa_object_insert_elem(opa_object_t *obj, opa_object_elem_t *new, size_t hash);
 static void __opa_set_add_elem(opa_set_t *set, opa_set_elem_t *new, size_t hash);
 
+OPA_INTERNAL
 int opa_value_type(opa_value *node)
 {
-	return node->type;
+    // For all intents and purposes, interned strings are strings,
+    // interned booleans are booleans.
+    // Only opa_value_free and opa_value_shallow_copy handle them
+    // separately, by refering to node->type directly.
+    switch (node->type)
+    {
+    case OPA_STRING_INTERNED:
+        return OPA_STRING;
+    case OPA_BOOLEAN_INTERNED:
+        return OPA_BOOLEAN;
+    default:
+        return node->type;
+    }
 }
 
 opa_value *opa_value_get_object(opa_object_t *obj, opa_value *key)
 {
     opa_object_elem_t *elem = opa_object_get(obj, key);
-
-    if (elem != NULL)
-    {
-        return elem->v;
-    }
-
-    return NULL;
+    return elem == NULL ? NULL : elem->v;
 }
 
 opa_value *opa_value_get_set(opa_set_t *set, opa_value *key)
 {
     opa_set_elem_t *elem = opa_set_get(set, key);
-
-    if (elem != NULL)
-    {
-        return elem->v;
-    }
-
-    return NULL;
+    return elem == NULL ? NULL : elem->v;
 }
 
 opa_value *opa_value_get_array_native(opa_array_t *arr, long long i)
 {
-    if (i >= arr->len)
-    {
-        return NULL;
-    }
-
-    return arr->elems[i].v;
+    return i >= arr->len ? NULL : arr->elems[i].v;
 }
 
 opa_value *opa_value_get_array(opa_array_t *arr, opa_value *key)
@@ -80,6 +77,7 @@ opa_value *opa_value_get_array(opa_array_t *arr, opa_value *key)
     return opa_value_get_array_native(arr, i);
 }
 
+OPA_INTERNAL
 opa_value *opa_value_get(opa_value *node, opa_value *key)
 {
     if (node != NULL)
@@ -273,9 +271,10 @@ size_t opa_value_length_string(opa_string_t *str)
     return str->len;
 }
 
+OPA_INTERNAL
 size_t opa_value_length(opa_value *node)
 {
-    switch (node->type)
+    switch (opa_value_type(node))
     {
     case OPA_ARRAY:
         return opa_value_length_array(opa_cast_array(node));
@@ -320,14 +319,6 @@ int opa_value_compare_number(opa_number_t *a, opa_number_t *b)
         return 0;
     }
 
-    if (a->repr == OPA_NUMBER_REPR_FLOAT && b->repr == OPA_NUMBER_REPR_FLOAT)
-    {
-        double da = opa_number_as_float(a);
-        double db = opa_number_as_float(b);
-
-        return opa_value_compare_float(da, db);
-    }
-
     mpd_t *ba = opa_number_to_bf(&a->hdr);
     mpd_t *bb = opa_number_to_bf(&b->hdr);
 
@@ -480,31 +471,31 @@ int opa_value_compare_set(opa_set_t *a, opa_set_t *b)
     return 0;
 }
 
+OPA_INTERNAL
 int opa_value_compare(opa_value *a, opa_value *b)
 {
-    if (a == NULL && b == NULL)
+    if (a == b)
     {
         return 0;
     }
-    else if (b == NULL)
+    if (b == NULL)
     {
         return 1;
     }
-    else if (a == NULL)
+    if (a == NULL)
     {
         return -1;
     }
-
-    if (a->type < b->type)
+    if (opa_value_type(a) < opa_value_type(b))
     {
         return -1;
     }
-    else if (b->type < a->type)
+    if (opa_value_type(b) < opa_value_type(a))
     {
         return 1;
     }
 
-    switch (a->type)
+    switch (opa_value_type(a))
     {
     case OPA_NULL:
         return 0;
@@ -570,7 +561,7 @@ fnv1a32(size_t hash, const void *input, size_t len)
 }
 
 size_t opa_boolean_hash(opa_boolean_t *b) {
-    return b->v == FALSE ? 0 : 1;
+    return b->v ? 0 : 1;
 }
 
 size_t opa_number_hash(opa_number_t *n) {
@@ -630,7 +621,7 @@ size_t opa_set_hash(opa_set_t *o) {
 }
 
 size_t opa_value_hash(opa_value *node) {
-    switch (node->type)
+    switch (opa_value_type(node))
     {
     case OPA_NULL:
         return 0;
@@ -648,12 +639,13 @@ size_t opa_value_hash(opa_value *node) {
         return opa_set_hash(opa_cast_set(node));
     }
 
-    return NULL;
+    return 0;
 }
 
+OPA_INTERNAL
 void opa_value_free(opa_value *node)
 {
-    switch (node->type)
+    switch (node->type) // bypass opa_value_type: don't free OPA_STRING_INTERNED
     {
     case OPA_NULL:
         opa_free(node);
@@ -679,11 +671,16 @@ void opa_value_free(opa_value *node)
     }
 }
 
+OPA_INTERNAL
 opa_value *opa_value_merge(opa_value *a, opa_value *b)
 {
+    if (a == NULL)
+    {
+        return b;
+    }
     if (opa_value_type(a) != OPA_OBJECT || opa_value_type(b) != OPA_OBJECT)
     {
-        return NULL;
+        return a;
     }
 
     opa_object_t *obj = opa_cast_object(a);
@@ -747,8 +744,6 @@ opa_value *opa_value_shallow_copy_number(opa_number_t *n)
 {
     switch (n->repr)
     {
-    case OPA_NUMBER_REPR_FLOAT:
-        return opa_number_float(n->v.f);
     case OPA_NUMBER_REPR_REF:
         return opa_number_ref(n->v.ref.s, n->v.ref.len);
     case OPA_NUMBER_REPR_INT:
@@ -811,7 +806,7 @@ opa_value *opa_value_shallow_copy_set(opa_set_t *s)
 
 opa_value *opa_value_shallow_copy(opa_value *node)
 {
-    switch (node->type)
+    switch (node->type) // bypass opa_value_type: pass OPA_STRING_INTERNED along as-is
     {
     case OPA_NULL:
         return node;
@@ -827,11 +822,50 @@ opa_value *opa_value_shallow_copy(opa_value *node)
         return opa_value_shallow_copy_object(opa_cast_object(node));
     case OPA_SET:
         return opa_value_shallow_copy_set(opa_cast_set(node));
+    case OPA_STRING_INTERNED:
+    case OPA_BOOLEAN_INTERNED:
+        return node;
     }
 
     return NULL;
 }
 
+static opa_value *__opa_tuple(opa_value *a, opa_value *b)
+{
+    opa_value *ret = opa_array_with_cap(2);
+    opa_array_t *tuple = opa_cast_array(ret);
+    opa_array_append(tuple, a);
+    opa_array_append(tuple, b);
+    return ret;
+}
+
+static void __opa_value_transitive_closure(opa_array_t *result, opa_array_t *path, opa_value *node)
+{
+    opa_array_append(result, __opa_tuple(&path->hdr, node));
+    opa_value *prev = NULL;
+    opa_value *curr = NULL;
+
+    while ((curr = opa_value_iter(node, prev)) != NULL)
+    {
+        opa_array_t *cpy = opa_cast_array(opa_value_shallow_copy_array(path));
+        opa_array_append(cpy, curr);
+        opa_value *child = opa_value_get(node, curr);
+        __opa_value_transitive_closure(result, cpy, child);
+        prev = curr;
+    }
+}
+
+OPA_BUILTIN
+opa_value *opa_value_transitive_closure(opa_value *v)
+{
+    opa_array_t *result = opa_cast_array(opa_array());
+    opa_array_t *path = opa_cast_array(opa_array());
+    __opa_value_transitive_closure(result, path, v);
+    return &result->hdr;
+}
+
+
+OPA_INTERNAL
 opa_value *opa_null()
 {
     opa_value *ret = (opa_value *)opa_malloc(sizeof(opa_value));
@@ -839,7 +873,7 @@ opa_value *opa_null()
     return ret;
 }
 
-opa_value *opa_boolean(int v)
+opa_value *opa_boolean_allocated(bool v)
 {
     opa_boolean_t *ret = (opa_boolean_t *)opa_malloc(sizeof(opa_boolean_t));
     ret->hdr.type = OPA_BOOLEAN;
@@ -847,6 +881,12 @@ opa_value *opa_boolean(int v)
     return &ret->hdr;
 }
 
+opa_value *opa_boolean(bool v)
+{
+    return opa_boolean_allocated(v);
+}
+
+OPA_INTERNAL
 opa_value *opa_number_size(size_t v)
 {
     opa_number_t *ret = (opa_number_t *)opa_malloc(sizeof(opa_number_t));
@@ -856,6 +896,7 @@ opa_value *opa_number_size(size_t v)
     return &ret->hdr;
 }
 
+OPA_INTERNAL
 opa_value *opa_number_int(long long v)
 {
     opa_number_t *ret = (opa_number_t *)opa_malloc(sizeof(opa_number_t));
@@ -865,15 +906,7 @@ opa_value *opa_number_int(long long v)
     return &ret->hdr;
 }
 
-opa_value *opa_number_float(double v)
-{
-    opa_number_t *ret = (opa_number_t *)opa_malloc(sizeof(opa_number_t));
-    ret->hdr.type = OPA_NUMBER;
-    ret->repr = OPA_NUMBER_REPR_FLOAT;
-    ret->v.f = v;
-    return &ret->hdr;
-}
-
+OPA_INTERNAL
 opa_value *opa_number_ref(const char *s, size_t len)
 {
     opa_number_t *ret = (opa_number_t *)opa_malloc(sizeof(opa_number_t));
@@ -896,6 +929,13 @@ opa_value *opa_number_ref_allocated(const char *s, size_t len)
     return &ret->hdr;
 }
 
+void opa_number_init_int(opa_number_t *n, long long v)
+{
+    n->hdr.type = OPA_NUMBER;
+    n->repr = OPA_NUMBER_REPR_INT;
+    n->v.i = v;
+}
+
 void opa_number_free(opa_number_t *n)
 {
     if (n->repr == OPA_NUMBER_REPR_REF)
@@ -913,8 +953,6 @@ int opa_number_try_int(opa_number_t *n, long long *i)
 {
     switch (n->repr)
     {
-    case OPA_NUMBER_REPR_FLOAT:
-        return -1;
     case OPA_NUMBER_REPR_INT:
         *i = n->v.i;
         return 0;
@@ -930,8 +968,6 @@ double opa_number_as_float(opa_number_t *n)
 {
     switch (n->repr)
     {
-    case OPA_NUMBER_REPR_FLOAT:
-        return n->v.f;
     case OPA_NUMBER_REPR_INT:
         return (double)n->v.i;
     case OPA_NUMBER_REPR_REF:
@@ -960,6 +996,7 @@ opa_value *opa_string(const char *v, size_t len)
     return &ret->hdr;
 }
 
+OPA_INTERNAL
 opa_value *opa_string_terminated(const char *v)
 {
     opa_string_t *ret = (opa_string_t *)opa_malloc(sizeof(opa_string_t));
@@ -1020,6 +1057,7 @@ opa_value *opa_array()
     return opa_array_with_cap(0);
 }
 
+OPA_INTERNAL
 opa_value *opa_array_with_cap(size_t cap)
 {
     opa_array_t *ret = (opa_array_t *)opa_malloc(sizeof(opa_array_t));
@@ -1083,6 +1121,7 @@ static opa_value *__opa_set_with_buckets(size_t buckets)
     return &ret->hdr;
 }
 
+OPA_INTERNAL
 opa_value *opa_set()
 {
     return __opa_set_with_buckets(OPA_SET_MIN_BUCKETS);
@@ -1100,12 +1139,7 @@ opa_value *opa_set_with_cap(size_t n)
     return __opa_set_with_buckets(buckets);
 }
 
-void opa_value_boolean_set(opa_value *v, int b)
-{
-    opa_boolean_t *ret = opa_cast_boolean(v);
-    ret->v = b;
-}
-
+OPA_INTERNAL
 void opa_value_number_set_int(opa_value *v, long long i)
 {
 	opa_number_t *ret = opa_cast_number(v);
@@ -1128,6 +1162,7 @@ void opa_array_free(opa_array_t *arr)
     opa_free(arr);
 }
 
+OPA_INTERNAL
 void opa_array_append(opa_array_t *arr, opa_value *v)
 {
     if (arr->len >= arr->cap)
@@ -1242,6 +1277,7 @@ void __opa_object_grow(opa_object_t *obj, size_t n) {
     opa_free(dst);
 }
 
+OPA_INTERNAL
 void opa_object_insert(opa_object_t *obj, opa_value *k, opa_value *v)
 {
     size_t hash = opa_value_hash(k);
@@ -1279,6 +1315,35 @@ static void __opa_object_insert_elem(opa_object_t *obj, opa_object_elem_t *new,
     obj->len++;
 }
 
+void opa_object_remove(opa_object_t *obj, opa_value *k)
+{
+    size_t hash = opa_value_hash(k);
+
+    size_t i = hash % obj->n;
+    opa_object_elem_t **prev = &obj->buckets[i];
+    opa_object_elem_t *curr = obj->buckets[i];
+    while (curr != NULL)
+    {
+        if (opa_value_compare(curr->k, k) == 0)
+        {
+            *prev = curr->next;
+            obj->len--;
+
+            opa_value_free(curr->k);
+            opa_value_free(curr->v);
+            opa_free(curr);
+
+            // TODO: Consider shrinking the object size. For now it will remain
+            // with its current size.
+
+            return;
+        }
+        prev = &curr->next;
+        curr = curr->next;
+    }
+    return;  // Key wasn't found, consider it deleted.
+}
+
 opa_object_elem_t *opa_object_get(opa_object_t *obj, opa_value *key)
 {
     size_t hash = opa_value_hash(key) % obj->n;
@@ -1378,6 +1443,7 @@ void __opa_set_grow(opa_set_t *set, size_t n) {
     opa_free(dst);
 }
 
+OPA_INTERNAL
 void opa_set_add(opa_set_t *set, opa_value *v)
 {
     size_t hash = opa_value_hash(v);
@@ -1428,3 +1494,186 @@ opa_set_elem_t *opa_set_get(opa_set_t *set, opa_value *v)
 
     return NULL;
 }
+
+// Validate that a path is non-null, an array value type,
+// has a length >0, and only contains strings.
+// Returns the size of the path, or -1 for an invalid path
+int _validate_json_path(opa_value *path)
+{
+    if (path == NULL || opa_value_type(path) != OPA_ARRAY)
+    {
+        return -1;
+    }
+
+    int path_len = opa_value_length(path);
+    if (path_len == 0)
+    {
+        return -1;
+    }
+
+    for (int i = 0; i < path_len-1; i++)
+    {
+        opa_value *v = opa_value_get_array_native(opa_cast_array(path), i);
+        if (opa_value_type(v) != OPA_STRING)
+        {
+            return -1;
+        }
+    }
+
+    return path_len;
+}
+
+// For the given `data` value set the provided value `v` at
+// the specified `path`. Requires objects for containers,
+// any portion of the path that is missing will be created.
+// The `path` must be an `opa_array_t` with at least one
+// element.
+//
+// Replaced objects will be freed.
+WASM_EXPORT(opa_value_add_path)
+opa_errc opa_value_add_path(opa_value *data, opa_value *path, opa_value *v)
+{
+    int path_len = _validate_json_path(path);
+
+    if (path_len < 1)
+    {
+        return OPA_ERR_INVALID_PATH;
+    }
+
+    // Follow the path, creating objects as needed.
+    opa_array_t *p = opa_cast_array(path);
+    opa_value *curr = data;
+
+    for (int i = 0; i < path_len-1; i++)
+    {
+        opa_value *k = opa_value_get_array_native(p, i);
+
+        opa_value *next = opa_value_get(curr, k);
+
+        if (next == NULL)
+        {
+            switch (curr->type)
+            {
+                case OPA_OBJECT:
+                    next = opa_object();
+                    opa_object_insert(opa_cast_object(curr), k, next);
+                    break;
+                default:
+                    return OPA_ERR_INVALID_TYPE;
+            }
+        }
+
+        curr = next;
+    }
+
+    opa_value *k = opa_value_get_array_native(p, path_len-1);
+
+    opa_value *old = opa_value_get(curr, k);
+
+    switch (curr->type)
+    {
+        case OPA_OBJECT:
+            opa_object_insert(opa_cast_object(curr), k, v);
+            break;
+        default:
+            return OPA_ERR_INVALID_TYPE;
+    }
+
+    if (old != NULL)
+    {
+        opa_value_free(old);
+    }
+
+    return OPA_ERR_OK;
+}
+
+// For the given `data` object delete the entry specified by `path`.
+// The `path` must be an `opa_array_t` with at least one
+// element.
+//
+// Deleted values will be freed.
+WASM_EXPORT(opa_value_remove_path)
+opa_errc opa_value_remove_path(opa_value *data, opa_value *path)
+{
+    int path_len = _validate_json_path(path);
+
+    if (path_len < 1)
+    {
+        return OPA_ERR_INVALID_PATH;
+    }
+
+    // Follow the path into data
+    opa_array_t *p = opa_cast_array(path);
+    opa_value *curr = data;
+
+    for (int i = 0; i < path_len-1; i++)
+    {
+        opa_value *k = opa_value_get_array_native(p, i);
+
+        opa_value *next = opa_value_get(curr, k);
+
+        if (next == NULL)
+        {
+            // We were unable to follow the full
+            // path, consider the target deleted
+            return OPA_ERR_OK;
+        }
+
+       curr = next;
+    }
+
+    opa_object_remove(opa_cast_object(curr), opa_value_get_array_native(p, path_len-1));
+
+    return OPA_ERR_OK;
+}
+
+// Lookup path in the passed mapping object. Returns 0 if it can't
+// be found, or of there's no function index leaf when we've run out
+// of path pieces.
+int opa_lookup(opa_value *mapping, opa_value *path) {
+    int path_len = _validate_json_path(path);
+    if (path_len < 1)
+    {
+        return 0;
+    }
+
+    opa_value *curr = mapping;
+
+    for (opa_value *idx = opa_value_iter(path, NULL); idx != NULL; idx = opa_value_iter(path, idx))
+    {
+        opa_value *key = opa_value_get(path, idx);
+        opa_value *next = opa_value_get(curr, key);
+        if (next == NULL)
+        {
+            return 0;
+        }
+        curr = next;
+    }
+    if (curr->type == OPA_NUMBER) {
+        long long i;
+        if (opa_number_try_int(opa_cast_number(curr), &i) == 0) {
+            return i;
+        }
+    }
+    return 0;
+}
+
+// global variable used for storing the parsed mapping JSON
+static opa_value *mapping;
+
+// Called from the WASM-generated '_initialize' function with the
+// address of the mapping string and its length. Parses the JSON
+// string it expects, sets the *mapping variable accordingly.
+OPA_INTERNAL
+void opa_mapping_init(const char *s, const int l) {
+    if (mapping == NULL) {
+        mapping = opa_json_parse(s, l);
+    }
+}
+
+// Lookup mapped function index from global mapping (initialized by
+// opa_mapping_init).
+OPA_INTERNAL
+int opa_mapping_lookup(opa_value *path) {
+    return opa_lookup(mapping, path);
+}
